/* Project filed -- Intelligent file cache and migration system
 * client.c -- Implementation of a client.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the actual client program.
 * NOTES
 * BUGS
 * TODO
 *    - Clean up, priv function definitions, often used routines to
 *      lib or client-lib, man pages, etc
  * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: client.c,v 1.13 2001/02/04 19:57:14 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <limits.h>
#include <string.h>
#include <signal.h>
#include "global.h"
#include "inet_interface.h"
#include "auth_interface.h"
#include "crypt_interface.h"
#include "misc_interface.h"
#include "configuration_interface.h"
#include "debug_interface.h"
#include "md5_interface.h"
#include "filesend_interface.h"
#include "protocol_interface.h"
#include "server_defs.h"
#include "server_client_defs.h"

/* Private type definitions */

typedef struct {
   char *server_host; /* must be freed since allocated dynamically by
                       * server.c's load_configuration()
                       */
   int server_port;

   char *client_name; /* must be freed */
   char *server_name; /* must be freed */

   char *client_secret; /* must be freed */
   char *server_secret; /* must be freed */
} client_config_t;

typedef struct {
   client_config_t config;

   char *program_name;

   char *file_name;

   direction_t direction;

   fd_t sd;

   protocol_info_t protocol_info;
} client_info_t;

/* Private function definitions */

static int do_processing(client_info_t *info);
static int do_operation(client_info_t *info);
static BOOL load_configuration(const char *config_file_name,
                               client_config_t *config);
static void usage(const char *program_name);
static BOOL get_next_file_read(void *caller_private,
                               fd_t *fd,
                               md5_hash_ptr_t last_file_md5sum,
                               BOOL *eof);
static BOOL get_next_file_write(void *caller_private,
                               fd_t *fd,
                               unsigned int *file_max,
                               md5_hash_ptr_t last_file_md5sum,
                               const BOOL last);

/* Function implementation */

int main(int argc, char **argv) {
   int result;

   if (argc != 4) {
      usage(argv[0]);
      /* NOT REACHED */
   }
   else if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
      perror("signal()");
      result = 1;
   }
   else {
      client_info_t info;
      
      info.program_name = argv[0];
      info.file_name = argv[3];
      
      if (!strcasecmp(argv[1], "save")) {
         info.direction = DIRECTION_TO_CACHE;
      }
      else if (!strcasecmp(argv[1], "restore")) {
         info.direction = DIRECTION_FROM_CACHE;
      }
      else {
         usage(argv[0]);
         /* NOT REACHED */
      }
      
      if (load_configuration(argv[2], &info.config)) {      
         info.sd = inet_connect_to_server(info.config.server_host,
                                          info.config.server_port);
      
         if (info.sd == -1) {
            debug_log(DEBUG_LEVEL_FATAL,
                      "%s: connection to server failed.\n",
                      info.program_name);
            result = 1;
         }
         else {
            result = do_processing(&info);
      
            if (close(info.sd)) {
               debug_perror(DEBUG_LEVEL_FATAL, "close()");
               if (result == 0) {
                  result = 1;
               }       
            }
         }
      
         free(info.config.server_host);
         free(info.config.client_name);
         free(info.config.server_name);
         free(info.config.client_secret);
         free(info.config.server_secret);
      }
      else {
         debug_log(DEBUG_LEVEL_FATAL,
                   "%s: error loading configuration %s.\n",
                   argv[0],
                   argv[1]);
         /* Not freeing party, done by exit() */
         result = 1;
      }
   }

   return result;
}

/* Private function implementation */

static int do_processing(client_info_t *info) {
   client_initial_message_t initial_message;

   initial_message.protocol_version = CLIENT_HELLO_PROTOCOL_VERSION;

   strncpy(&initial_message.client_name[0],
           info->config.client_name,
           sizeof(initial_message.client_name) - 1);
   initial_message.client_name[sizeof(initial_message.client_name) - 1] = '\0';

   strncpy(&initial_message.server_name[0],
           info->config.server_name,
           sizeof(initial_message.server_name) - 1);
   initial_message.server_name[sizeof(initial_message.server_name) - 1] = '\0';

   if (write(info->sd, &initial_message, sizeof(initial_message))
       == sizeof(initial_message)) {
      int result = 1; /* failed */

      /* Prove to the server we know the secret */
      if (!auth_challenge_receive_and_respond
              (info->sd,
               info->config.client_secret,
               strlen(info->config.client_secret))) {
         debug_log(DEBUG_LEVEL_FATAL, "server doesn't know my key.\n");
         return 1;
      }

      /* Verify that the server can hash the cleartext challenge
       * with its server secret.
       */
      if (!auth_random_cleartext_challenge_issue_and_verify
              (info->sd,
               info->config.server_secret,
               strlen(info->config.server_secret))) {
         debug_log(DEBUG_LEVEL_FATAL,
                   "server doesn't know his key, or previous wrong.\n");
         return 1;
      }

      debug_log(DEBUG_LEVEL_INFO,
                "SERVER-CLIENT authentification succeeded.\n");

      /* real starts here  */

      info->protocol_info.decrypt_context_1 = NULL;
      info->protocol_info.encrypt_context_1 = NULL;
      info->protocol_info.decrypt_context_2 = NULL;
      info->protocol_info.encrypt_context_2 = NULL;

      info->protocol_info.decrypt_initial_1 = TRUE;
      info->protocol_info.decrypt_initial_2 = TRUE;
      info->protocol_info.encrypt_initial_1 = TRUE;
      info->protocol_info.encrypt_initial_2 = TRUE;

      if ((info->protocol_info.decrypt_context_1 = crypt_context_allocate())
          && (info->protocol_info.encrypt_context_1 = crypt_context_allocate())
          && (info->protocol_info.decrypt_context_2 = crypt_context_allocate())
          && (info->protocol_info.encrypt_context_2
              = crypt_context_allocate())) {
         unsigned char random_crypt_key[RANDOM_CRYPT_KEY_SIZE + 1]; /* \0 */

         if (protocol_receive_and_decrypt_session_key
                (info->sd,
                 random_crypt_key,
                 sizeof(random_crypt_key) - 1,
                 info->config.server_secret,
                 strlen
                   (info->config.server_secret))) {
            random_crypt_key[sizeof(random_crypt_key) - 1] = '\0';
            if (crypt_context_initialize(info->protocol_info.decrypt_context_1,
                                         random_crypt_key,
                                         sizeof(random_crypt_key) - 1)
	        && crypt_context_initialize(info->protocol_info
                                               .encrypt_context_1,
					    random_crypt_key,
					    sizeof(random_crypt_key) - 1)) {
	       if (protocol_receive_and_decrypt_session_key
                      (info->sd,
                       random_crypt_key,
                       sizeof(random_crypt_key) - 1,
                       info->config.client_secret,
                       strlen
                         (info->config.client_secret))) {
		  if (crypt_context_initialize
                         (info->protocol_info.decrypt_context_2,
                          random_crypt_key,
                          sizeof(random_crypt_key) -1)
                      && crypt_context_initialize
                         (info->protocol_info.encrypt_context_2,
                          random_crypt_key,
                          sizeof(random_crypt_key) -1)) {
                     debug_log(DEBUG_LEVEL_INFO,
                               "SESSION ENCRYPTION KEYS SET.\n");

                     result = do_operation(info);
                  }
		  else {
		     debug_log(DEBUG_LEVEL_FATAL, "weak key 2.\n");
		  }
               }
	       else {
		  debug_log(DEBUG_LEVEL_FATAL,
                            "can't receive session key 2.\n");
	       }
            }
            else {
               debug_log(DEBUG_LEVEL_FATAL, "weak key.\n");
            }
         }
         else {
            debug_log(DEBUG_LEVEL_FATAL, "can't receive session key.\n");
         }
      }
      else {
         debug_log(DEBUG_LEVEL_FATAL, "no crypt context.\n");
      }


      if (info->protocol_info.decrypt_context_1) {
         crypt_context_free(&(info->protocol_info.decrypt_context_1));
         if (info->protocol_info.encrypt_context_1) {
            crypt_context_free(&(info->protocol_info.encrypt_context_1));
            if (info->protocol_info.decrypt_context_2) {
               crypt_context_free(&(info->protocol_info.decrypt_context_2));
               if (info->protocol_info.encrypt_context_2) {
                  crypt_context_free(&(info->protocol_info.encrypt_context_2));
               }
            }
         }
      }

      return result;
   }
   else {
      debug_perror(DEBUG_LEVEL_FATAL, "write(): can't send initial message");
   }

   return 1;
}

static int do_operation(client_info_t *info) {
   int result = 0;
   session_dumper_info_t hello;

   /* BUGS
    *    - server at this time needs this information.
    */
   hello.block_size = 131072;
   hello.blocks = (350*1024*1024)/131072;
   hello.offset = 0;

   /* BUGS
    *    - file name might be silently truncated.
    */
   strncpy(hello.file_name, info->file_name, sizeof(hello.file_name));
   hello.file_name[sizeof(hello.file_name) - 1] = '\0';

   hello.direction = info->direction;
   hello.flags = SESSION_FLAGS_FAST;

   if (protocol_send_data(info->sd,
			  &info->protocol_info,
			  (const unsigned char *) &hello,
			  sizeof(hello),
			  TRUE)) {
      if (info->direction == DIRECTION_TO_CACHE) {
	 debug_log(DEBUG_LEVEL_INFO, "stdin to server ...\n");

         info->protocol_info.user_defined = (void *) TRUE;

         if (!filesend_send_file(info->sd,
                                 (void *) &info->protocol_info,
                                 protocol_encrypt,
                                 get_next_file_read)) {
            result = 1;
            debug_log(DEBUG_LEVEL_ERROR, "filesend_send_file() failed.\n");
         }
      }
      else {
         unsigned char md5_hash[MD5_HASH_STRING_LENGTH];

	 debug_log(DEBUG_LEVEL_INFO, "server to stdout ...\n");

         if (!filesend_receive_file(info->sd,
                                    (void *) &info->protocol_info,
                                    protocol_decrypt,
                                    get_next_file_write,
                                    md5_hash)) {
            result = 1;
            debug_log(DEBUG_LEVEL_ERROR, "filesend_receive_file() failed.\n");
         }
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "Can't say hello.\n");
      result = 1;
   }

   return result;
}


static BOOL load_configuration(const char *config_file_name,
                               client_config_t *config) {
   static const configuration_description_t configuration_array[]
      = {
           {
              FALSE,
              {
                 {
                    "server_host",
                    CONFIGURATION_ITEM_TYPE_STRING
                 },
                 { "" }
              }
           },
           {
              FALSE,
              {
                 {
                    "server_port",
                    CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER
                 },
                 { 0 }
              }
           },
           {
              FALSE,
              {
                 {
                    "client_name",
                    CONFIGURATION_ITEM_TYPE_STRING
                 },
                 { "" }
              }
           },
           {
              FALSE,
              {
                 {
                    "server_name",
                    CONFIGURATION_ITEM_TYPE_STRING
                 },
                 { "" }
              }
           },
           {
              FALSE,
              {
                 {
                    "client_secret",
                    CONFIGURATION_ITEM_TYPE_STRING
                 },
                 { "" }
              }
           },
           {
              FALSE,
              {
                 {
                    "server_secret",
                    CONFIGURATION_ITEM_TYPE_STRING
                 },
                 { "" }
              }
           }
        };
   configuration_context_t context;

   context = configuration_initialize(&configuration_array[0],
                                      sizeof(configuration_array)
                                      / sizeof(configuration_array[0]));

   if (context) {
      if (configuration_load(context, config_file_name)) {
         configuration_item_t item;

         item.type.type = CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER;
         item.type.name = "server_port";
         if (configuration_get(context, &item)) {
            config->server_port = item.object.unsigned_integer;

            item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
            item.type.name = "server_host";

            if (configuration_get(context, &item)) {
               config->server_host = strdup(item.object.string);
               if (config->server_host) {
                  item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                  item.type.name = "client_name";

                  if (configuration_get(context, &item)) {
                     config->client_name = strdup(item.object.string);
                     if (config->client_name) {
                        item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                        item.type.name = "server_name";

                        if (configuration_get(context, &item)) {
                           config->server_name = strdup(item.object.string);
                           if (config->server_name) {
                              item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                              item.type.name = "client_secret";

                              if (configuration_get(context, &item)) {
                                 config->client_secret
                                    = strdup(item.object.string);
                                 if (config->client_secret) {
                                    item.type.type
                                       = CONFIGURATION_ITEM_TYPE_STRING;
                                    item.type.name = "server_secret";

                                    if (configuration_get(context, &item)) {
                                       config->server_secret
                                          = strdup(item.object.string);
                                       if (config->server_secret) {
                                          return TRUE;
                                       }
                                       else {
                                          debug_log(DEBUG_LEVEL_ERROR,
                                                    "no memory for server"
                                                    " secret.\n");
                                       }
                                    }
                                    else {
                                       debug_log(DEBUG_LEVEL_ERROR,
                                                 "missing server secret"
                                                 "in %s.\n",
                                                 config_file_name);
                                    }
                                 }
                                 else {
                                    debug_log(DEBUG_LEVEL_ERROR,
                                              "no memory for client"
                                              " secret.\n");
                                 }
                              }
                              else {
                                 debug_log(DEBUG_LEVEL_ERROR,
                                           "missing client secret in %s.\n",
                                           config_file_name);
                              }
                           }
                           else {
                              debug_log(DEBUG_LEVEL_ERROR,
                                        "no memory for server_name");
                           }
                        }
                        else {
                           debug_log(DEBUG_LEVEL_ERROR,
                                     "missing server_name in %s.\n",
                                     config_file_name);
                        }
                     }
                     else {
                        debug_log(DEBUG_LEVEL_ERROR,
                                  "no memory for client_name");
                     }
                  }
                  else {
                     debug_log(DEBUG_LEVEL_ERROR,
                               "missing client_name in %s.\n",
                               config_file_name);
                  }
               }
               else {
                  debug_log(DEBUG_LEVEL_ERROR, "no memory for server_host");
               }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR,
                         "missing server_host in %s.\n",
                         config_file_name);
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "missing server_port in %s.\n",
                      config_file_name);
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "can't load configuration from file %s.\n",
                   config_file_name);
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "configuration module initialization failed.\n");
   }

   return FALSE;
}

static void usage(const char *program_name) {
   debug_log(DEBUG_LEVEL_FATAL,
             "%s save|restore config-file file-name\n%s: bad args.\n",
             program_name,
             program_name);
   exit(2);
   /* NOT REACHED */
}

static BOOL get_next_file_read(void *caller_private,
                               fd_t *fd,
                               md5_hash_ptr_t last_file_md5sum,
                               BOOL *eof) {
   *eof = FALSE;

   if (((protocol_info_t *) caller_private)->user_defined) {
      *fd = dup(fileno(stdin));
      ((protocol_info_t *) caller_private)->user_defined = FALSE;
      if (*fd == -1) {
         return FALSE;
      }
      else {
         return TRUE;
      }
   }
   else {
      *fd = -1;
      return TRUE;
   }
   /* NOT REACHED */
}

static BOOL get_next_file_write(void *caller_private,
                               fd_t *fd,
                               unsigned int *file_max,
                               md5_hash_ptr_t last_file_md5sum,
                               const BOOL last) {
   if (!last) {
      *fd = dup(fileno(stdout));

      *file_max = UINT_MAX;

      return (*fd != -1);
   }

   return FALSE;
}
