/* Project filed -- Intelligent file cache and migration system
 * auth.c -- Authentification functions.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the auth module. The auth modules offers simple authentification
 *    facilities.
 * NOTES
 * BUGS
 *    - Replay attack. Should add a timestamp and a 5 min
 *      delay, hoping that same random sequences do not
 *      happen within 5 mins, and that servers/clients are
 *      synchronized within 5 mins. However, the challenge space
 *      is quite big, and the random generator very good.
 *    - No timeout while in negociation, DoS possible.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: auth.c,v 1.3 2000/04/12 19:48:29 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include "global.h"
#include "../misc/misc_interface.h"
#include "../md5/md5_interface.h"
#include "auth_interface.h"

/* Type definitions */

#define CHALLENGE_LENGTH 64

/* Private function definitions */

/* NAME
 *    compute_hash
 * DESCRIPTION
 *    Computes a hash from data in buffer and a secret key.
 * RESULT
 *    BOOL TRUE if ok, else error (internal).
 * NOTES
 * BUGS
 * TODO
 */
static BOOL compute_hash(const unsigned char *buffer,
                         const int buffer_length,
                         const unsigned char *secret,
                         const int secret_length,
                         unsigned char *hash);

/* Function implementation */

BOOL auth_challenge_receive_and_respond(fd_t sd,
                                        const unsigned char *secret,
                                        const int secret_length) {
   unsigned char buffer[CHALLENGE_LENGTH];
   unsigned char computed_hash[MD5_HASH_STRING_LENGTH];

   /* Receives the challenge */
   if (misc_all_read(sd, buffer, sizeof(buffer)) != sizeof(buffer)) {
      return FALSE;
   }

   /* Compute the hash */
   if (!compute_hash(buffer,
	 	     sizeof(buffer),
		     secret,
                     secret_length,
		     computed_hash)) {
      return FALSE;
   }

   /* Send the hash */
   if (write(sd, computed_hash, sizeof(computed_hash))
       != sizeof(computed_hash)) {
      perror("write()");
      return FALSE;
   }

   return TRUE; /* fine */
}

BOOL auth_random_cleartext_challenge_issue_and_verify
        (fd_t sd,
         const unsigned char *secret,
         const int secret_length) {
   unsigned char buffer[CHALLENGE_LENGTH];
   unsigned char hash[MD5_HASH_STRING_LENGTH];
   unsigned char computed_hash[MD5_HASH_STRING_LENGTH];

   /* Generate a random challenge */
   if (!misc_get_random_data(buffer, sizeof(buffer))) {
      return FALSE;
   }

   /* Send the challenge */
   if (write(sd, buffer, sizeof(buffer)) != sizeof(buffer)) {
      perror("write()");
      return FALSE;
   }

   /* Receive the hash */
   if (misc_all_read(sd, hash, sizeof(hash)) != sizeof(hash)) {
      return FALSE;
   }


   /* Compute the expected hash */
   if (!compute_hash(buffer,
	 	     sizeof(buffer),
		     secret,
                     secret_length,
		     computed_hash)) {
      return FALSE;
   }

   /* Compare */
   if (memcmp(hash, computed_hash, sizeof(computed_hash))) {
      return FALSE;
   }

   return TRUE; /* fine */
}

/* Private function implementation */

static BOOL compute_hash(const unsigned char *buffer,
                         const int buffer_length,
                         const unsigned char *secret,
                         const int secret_length,
                         unsigned char *hash) {
   BOOL result = FALSE;
   void *md5_context = md5_context_allocate();

   if (md5_context) {
      md5_context_initialize(md5_context);

      md5_compute_hash(md5_context, buffer, buffer_length, hash, FALSE);
      md5_compute_hash(md5_context, secret, secret_length, hash, TRUE);

      md5_context_free(&md5_context);

      result = TRUE;
   }

   return result;
}
