/* Project filed -- Intelligent file cache and migration system
 * buffer.c -- Buffer routines
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the buffer module. The buffer module offers buffered
 *    efficient I/O from and to UNIX file descriptors.
 * NOTES
 * BUGS
 *    - Could have used the process support library, and the messaging
 *      system ...
 *    - Shared memory size and chunk size should be configurable
 *      at run time.
 *    - Maybe some of the shared memory should be locked memory (?)
 *    - Could do initial filling of shared memory if requested by caller
 *      (e.g. disk to tape, in father).
 *    - I/O errors should be tellable from child to father if required.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: buffer.c,v 1.6 2000/08/06 18:46:49 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <fcntl.h>
#include "global.h"
#include "../misc/misc_interface.h"
#include "../debug/debug_interface.h"
#include "buffer_interface.h"

/* Private type definitions */

#define BUFFER_CHUNK_SIZE 32768UL /* I/O size */
#define BUFFER_AMOUNT 64UL
#define BUFFER_SHARED_MEMORY_SIZE (BUFFER_CHUNK_SIZE * BUFFER_AMOUNT)

#undef  BUFFER_DEBUG

#ifdef BUFFER_DEBUG
#define BUFFER_ENABLE_STATISTICS
#ifdef BUFFER_ENABLE_STATISTICS
#define BUFFER_DISPLAY_STATISTICS_AT_END
#define BUFFER_DISPLAY_STATISTICS_AT_FATHER_SLEEP
#endif /* BUFFER_ENABLE_STATISTICS */

#undef  BUFFER_DEBUG_QUEUE
#undef  BUFFER_DEBUG_CHILD
#undef  BUFFER_DEBUG_FATHER
#endif /* BUFFER_DEBUG */

/* BUGS
 *    - Should be defined somewhere else.
 */
#define PIPE_SIZE 4096

#define INVALID_LENGTH 0UL /* for correct completion */
typedef struct {
   addr_t address;
   size_t size;
} buffer_command_t; /* from, and to child */

typedef struct {
   unsigned int next_write;
   unsigned int next_read;

   addr_t buffer;
   size_t size;
   size_t chunk_size;
   unsigned int numbers;
} queue_t;

/* Private function definitions */

/* NAME
 *    child_processing
 * DESCRIPTION
 *    The child main work function.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void child_processing(fd_t output,
                             fd_t from_father,
                             fd_t to_father,
                             addr_t address,
                             size_t size);
/* NAME
 *    father_processing
 * DESCRIPTION
 *    The father main work function.
 * RESULT
 *    BOOL TRUE if succeeded.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL father_processing(fd_t input,
                              fd_t to_child,
                              fd_t from_child,
                              addr_t address,
                              size_t size,
                              int *the_errno,
                              BOOL *input_error,
                              BOOL *child_died,
                              pid_t child_pid);

/* NAME
 *    create_shared_memory_segment
 * DESCRIPTION
 *    Creates a shared memory segment to be used for father/child data
 *    transfer.
 * RESULT
 *    BOOL TRUE if the shared memory segment of size size could be
 *    created properly.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL create_shared_memory_segment(addr_t *address, size_t size);

/* NAME
 *    free_shared_memory_segment
 * DESCRIPTION
 *    Free a shared memory segment, specified by *address and size.
 *    Zeroes *address on return.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void free_shared_memory_segment(addr_t *address, size_t size);

/* NAME
 *    queue_initialize
 * DESCRIPTION
 *    Initializes a queue structure for later usage.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void queue_initialize(queue_t *queue,
                             addr_t addr,
                             size_t size,
                             size_t chunk_size);

/* NAME
 *    queue_slot_allocate_for_writing
 * DESCRIPTION
 *    Allocates a queue entry for the father to write into. The
 *    allocation may fail if the queue is full: in that case the father
 *    should block on the from child pipe read() operation.
 * RESULT
 *    BOOL, TRUE if entry allocated (buffer and buffer_size will be set),
 *          FALSE if queue full.
 * NOTES
 *    - Updates the queue structure.
 * BUGS
 * TODO
 */
static BOOL queue_slot_allocate_for_writing(queue_t *queue,
                                            addr_t *buffer,
                                            size_t *buffer_size);

/* NAME
 *    queue_send_completed
 * DESCRIPTION
 *    Tells the child that this buffer is now ready for processing (writing).
 * RESULT
 *    FALSE if the information couldn't be send to the child (fatal error).
 * NOTES
 *    - The queue entry must have been allocated with
 *         queue_slot_allocate_for_writing()
 *      and the associated buffer filled before calling this function.
 *    - Updates the queue structure.
 * BUGS
 * TODO
 */
static BOOL queue_send_completed(fd_t to_child,
                                 queue_t *queue,
                                 addr_t buffer,
                                 size_t actual);

/* NAME
 *    queue_done
 * DESCRIPTION
 *    Moves the queue done pointer ahead (one entry completed in child,
 *    in order).
 * RESULT
 *    NONE
 * NOTES
 *    - Updates the queue structure.
 * BUGS
 * TODO
 */
static void queue_done(queue_t *queue);

/* Function implementation */

BOOL buffer_do_io(fd_t input,
                  fd_t output,
                  int *the_errno,
                  BOOL *input_error) {
   BOOL result = FALSE;
   pid_t child_pid;
   fd_t comm_to_child[2];
   fd_t comm_from_child[2];
   addr_t address;
   size_t size = BUFFER_SHARED_MEMORY_SIZE;
   int status;

   *the_errno = ENOMEM; /* BUGS */

   if (PIPE_SIZE
       < (BUFFER_AMOUNT
          * sizeof(buffer_command_t))) {
      debug_log(DEBUG_LEVEL_FATAL,
                "Deadlock possible. Decrease BUFFER_AMOUNT\n");
   }
   else if (!create_shared_memory_segment(&address, size)) {
      debug_log(DEBUG_LEVEL_ERROR, "Can't create shmem segment.\n");
   }
   else if (pipe(comm_to_child)) {
      debug_perror(DEBUG_LEVEL_ERROR, "pipe()");
      free_shared_memory_segment(&address, size);
   }
   else if (pipe(comm_from_child)) {
      close(comm_to_child[0]); /* result ignored */
      close(comm_to_child[1]); /* result ignored */
      free_shared_memory_segment(&address, size);
   }
   else {
      BOOL child_died = FALSE;

      switch ((child_pid = fork())) {
	 case 0: /* child */
	    close(input); /* result ignored */
            close(comm_to_child[1]);  /* result ignored */
            close(comm_from_child[0]);  /* result ignored */

            /* Get a message, transfer that segment, send a message
             * saying that done.
             */

            child_processing(output,
                             comm_to_child[0],
                             comm_from_child[1],
                             address,
                             size);

            /* Don't need to close files, will be done by _exit() */

            _exit(0);
	    break;
	 case -1: /* error */
	    debug_perror(DEBUG_LEVEL_ERROR, "fork()");
            close(comm_to_child[0]); /* result ignored */
            close(comm_to_child[1]); /* result ignored */
            close(comm_from_child[0]); /* result ignored */
            close(comm_from_child[1]); /* result ignored */
	    break;
	 default: /* father */
            close(comm_to_child[0]);  /* result ignored */
            close(comm_from_child[1]);  /* result ignored */

            result = father_processing(input,
                                       comm_to_child[1],
                                       comm_from_child[0],
                                       address,
                                       size,
                                       the_errno,
                                       input_error,
                                       &child_died,
                                       child_pid);

            close(comm_to_child[1]);
            close(comm_from_child[0]);

            if (!child_died) {
	       if (waitpid(child_pid, &status, 0) != child_pid) {
		  debug_perror(DEBUG_LEVEL_ERROR, "waitpid()");
		  result = FALSE;
	       }
            }
            break;
      }

      free_shared_memory_segment(&address, size);
   }

   return result;
}

static void child_processing(fd_t output,
                             fd_t from_father,
                             fd_t to_father,
                             addr_t address,
                             size_t size) {
   buffer_command_t command;

   do {
      BOOL error = FALSE;

      if (misc_all_read(from_father,
                        (unsigned char *) &command,
                        sizeof(command))
          != sizeof(command)) {
         break;
      }

#ifdef BUFFER_DEBUG_CHILD
      debug_log(DEBUG_LEVEL_INFO,
                "CHILD: got addr 0x%x size %d.\n", 
                command.address,
                command.size);
#endif /* BUFFER_DEBUG_CHILD */

      if (command.size) {
         /* BUGS
          *    - well, with tapes and files we should not
          *      be so permissive.
          * TODO
          *    - Modify the interface specification.
          */
         error = (misc_all_write(output,
                                 command.address,
                                 command.size) != command.size);
      }

      /* send back as confirmation */

      if (error || (misc_all_write(to_father,
                                  (unsigned char *) &command,
                                  sizeof(command))
                    != sizeof(command))) {
         command.size = 0; /* quit, error */
      }
   } while (command.size);
} 

static BOOL father_processing(fd_t input,
                              fd_t to_child,
                              fd_t from_child,
                              addr_t address,
                              size_t size,
                              int *the_errno,
                              BOOL *input_error,
                              BOOL *child_died,
                              pid_t child_pid) {
   BOOL finishing = FALSE;
   BOOL finished = FALSE;
   BOOL had_output_error = FALSE;
   BOOL had_input_error = FALSE;
   BOOL next_must_be_eof = FALSE;
   queue_t queue;
   int status;
#ifdef BUFFER_ENABLE_STATISTICS
   unsigned int block_counter = 0;
   unsigned int blocks_in_queue = 0;
   unsigned int max_in_queue = 0;
   unsigned int times_wait = 0;
   unsigned int times_empty = 0;
#endif /* BUFFER_ENABLE_STATISTICS */

   *child_died = FALSE;

   queue_initialize(&queue, address, size, BUFFER_CHUNK_SIZE);

   while (!finished && !had_output_error & !had_input_error && !*child_died) {
      addr_t buffer;
      size_t buffer_size;
      BOOL wait_for_event = FALSE;

      /* NOTES
       *    - Could only deadlock if one of the pipes gets full, so
       *      more than PIPE_SIZE / MAX(sizeof(buffer_command_t),
       *                           sizeof(buffer_answer_t))
       *      buffers, which is why we check it in buffer_do_io().
       *    - (actual == 0) means end of stream.
       */

      if (queue_slot_allocate_for_writing(&queue, &buffer, &buffer_size)) {
         ssize_t actual = misc_all_read(input, buffer, buffer_size);

         if (actual == 0) {
            finishing = TRUE;
         }
         else if (actual > buffer_size) {
            had_input_error = TRUE;
         }
         else if (actual < buffer_size) {
            if (next_must_be_eof) {
               debug_log(DEBUG_LEVEL_ERROR, "Multiple Partial-sized block.\n");
               had_input_error = TRUE;
            }
            else {
               debug_log(DEBUG_LEVEL_WARN,
                         "Partial-sized block (ignore if no other error).\n");
            }
            next_must_be_eof = TRUE;
         }
         else if (next_must_be_eof) {
            debug_log(DEBUG_LEVEL_ERROR, "EOF was expected.\n");
            had_input_error = TRUE;
         }
          
         if (!had_input_error
             && !queue_send_completed(to_child,
                                      &queue,
                                      buffer,
                                      actual)) {
            had_output_error = TRUE;
         }
#ifdef BUFFER_ENABLE_STATISTICS
         else {
	    blocks_in_queue++;
            max_in_queue = MAX(max_in_queue, blocks_in_queue);
            block_counter++;
         }
#endif /* BUFFER_ENABLE_STATISTICS */
      }
      else {
         wait_for_event = TRUE; /* the next read() will block */
      }

      if (!had_input_error && !had_output_error) {
         BOOL got_some;
         buffer_command_t command;

	 if (!wait_for_event) {
            int flags = fcntl(from_child, F_GETFL);

            if (flags != -1) {
               fcntl(from_child, F_SETFL, flags | O_NDELAY);
            }

            /* errors ignored */
	 }
#ifdef BUFFER_ENABLE_STATISTICS
         else {
            times_wait++;
#ifdef BUFFER_DISPLAY_STATISTICS_AT_FATHER_SLEEP
            debug_log(DEBUG_LEVEL_INFO,
                      "Sleeping: blk %u blk_q %u max_q %u sleep %u e %u.\n",
                      block_counter,
                      blocks_in_queue,
                      max_in_queue,
                      times_wait,
                      times_empty);
#endif /* BUFFER_DISPLAY_STATISTICS_AT_FATHER_SLEEP */
         }
#endif /* BUFFER_ENABLE_STATISTICS */

         do {
            got_some = FALSE;

	    if (misc_all_read(from_child,
			      (unsigned char *) &command,
			      sizeof(command))
		== sizeof(command)) {
	       if ((command.size == 0) && finishing) {
		  finished = TRUE;
	       }
	       else {
		  queue_done(&queue);
#ifdef BUFFER_ENABLE_STATISTICS
                  blocks_in_queue--;
                  if (blocks_in_queue == 0) {
                     times_empty ++;
                  }
#endif /* BUFFER_ENABLE_STATISTICS */
#ifdef BUFFER_DEBUG_FATHER
		  debug_log(DEBUG_LEVEL_INFO,
			    "FATHER: done addr 0x%x size %d.\n", 
			    command.address,
			    command.size);
#endif /* BUFFER_DEBUG_FATHER */
		  got_some = TRUE;
	       }
	    }
	    else if (errno != EAGAIN) {
               debug_perror(DEBUG_LEVEL_ERROR, "misc_all_read()");
	       had_output_error = TRUE;
	    }

            /* If we were waiting for an event in blocking fashion,
             * and we got it, disable the waiting and make so that
             * we can group the queries.
             */
            if (got_some && wait_for_event) {
	       int flags = fcntl(from_child, F_GETFL);

               wait_for_event = FALSE;

	       if (flags != -1) {
		  fcntl(from_child, F_SETFL, flags | O_NDELAY);
	       }

	       /* errors ignored */
            }

            if (!finished
                && waitpid(child_pid, &status, WNOHANG)) {
               *child_died = TRUE;
            }
         } while (got_some && !*child_died);

	 if (!wait_for_event) {
            int flags = fcntl(from_child, F_GETFL);

            if (flags != -1) {
               fcntl(from_child, F_SETFL, flags & ~(O_NDELAY));
            }
	 }
      }
   }

#ifdef BUFFER_DISPLAY_STATISTICS_AT_END
            debug_log(DEBUG_LEVEL_INFO,
                      "Summary: blk %u blk_q %u max_q %u sleep %u e %u.\n",
                      block_counter,
                      blocks_in_queue,
                      max_in_queue,
                      times_wait,
                      times_empty);
#endif /* BUFFER_DISPLAY_STATISTICS_AT_END */

   if (*child_died) {
      had_output_error = TRUE;
      debug_log(DEBUG_LEVEL_ERROR, "Child died unexpectedly.\n");
   }

   if (had_input_error) {
      *input_error = TRUE;
   }
   else {
      *input_error = FALSE;
   }

   return !had_output_error && !had_input_error;
}

static BOOL create_shared_memory_segment(addr_t *address, size_t size) {
   int shm_id;

   shm_id = shmget(IPC_PRIVATE, size, IPC_CREAT /* implicit */
                                      | IPC_EXCL
                                      | (SHM_R | SHM_W));
   if (shm_id == -1) {
      debug_perror(DEBUG_LEVEL_ERROR, "shmget");
   }
   else {
      *address = shmat(shm_id, NULL, 0);
      /* BUGS
       *    - Not nice next line.
       */
      if (*address == (unsigned char *) -1) {
         debug_perror(DEBUG_LEVEL_ERROR, "shmat");
      }
      else {
         return TRUE;
      }
   }

   return FALSE;
}

static void free_shared_memory_segment(addr_t *address, size_t size) {
   shmdt(*address);
   *address = NULL;
}

/* IMPLEMENTATION NOTES
 *    - We implement the queue with two pointers to elements,
 *      reader and writer and a queue element size.
 *    - The queue is empty when
 *         reader == writer
 *    - A new element can be added in queue only if (writer + 1) modulo
 *      size is not equal to reader.
 *    - An element is available for reading only if reader is not
 *      equal to writer. But this we don't really care about.
 *    - reader and writer are opposite to the actual use by buffer.
 *      size is numbers.
 */

static void queue_initialize(queue_t *queue,
                             addr_t addr,
                             size_t size,
                             size_t chunk_size) {
   queue->next_write = 0;
   queue->next_read = 0;

   queue->buffer = addr;
   queue->size = size;
   queue->chunk_size = chunk_size;
   queue->numbers = size / chunk_size;
}

static BOOL queue_slot_allocate_for_writing(queue_t *queue,
                                            addr_t *buffer,
                                            size_t *buffer_size) {
   if (((queue->next_write + 1) % queue->numbers) != queue->next_read) {
      *buffer = (queue->next_write * queue->chunk_size) + queue->buffer;
      *buffer_size = queue->chunk_size;

      /* Integrity check */
      if ((*buffer + *buffer_size) > (queue->buffer + queue->size)) {
         debug_log(DEBUG_LEVEL_FATAL,
                   "Integrity check failed.\n");
         return FALSE;
      }

      return TRUE;
   }
   else {
      return FALSE;
   }
}

static BOOL queue_send_completed(fd_t to_child,
                                 queue_t *queue,
                                 addr_t buffer,
                                 size_t actual) {
   buffer_command_t command = { buffer, actual };

   queue->next_write = (queue->next_write + 1) % queue->numbers;

#ifdef BUFFER_DEBUG_QUEUE
   debug_log(DEBUG_LEVEL_INFO,
             "FATHER: send: entry %u / %u.\n",
             queue->next_write,
             queue->next_read);
#endif /* BUFFER_DEBUG_QUEUE */

   return (misc_all_write(to_child,
                          (unsigned char *) &command,
                          sizeof(command))
           == sizeof(command));
}

static void queue_done(queue_t *queue) {
   queue->next_read = (queue->next_read + 1) % queue->numbers;

#ifdef BUFFER_DEBUG_QUEUE
   debug_log(DEBUG_LEVEL_INFO,
             "FATHER: compl: entry %u / %u.\n",
             queue->next_write,
             queue->next_read);
#endif /* BUFFER_DEBUG_QUEUE */
}