/* Project filed -- Intelligent file cache and migration system
 * chg.c -- High level changer support module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the chg module. The chg module offers a high level
 *    medium changer interface.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: chg.c,v 1.6 2000/07/08 09:39:10 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "global.h"
#include "../debug/debug_interface.h"
#include "../scsi/scsi_interface.h"
#include "chg_interface.h"

/* Private type definitions */

#undef  DEBUG_ELEMENT_STATUS
#undef  DEBUG_SCSI_COMMANDS

#define SCSI_U24_TO_UINT(x, y, z) \
 ((((unsigned long int) (x)) << 16) \
  + (((unsigned long int) (y)) << 8) \
  + ((unsigned long int) (z)))

#define SCSI_U16_TO_UINT(x, y) SCSI_U24_TO_UINT(0, (x), (y))

#define CHG_INITIALIZE_ELEMENT_STATUS_CMD_LEN 6
#define CHG_READ_ELEMENT_STATUS_CMD_LEN 12
#define CHG_MOVE_MEDIUM_CMD_LEN 12

#define CHG_INITIALIZE_ELEMENT_STATUS 7
#define CHG_READ_ELEMENT_STATUS 0xB8
#define CHG_MOVE_MEDIUM 0xA5

#define CHG_READ_ELEMENT_STATUS_BUFFER_SIZE 500 /* @@@ */

#define CHG_TIMEOUT_SMALL 20
#define CHG_TIMEOUT_BIG (5 * 60)

#define MAX_UNIT_ATTENTION_RETRIES 3

#define SCSI_SENSE_UNIT_ATTENTION 6

/* Private function definitions */

/* NAME
 *    chg_scsi_execute
 * DESCRIPTION
 *    Wrapper for scsi_execute().
 * RESULT
 *    BOOL TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
static BOOL chg_scsi_execute(chg_context_t context,
			     const char *description,
			     unsigned char *scsi_cmd,
			     unsigned char scsi_cmd_length,
			     unsigned char *input_buffer,
			     unsigned int *input_buffer_length,
			     unsigned char *output_buffer,
			     unsigned int *output_buffer_length,
                             scsi_timeout_t timeout);

/* NAME
 *    process_elt_descriptor
 * DESCRIPTION
 *    Processes the current READ ELEMENT STATUS descriptor,
 *    copying the relevant data to the appropriate chg_elt_t
 *    structure.
 * RESULT
 *    TRUE if descriptor valid, or if *size and *elt where
 *    NULL (no processing).
 * NOTES
 * BUGS
 * TODO
 */
static BOOL process_elt_descriptor(unsigned char type,
				   unsigned int elt_len,
				   chg_elt_t **elt,
				   int *size,
				   unsigned char *current);

/* Function implementation */

chg_context_t chg_initialize(const char *changer_name) {
   if (!strncmp("scsi:", changer_name, 5)) {
      return (chg_context_t) scsi_initialize(changer_name + 5);
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "chg_initialize(): unsupported transport.\n");
      return NULL;
   }
}

void chg_deinitialize(chg_context_t *context) {
   scsi_deinitialize((scsi_context_t *) context);
}


BOOL chg_initialize_element_status(chg_context_t context,
                                   chg_elt_t *elements,
                                   int size,
                                   unsigned int *actual_size) {
   unsigned char cmd[CHG_INITIALIZE_ELEMENT_STATUS_CMD_LEN]
      = { CHG_INITIALIZE_ELEMENT_STATUS,
          0,
          0, 0, 0, 0
        };

   if (size && (elements == NULL)) {
      debug_log(DEBUG_LEVEL_WARN,
                "NULL elements, but non 0 size. Can't proceed.\n");
      return FALSE;
   }

   *actual_size = 0;

   /* We assume we have a big enough buffer, and we only compute the
    * not enough from the user passed buffer. No tricky length
    * computation.
    */
   if (chg_scsi_execute(context,
                        "Initialize Element Status",
                        cmd,
                        sizeof(cmd),
                        NULL,
                        NULL,
                        NULL,
                        NULL,
                        CHG_TIMEOUT_BIG)) {
      unsigned char buffer[CHG_READ_ELEMENT_STATUS_BUFFER_SIZE];
      unsigned char cmd[CHG_READ_ELEMENT_STATUS_CMD_LEN]
	 = { CHG_READ_ELEMENT_STATUS,
	     0,
             0, 0,
             0xFF, 0xFF,
             0,
             sizeof(buffer) >> 16,
             sizeof(buffer) >> 8,
             sizeof(buffer) & 0xFF,
             0,
             0
           };
      unsigned int actual_length = sizeof(buffer);

      if (chg_scsi_execute(context,
                           "Read Element Status",
                           cmd,
                           sizeof(cmd),
                           buffer,
                           &actual_length,
                           NULL,
                           NULL,
                           CHG_TIMEOUT_SMALL)
         && (actual_length >= 8)) {
         unsigned char *current = buffer;
         chg_elt_number_t first;
         chg_elt_number_t count;
         unsigned int byte_size;

         first = SCSI_U16_TO_UINT(current[0], current[1]);
         count = SCSI_U16_TO_UINT(current[2], current[3]);
         byte_size = SCSI_U24_TO_UINT(current[5], current[6], current[7]);

#ifdef DEBUG_ELEMENT_STATUS
         debug_log(DEBUG_LEVEL_INFO,
                   "first %d count %d byte_size %d.\n",
                   first,
                   count,
                   byte_size);
#endif /* DEBUG_ELEMENT_STATUS */

         actual_length -= 8;
         actual_length = MIN(actual_length, byte_size);

         current = &current[8];

         while (actual_length >= 8) {
            unsigned char type = current[0];
            unsigned int elt_len = SCSI_U16_TO_UINT(current[2], current[3]);
            unsigned int block_length = SCSI_U24_TO_UINT(current[5],
                                                    current[6],
                                                    current[7]);

#ifdef DEBUG_ELEMENT_STATUS
            debug_log(DEBUG_LEVEL_INFO,
                      "type %d elt_len %d block_length %d.\n",
                      type,
                      elt_len,
                      block_length);
#endif /* DEBUG_ELEMENT_STATUS */

            actual_length -= 8;
            current = &current[8];

            while ((actual_length >= elt_len) && (block_length >= elt_len)) {
               (*actual_size)++;
               if (!process_elt_descriptor(type,
					   elt_len,
					   &elements,
					   &size,
					   current)) {
                  debug_log(DEBUG_LEVEL_WARN, "elt descriptor incorrect.\n");
                  actual_length = 0;
                  break;
               }

               current += elt_len;
               block_length -= elt_len;
               actual_length -= elt_len;
            }

            if (block_length) {
               debug_log(DEBUG_LEVEL_WARN,
                         "End not on block boundary (%d).\n", 
                         block_length);
            }
         }
         
         return (actual_length == 0);
      }
      else {
         return FALSE;
      }
   }
   else {
      return FALSE;
   }
}

BOOL chg_enable_operator_access(chg_context_t context) {
   return TRUE;
}

BOOL chg_disable_operator_access(chg_context_t context) {
   return TRUE;
}

chg_error_t chg_get_error(chg_context_t context) {
   scsi_error_t scsi_error = scsi_get_error((scsi_context_t) context);
   chg_error_t chg_error;

   /* Very simple implementation for now */

   if (scsi_error.sense != 0) {
      chg_error = CHG_ERROR_OPERATOR_INTERVENTION_REQUIRED;
   }

   return chg_error;
}

const char *chg_error_description(chg_error_t error) {
   switch (error) {
      case CHG_ERROR_OPERATOR_INTERVENTION_REQUIRED:
         return "Operator intervention required";
         break;
      default:
         return "Unknown chg error";
         break;
   }
   /* NOT REACHED */
}

const char *chg_type_description(chg_elt_type_t type) {
   switch (type) {
      case CHG_ELT_TYPE_MEDIUM_TRANSPORT:
         return "MEDIUM TRANSPORT";
         break;
      case CHG_ELT_TYPE_STORAGE:
         return "STORAGE";
         break;
      case CHG_ELT_TYPE_IMPORT_EXPORT:
         return "IMPORT/EXPORT";
         break;
      case CHG_ELT_TYPE_DATA_TRANSFER:
         return "DATA TRANSFER";
         break;
      default:
 	 return "UNKNOWN";
         break;
   }
}

BOOL chg_move_medium(chg_context_t context,
                     chg_elt_number_t arm,
                     chg_elt_number_t source,
                     chg_elt_number_t destination) {
   unsigned char cmd[CHG_MOVE_MEDIUM_CMD_LEN]
      = { CHG_MOVE_MEDIUM,
          0,
          arm >> 8,
          arm,
          source >> 8,
          source, 
          destination >> 8,
          destination,
          0, 0, 0, 0
        };


   return chg_scsi_execute(context,
                           "Move Medium",
                           cmd,
                           sizeof(cmd),
                           NULL, NULL,
                           NULL, NULL,
                           CHG_TIMEOUT_BIG);
}

/* Private function implementation */

static BOOL chg_scsi_execute(chg_context_t context,
			     const char *description,
			     unsigned char *scsi_cmd,
			     unsigned char scsi_cmd_length,
			     unsigned char *input_buffer,
			     unsigned int *input_buffer_length,
			     unsigned char *output_buffer,
			     unsigned int *output_buffer_length,
                             scsi_timeout_t timeout) {
   int retries = MAX_UNIT_ATTENTION_RETRIES;

#ifdef DEBUG_SCSI_COMMANDS
   debug_log(DEBUG_LEVEL_INFO, "chg_scsi: %s.\n", description);
#endif /* DEBUG_SCSI_COMMANDS */

   do {
      if (scsi_execute((scsi_context_t) context,
		       scsi_cmd,
		       scsi_cmd_length,
		       input_buffer,
		       input_buffer_length,
		       output_buffer,
		       output_buffer_length,
		       timeout)) {
	 return TRUE;
      }
      else {
	 scsi_error_t scsi_error = scsi_get_error((scsi_context_t) context);

	 debug_log(DEBUG_LEVEL_WARN,
		   "%s: SCSI Error 0x%x 0x%x 0x%x: %s.\n",
		   description,
		   scsi_error.sense,
		   scsi_error.asc,
		   scsi_error.ascq,
		   scsi_error_description(scsi_error));
         if (scsi_error.sense != SCSI_SENSE_UNIT_ATTENTION) {
            return FALSE;
         }
      }
   } while (retries--);

   return FALSE;            
}

static BOOL process_elt_descriptor(unsigned char type,
				   unsigned int elt_len,
				   chg_elt_t **elt,
				   int *size,
				   unsigned char *current) {
   BOOL result = FALSE;

   /* This is so easy because we have chosen our chg_elt_type_t
    * appropriately.
    */

#ifdef DEBUG_ELEMENT_STATUS
   debug_log(DEBUG_LEVEL_INFO,
             "ped: type 0x%x len %d size %d [0x%x 0x%x] 0x%x.\n",
             type,
             elt_len,
             *size,
             elt,
             *elt,
             current);
#endif /* DEBUG_ELEMENT_STATUS */

   if (*size) {
      (*elt)->type = type;
      (*elt)->number = SCSI_U16_TO_UINT(current[0], current[1]);

      switch (type) {
	 case CHG_ELT_TYPE_MEDIUM_TRANSPORT:
	    result = TRUE;
	    break;
	 case CHG_ELT_TYPE_STORAGE:
	    (*elt)->elt.storage.full = current[2] & 1;
	    result = TRUE;
	    break;
	 case CHG_ELT_TYPE_IMPORT_EXPORT:
	    (*elt)->elt.import_export.full = current[2] & 1;
	    result = TRUE;
	    break;
	 case CHG_ELT_TYPE_DATA_TRANSFER:
	    (*elt)->elt.data_transfer.full = current[2] & 1;
            /* BUGS
             *    - For now.
             */
	    (*elt)->elt.data_transfer.from = CHG_ELT_NUMBER_INVALID;
	    result = TRUE;
	    break;
      }

      if (result) {
	 (*size)--;
	 (*elt)++;
      }
   }
   else {
      result = TRUE;
   }

   return result;
}
