/* Project filed -- Intelligent file cache and migration system
 * chg_interface.h -- Interface definitions for the chg module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    chg module. The chg modules offers simple medium changer
 *    services.
 * NOTES
 *    - The first version will be very simple.
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: chg_interface.h,v 1.5 2000/07/08 09:39:10 schaefer Exp $
 */

#ifndef _CHG_INTERFACE_H
#define _CHG_INTERFACE_H

/* Type definitions */

#define CHG_ELT_NUMBER_INVALID 65535
typedef unsigned int chg_elt_number_t;

typedef enum {
   CHG_ELT_TYPE_MEDIUM_TRANSPORT = 1,
   CHG_ELT_TYPE_STORAGE = 2,          /* slot */
   CHG_ELT_TYPE_IMPORT_EXPORT = 3,
   CHG_ELT_TYPE_DATA_TRANSFER = 4,    /* drive */
} chg_elt_type_t;

typedef struct {
   int unused;   
} chg_elt_medium_transport_t;

typedef struct {
   BOOL full;
} chg_elt_storage_t;

typedef struct {
   BOOL full;
} chg_elt_import_export_t;

typedef struct {
   BOOL full;

   chg_elt_number_t from; /* The slot that medium came from. This is
                           * either set to CHG_ELT_NUMBER_INVALID, or
                           * to the information returned by the changer.
                           */
} chg_elt_data_transfer_t;

typedef struct {
   chg_elt_type_t type;
   chg_elt_number_t number;

   void *user_data; /* for user to store data, NULLified by
                     * chg_initialize_element_status()
                     */

   union {
      chg_elt_medium_transport_t medium_transport;
      chg_elt_storage_t storage;
      chg_elt_import_export_t import_export;
      chg_elt_data_transfer_t data_transfer;
   } elt;
} chg_elt_t;

typedef enum {
   CHG_ERROR_RETRIABLE,
   CHG_ERROR_OPERATOR_INTERVENTION_REQUIRED,
   CHG_ERROR_SOURCE_ELEMENT_EMPTY,
   CHG_ERROR_DESTINATION_ELEMENT_FULL,
   CHG_ERROR_INITIALIZE_ELEMENT_STATUS_REQUIRED /* tapes also might have
                                                 * moved.
                                                 */
} chg_error_t;

typedef void * chg_context_t;

/* Function definitions */

/* NAME
 *    chg_initialize
 * DESCRIPTION
 *    Initializes a chg module instance. Passed is the changer_name,
 *    which describes a changer device name.
 * RESULT
 *    chg_context_t, NULL if initialization failed.
 * NOTES
 *    - At this time the only supported changer_name is:
 *         "scsi:/dev/sgX"
 *      where X is the number of the SCSI generic interface
 *      on Linux.
 * BUGS
 * TODO
 */
chg_context_t chg_initialize(const char *changer_name);

/* NAME
 *    chg_deinitialize
 * DESCRIPTION
 *    Deinitializes a chg module instance.
 * RESULT
 *    NONE
 * NOTES
 *    - context will be set to NULL
 * BUGS
 * TODO
 */
void chg_deinitialize(chg_context_t *context);

/* NAME
 *    chg_initialize_element_status
 * DESCRIPTION
 *    Re-reads the element status from the device. Will fill the
 *    table with upto size elements. It will return the size
 *    which would have been necessary in actual_size.
 * RESULT
 *    BOOL TRUE if ok
 * NOTES
 *    - Will NULLify the user_data pointer for every returned
 *      element.
 *    - elements can be specified as NULL if size is 0, e.g. to
 *      know what to allocate.
 *    - It must be verified that the element numbers, within a type,
 *      are ordered (as assumed in changer.c).
 * BUGS
 * TODO
 */
BOOL chg_initialize_element_status(chg_context_t context,
                                   chg_elt_t *elements,
                                   int size,
                                   unsigned int *actual_size);

/* NAME
 *    chg_enable_operator_access
 * DESCRIPTION
 *    Allows the operator to access the device, for example
 *    changing the state of the slots.
 * RESULT
 *    BOOL TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
BOOL chg_enable_operator_access(chg_context_t context);

/* NAME
 *    chg_disable_operator_access
 * DESCRIPTION
 *    Prevents the operator from accessing the device, for example
 *    from changing the state of the slots.
 * RESULT
 *    BOOL TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
BOOL chg_disable_operator_access(chg_context_t context);

/* NAME
 *    chg_get_error
 * DESCRIPTION
 *    Returns an error code for the given instance. This should be
 *    called when a function returns FALSE.
 * RESULT
 *    chg_error_t
 * NOTES
 * BUGS
 * TODO
 */
chg_error_t chg_get_error(chg_context_t context);

/* NAME
 *    chg_error_description
 * DESCRIPTION
 *    Returns an error description for the given chg error.
 * RESULT
 *    const char *
 * NOTES
 * BUGS
 * TODO
 */
const char *chg_error_description(chg_error_t error);

/* NAME
 *    chg_type_description
 * DESCRIPTION
 *    Returns an type description for the given element type.
 * RESULT
 *    const char *
 * NOTES
 * BUGS
 * TODO
 */
const char *chg_type_description(chg_elt_type_t type);

/* NAME
 *    chg_move_medium
 * DESCRIPTION
 *    Executes the specified MOVE MEDIUM operation.
 * RESULT
 *    BOOL TRUE if successful.
 * NOTES
 *    - If the source is a storage element, and the destination
 *      a data transfer element.
 * BUGS
 * TODO
 */
BOOL chg_move_medium(chg_context_t context,
                     chg_elt_number_t arm,
                     chg_elt_number_t source,
                     chg_elt_number_t destination);
                     
#endif /* !_CHG_INTERFACE_H */