/* Project filed -- Intelligent file cache and migration system
 * configuration.c -- Configuration module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This module implements a generic configurator, using text file
 *    data as input or input, and generic data types.
 * NOTES
 * BUGS
 * TODO
 *    - Double check the list stuff, possibly externalize into a list module.
 *    - Maybe instead of fprintf()s return exact error codes, and implement
 *      a strerror() function for the module user.
 *    - the functions searching for an item in the configured option list,
 *      and the correct test in _set(), and the use of it for type
 *      parsing in parse_line().
 * BASED-ON
 *    Original work
 * MODIFICATION-HISTORY
 * $Id: configuration.c,v 1.3 2000/04/20 17:08:07 schaefer Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "global.h"
#include "../misc/misc_interface.h"
#include "configuration_interface.h"
#include "../lists/lists_interface.h"

/* Private definitions */

#define STRING_BOOLEAN_TRUE "True"
#define STRING_BOOLEAN_FALSE "False"

/* Private types definitions */

typedef struct {
   list_item_header_t list_header;

   configuration_item_t item;
} configuration_data_item_t;

typedef struct {
   const configuration_description_t *descriptions;
   configuration_type_unsigned_integer_t description_amount;

   list_head_t data_items; /* list of configuration_data_item_t */
} configuration_info_t;

typedef struct {
   const configuration_item_t *item;
   configuration_data_item_t *data;
} configuration_find_private_data_t;

/* Private macros */

#define LIST_ITEM_TO_DATA_ITEM(item) \
   LIST_ITEM_TO_OBJECT((item), configuration_data_item_t)

/* Private function definitions */

/* NAME
 *    configuration_clear_actor
 * DESCRIPTION
 *    This function removes from the list and frees a single data item,
 *    including a possible string.
 * RESULT
 *    TRUE (never stops until end)
 * NOTES
 *    - private_2 must contain the configuration_info_t *.
 *    - This function calls configuration_free_actor().
 * BUGS
 * TODO
 */
static BOOL configuration_clear_actor(list_item_header_t *item,
                                      list_counter_t counter,
                                      void *private_1,
                                      void *private_2);

/* NAME
 *    configuration_free_actor
 * DESCRIPTION
 *    This function removes from the list and frees a single data item,
 *    including a possible string.
 * RESULT
 *    TRUE (never stops until end)
 * NOTES
 *    - private_2 must contain the configuration_info_t *.
 * BUGS
 * TODO
 */
static BOOL configuration_free_actor(list_item_header_t *item,
                                     list_counter_t counter,
                                     void *private_1,
                                     void *private_2);

/* NAME
 *    configuration_display_actor
 * DESCRIPTION
 *    This function prints a (key, value) pair suitable for saving in a
 *    configuration file. The file is passed in private_1. It is a FILE *
 *    so that stdio can be used.
 * RESULT
 *    FALSE if file I/O error, or unknown type encountered.
 * NOTES
 *    - private_1 must contain the FILE *.
 *    - private_2 must contain the configuration_info_t *.
 * BUGS
 * TODO
 */
static BOOL configuration_display_actor(list_item_header_t *item,
                                        list_counter_t counter,
                                        void *private_1,
                                        void *private_2);

/* NAME
 *    configuration_find_actor
 * DESCRIPTION
 *    This function attempts to find an item in the configuration list matching
 *    specifications given in private_1 (as a configuration_find_data_t *
 *    structure).
 * RESULT
 *    FALSE if matches, TRUE else.
 * NOTES
 *    - special structure pointed by private_1.
 *    - private_2 must contain the configuration_info_t *.
 * BUGS
 * TODO
 */
static BOOL configuration_find_actor(list_item_header_t *item,
                                     list_counter_t counter,
                                     void *private_1,
                                     void *private_2);

/* NAME
 *    parse_line
 * DESCRIPTION
 *    Parses the specified line (removing final \n if any), according to
 *    the defined keywords, and insert the configuration items in the list.
 * RESULT
 *    TRUE if parsing ok
 * NOTES
 * BUGS
 * TODO
 */
static BOOL parse_line(configuration_info_t *info,
                       unsigned char *buffer);

/* NAME
 *    get_description
 * DESCRIPTION
 *    Sets description pointing to the description matching the given
 *    name.
 * RESULT
 *    TRUE if a match was found
 * NOTES
 *    - The caller must then verify the type is the correct one.
 * BUGS
 * TODO
 */
static BOOL get_description(configuration_info_t *info,
                            const char *name,
                            configuration_description_t **description);

/* Function implementation */

configuration_context_t configuration_initialize
   (const configuration_description_t *descriptions,
    const configuration_type_unsigned_integer_t description_amount) {
   configuration_info_t *info
      = (configuration_info_t *) malloc(sizeof(configuration_info_t));

   if (info) {
      info->descriptions = descriptions;
      info->description_amount = description_amount;
      list_initialize(&info->data_items);
   }

   return (configuration_context_t) info;
}

void configuration_deinitialize(configuration_context_t *context) {
   configuration_info_t *info = (configuration_info_t *) *context;

   configuration_clear(info);

   free(info);

   *context = NULL;
}

BOOL configuration_load(configuration_context_t context,
                        const char *configuration_file_name) {
   configuration_info_t *info = (configuration_info_t *) context;
   BOOL result = TRUE;
   FILE *f;

   f = fopen(configuration_file_name, "r");
   if (f) {
      BOOL aborted = FALSE;

      do {
         char buffer[MAX_LINE_LENGTH];

         if (fgets(buffer, sizeof(buffer), f)) {
            aborted = !parse_line(info, buffer);
         }

      } while (!feof(f) && !ferror(f) && !aborted);

      result = !ferror(f) && !aborted;

      if (fclose(f)) {
         perror("configuration_load: fclose()");
         result = FALSE;
      }
   }
   else {
      perror("configuration_load: fopen()");
      result = FALSE;
   }

   return result;
}

BOOL configuration_save(configuration_context_t context,
                        const char *configuration_file_name) {
   configuration_info_t *info = (configuration_info_t *) context;
   BOOL result = TRUE;
   FILE *f;

   f = fopen(configuration_file_name, "w");
   if (f) {
      if (!list_iterator(&info->data_items,
                         NULL,
                         list_next,
                         configuration_display_actor,
                         (void *) f,
                         (void *) info,
                         NULL)) {
         fprintf(stderr, "configuration_save: I/O error or incorrect type.\n");
         result = FALSE;
      }

      if (fclose(f)) {
         perror("configuration_save: fclose()");
         result = FALSE;
      }
   }
   else {
      perror("configuration_save: fopen()");
      result = FALSE;
   }

   return result;
}

void configuration_clear(configuration_context_t context) {
   configuration_info_t *info = (configuration_info_t *) context;

   if (!list_iterator(&info->data_items,
                      NULL,
                      list_next,
                      configuration_clear_actor,
                      NULL,
                      (void *) info,
                      NULL)) {
      fprintf(stderr, "configuration_clear: assumption failed.\n");
   }
}

BOOL configuration_get(configuration_context_t context,
                       configuration_item_t *item) {
   configuration_info_t *info = (configuration_info_t *) context;
   configuration_find_private_data_t data = { item, NULL };

   if (!list_iterator(&info->data_items,
                      NULL,
                      list_next,
                      configuration_find_actor,
                      (void *) &data,
                      (void *) info,
                      NULL)) {
      memcpy(item, &data.data->item, sizeof(*item));

      return TRUE;
   }
   else {
      configuration_description_t *description;

      if (get_description(info, item->type.name, &description)) {
         if ((description->item.type.type == item->type.type)
             && description->default_ok) {
            memcpy(item, &description->item, sizeof(*item));
            return TRUE;
         }
      }

      return FALSE;
   }
   /* NOT REACHED */
}

BOOL configuration_set(configuration_context_t context,
                       const configuration_item_t *item) {
   configuration_info_t *info = (configuration_info_t *) context;
   configuration_description_t *description;

   if (get_description(info, item->type.name, &description)
       && /* evaluation order */ (description->item.type.type
                                  == item->type.type)) {
      configuration_find_private_data_t data = { item, NULL };
      configuration_data_item_t *data_item;

      data_item
         = (configuration_data_item_t *)
              malloc(sizeof(configuration_data_item_t));
      if (data_item) {
         /* Create new item */

         memcpy(&data_item->item, item, sizeof(data_item->item));
         data_item->item.type.name = strdup(item->type.name);
         if (data_item->item.type.type == CONFIGURATION_ITEM_TYPE_STRING) {
            data_item->item.object.string = strdup(item->object.string);
         }

         if ((data_item->item.type.name == NULL)
             || ((data_item->item.type.type == CONFIGURATION_ITEM_TYPE_STRING)
                  && (data_item->item.object.string == NULL))) {
            configuration_free_actor(&data_item->list_header,
                                     0,
                                     NULL,
                                     (void *) info);
            return FALSE;
         }

         if (!list_iterator(&info->data_items,
                            NULL,
                            list_next,
                            configuration_find_actor,
                            (void *) &data,
                            (void *) info,
                            NULL)) {
            /* Delete old item */
            configuration_clear_actor(&data.data->list_header,
                                      0,
                                      NULL,
                                      (void *) info);
         }

         /* Add new item */
         list_append_tail(&info->data_items,
                          &data_item->list_header);
         return TRUE;
      }
      else {
         return FALSE;
      }
   }
   else {
      fprintf(stderr, "configuration_set: unknown option/type %s.\n",
                      item->type.name);
      return FALSE;
   }
   /* NOT REACHED */
}

/* Private function implementation */

static BOOL configuration_free_actor(list_item_header_t *item,
                                     list_counter_t counter,
                                     void *private_1,
                                     void *private_2) {
   configuration_data_item_t *data
      = LIST_ITEM_TO_DATA_ITEM(item);

   if (data->item.type.name) {
      free(data->item.type.name);
      data->item.type.name = NULL;
   }

   if (data->item.type.type == CONFIGURATION_ITEM_TYPE_STRING) {
      if (data->item.object.string) {
         free(data->item.object.string);
         data->item.object.string = NULL;
      }
   }

   free(data);

   return TRUE;
}

static BOOL configuration_clear_actor(list_item_header_t *item,
                                      list_counter_t counter,
                                      void *private_1,
                                      void *private_2) {
   configuration_info_t *info
      = (configuration_info_t *) private_2;

   list_remove_item(&info->data_items,
                    item);

   return configuration_free_actor(item,
                                   counter,
                                   private_1,
                                   private_2);
}

static BOOL configuration_display_actor(list_item_header_t *item,
                                        list_counter_t counter,
                                        void *private_1,
                                        void *private_2) {
   configuration_data_item_t *data
      = LIST_ITEM_TO_DATA_ITEM(item);
   FILE *f = (FILE *) private_1;

   fprintf(f, "%s ", data->item.type.name);

   switch (data->item.type.type) {
      case CONFIGURATION_ITEM_TYPE_SIGNED_INTEGER:
         fprintf(f, "%d", data->item.object.signed_integer);
         break;
      case CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER:
         fprintf(f, "%u", data->item.object.unsigned_integer);
         break;
      case CONFIGURATION_ITEM_TYPE_STRING:
         fprintf(f, "%s", data->item.object.string);
         break;
      case CONFIGURATION_ITEM_TYPE_BOOLEAN:
         fprintf(f, "%s",
                 data->item.object.boolean
                    ? STRING_BOOLEAN_TRUE
                    : STRING_BOOLEAN_FALSE);
         break;
      default:
         fprintf(f, "[Unknown type error]\n");
         return FALSE; /* error */
         break;
   }

   fprintf(f, "\n");

   return !ferror(f);
}

static BOOL configuration_find_actor(list_item_header_t *item,
                                     list_counter_t counter,
                                     void *private_1,
                                     void *private_2) {
   configuration_data_item_t *data
      = LIST_ITEM_TO_DATA_ITEM(item);
   configuration_find_private_data_t *seeked
      = (configuration_find_private_data_t *) private_1;

   if ((seeked->item->type.type == data->item.type.type)
       && !strcmp(seeked->item->type.name, data->item.type.name)) {
      seeked->data = data;
      return FALSE; /* found */
   }

   return TRUE; /* not yet found */
}

static BOOL get_description(configuration_info_t *info,
                            const char *name,
                            configuration_description_t **description) {
   configuration_description_t *current
      = (configuration_description_t *) info->descriptions;
   configuration_type_unsigned_integer_t amount = info->description_amount;

   while (amount--) {
      if (!strcmp(name, current->item.type.name)) {
         *description = current;
         return TRUE;
      }
      current++;
   }
   
   return FALSE;
}

static BOOL parse_line(configuration_info_t *info,
                       unsigned char *buffer) {
   misc_delete_spaces(&buffer); /* and final \n */

   /* Only supports comments at start of line (possibly after spaces) */
   if (buffer[0] == '#') {
      buffer[0] = '\0';
   }

   if (buffer[0]) {
      char *name = strsep((char **) &buffer, " "); /* only accepts 0x20 */

      if (name) {
         misc_delete_spaces((unsigned char **) &name);
         if (name[0]) {
            char *parameter = strsep((char **) &buffer, " "); /* same */

            if (parameter) {
               misc_delete_spaces((unsigned char **) &parameter);
               if (parameter[0]) {
                  configuration_description_t *description;

                  if (get_description(info, name, &description)) {
                     configuration_item_t item;
                     int count = 0;

		     switch (description->item.type.type) {
			case CONFIGURATION_ITEM_TYPE_SIGNED_INTEGER:
			   count = sscanf(parameter,
                                          "%d",
                                          &item.object.signed_integer);
			   break;
			case CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER:
			   count = sscanf(parameter,
                                          "%u",
                                          &item.object.unsigned_integer);
			   break;
			case CONFIGURATION_ITEM_TYPE_STRING:
                           item.object.string = parameter; /* no strdup here */
                           count = 1;
			   break;
			case CONFIGURATION_ITEM_TYPE_BOOLEAN:
			   if (!strcmp(STRING_BOOLEAN_TRUE,
                                      parameter)) {
                              item.object.boolean = TRUE;
                              count = 1;
                           }
                           else if (!strcmp(STRING_BOOLEAN_FALSE,
                                            parameter)) {
                              item.object.boolean = FALSE;
                              count = 1;
                           }
			   break;
			default:
			   break;
		     }

                     if (count) {
                        item.type.name = description->item.type.name;
                        item.type.type = description->item.type.type;  

                        if (configuration_set((void *) info, &item)) {
                           return TRUE;
                        }
                     }
                  }
               }
            }           
         }
      }
   }
   else {
      return TRUE; /* ignore empty/comment lines */
   }

   return FALSE;
}
