/* Project filed -- Intelligent file cache and migration system
 * configuration_interface.h -- Interface definitions for the Configuration
 *                              module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    configuration module. The configuration modules offers facilities
 *    to handle plain text configuration files.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: configuration_interface.h,v 1.2 2000/09/20 18:24:45 schaefer Exp $
 */

#ifndef _CONFIGURATION_INTERFACE_H
#define _CONFIGURATION_INTERFACE_H

/* Type definitions */

typedef signed int configuration_type_signed_integer_t;
typedef unsigned int configuration_type_unsigned_integer_t;
typedef char * configuration_type_string_t; /* strings are malloc()ated,
                                             * to be free()d by caller.
                                             */
typedef signed char configuration_type_boolean_t;

typedef enum {
   CONFIGURATION_ITEM_TYPE_SIGNED_INTEGER,
   CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER,
   CONFIGURATION_ITEM_TYPE_STRING,
   CONFIGURATION_ITEM_TYPE_BOOLEAN
} configuration_item_type_t;

typedef struct {
   configuration_type_string_t name;
   configuration_item_type_t type;
} configuration_type_t;

typedef struct {
   configuration_type_t type;

   /* Now, depending on the item type.type, follows the data */
   union {
      configuration_type_signed_integer_t signed_integer;
      configuration_type_unsigned_integer_t unsigned_integer;
      configuration_type_string_t string;
      configuration_type_boolean_t boolean;
   } object;
} configuration_item_t;

typedef struct {
   BOOL default_ok; /* this item has a default value, given below */

   configuration_item_t item;
} configuration_description_t;

typedef void *configuration_context_t;

/* Function definitions */

/* NAME
 *    configuration_initialize
 * DESCRIPTION
 *    Allocates and initializes a configuration_context_t structure, which must
 *    be passed afterwards to all the configuration module functions. The
 *    caller passes an array of configuration_description_t structures
 *    (description_amount), containing the name and type of all the
 *    configuration parameters.
 * RESULT
 *    A configuration_context_t, acting as a context for the configuration
 *    module. If memory cannot be allocated, NULL is returned.
 * NOTES
 *    - Normal life cycle of the configuration module is:
 *         configuration_initialize()
 *            repeat:
 *               configuration_load()
 *                  repeat:
 *                     configuration_{set|get}()
 *               [configuration_save()]
 *         configuration_deinitialize()
 *    - The module owns the description array, until
 *      configuration_deinitialize(), where the caller will own it again.
 * BUGS
 * TODO
 */
configuration_context_t configuration_initialize
   (const configuration_description_t *descriptions,
    const configuration_type_unsigned_integer_t description_amount);

/* NAME
 *    configuration_deinitialize
 * DESCRIPTION
 *    Frees the allocated context.
 * RESULT
 *    NONE
 * NOTES
 *    - Side effect: the context is set to NULL.
 * BUGS
 * TODO
 */
void configuration_deinitialize(configuration_context_t *context);


/* NAME
 *    configuration_load
 * DESCRIPTION
 *    Parses the specified file name, adding/replacing the found variables to
 *    the context, with their values.
 * RESULT
 *    TRUE if the format of the configuration file is right and there was no
 *    error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL configuration_load(configuration_context_t context,
                        const char *configuration_file_name);

/* NAME
 *    configuration_save
 * DESCRIPTION
 *    Saves the current context into the specified text file.
 * RESULT
 *    TRUE if there was no error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL configuration_save(configuration_context_t context,
                        const char *configuration_file_name);

/* NAME
 *    configuration_clear
 * DESCRIPTION
 *    Clears the context.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void configuration_clear(configuration_context_t context);

/* NAME
 *    configuration_get
 * DESCRIPTION
 *    Retrieves the value associated with the named item. The value
 *    will be returned as a pointer valid until configuration_deinitialize()
 *    or any operation (such as configuration_clear(), configuration_load()).
 * RESULT
 *    FALSE if not found, TRUE else.
 * NOTES
 * BUGS
 * TODO
 */
BOOL configuration_get(configuration_context_t context,
                       configuration_item_t *item);

/* NAME
 *    configuration_set
 * DESCRIPTION
 *    Sets a configuration value. The value will be copied,
 *    including the string.
 * RESULT
 *    FALSE if any error, else TRUE.
 * NOTES
 * BUGS
 * TODO
 */
BOOL configuration_set(configuration_context_t context,
                       const configuration_item_t *item);

#endif /* !CONFIGURATION_INTERFACE_H */
