/* Project filed -- Intelligent file cache and migration system
 * filesend.c -- Filesend routines
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the filesend module. The filesend module offers simple file
 *    sending and receiving routines (including encryption).
 * NOTES
 * BUGS
 *    - Same problem as usual: big endianness, etc.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: filesend.c,v 1.5 2001/02/04 19:57:15 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "global.h"
#include "../misc/misc_interface.h"
#include "../debug/debug_interface.h"
#include "../md5/md5_interface.h"
#include "filesend_interface.h"

/* Private type definitions */

typedef struct {
   unsigned long int block_count;
   unsigned long int block_size;
} block_header_t;

/* Function implementation */

#define BUFFER_LENGTH 32768 /* @@@ */

BOOL filesend_send_file(fd_t socket,
                        void *caller_private,
                        void encrypt(void *caller_private,
                                     unsigned char *in_buffer,
                                     unsigned char *out_buffer,
                                     unsigned int buffer_length,
                                     BOOL normal),
                        BOOL get_next_file(void *caller_private,
                                           fd_t *fd,
                                           md5_hash_ptr_t last_file_md5sum,
                                           BOOL *eof /* end of stream */)) {
   BOOL result = TRUE;
   void *md5_context = md5_context_allocate();
   void *md5_sub_file_context = md5_context_allocate();

   if (md5_context && md5_sub_file_context) {
      unsigned char in_buffer[BUFFER_LENGTH];
      unsigned char out_buffer[BUFFER_LENGTH];
      fd_t f = -1;
      BOOL finished_input = FALSE;
      block_header_t block_header = { 0, 0 };

      md5_context_initialize(md5_context);
      md5_context_initialize(md5_sub_file_context); /* for cleanessness */

      while (!finished_input && result) {
	 size_t amount_read = 0;
	 unsigned char *tmp_buffer = in_buffer;

	 /* Collect data as much as we can, passing over files. */

	 while (!finished_input && (amount_read < BUFFER_LENGTH) && result) {
	    if (f == -1) {
               BOOL eof = FALSE;
	       unsigned char sub_file_hash[MD5_HASH_STRING_LENGTH];

               /* Finalize the md5sum even if it is not valid (first) */
               md5_compute_hash(md5_sub_file_context,
                                NULL,
                                0,
                                sub_file_hash,
                                TRUE);

	       if (!get_next_file(caller_private, &f, sub_file_hash, &eof)) {
                  debug_log(DEBUG_LEVEL_ERROR, "get_next_file().\n");
                  result = eof;
		  f = -1;
	       }

               md5_context_initialize(md5_sub_file_context);

               if (f == -1) {
		  finished_input = TRUE;
                  debug_log(DEBUG_LEVEL_INFO, "finished input files.\n");
               }
	    }

	    if (!finished_input) {
	       ssize_t tmp_read
		  = read(f, tmp_buffer, BUFFER_LENGTH - amount_read);

	       if (tmp_read < 0) {
		  result = FALSE;
		  debug_perror(DEBUG_LEVEL_ERROR,
                               "filesend_send_file: read()");
		  if (close(f)) {
		     debug_perror(DEBUG_LEVEL_ERROR,
                                  "filesend_send_file: close()");
		  }               
	       }
	       else if (tmp_read) {
		  md5_compute_hash(md5_sub_file_context,
				   tmp_buffer,
				   tmp_read,
				   NULL,
				   FALSE);
		  amount_read += tmp_read;
		  tmp_buffer += tmp_read;
	       }
	       else {
		  if (close(f)) {
		     result = FALSE;
		     debug_perror(DEBUG_LEVEL_ERROR,
                                  "filesend_send_file: close()");
		  }
		  f = -1;
	       }
	    }
	 }

	 if (result) {
	    /* f was closed if it was open, unless it is not finished yet. */

	    /* Compute md5 */

	    md5_compute_hash(md5_context,
			     in_buffer /* in clear */, 
			     amount_read, /* could be zero */
			     NULL,
			     FALSE);

            if (amount_read) {
	       /* Encrypt data */

	       if (encrypt) {
		  encrypt(caller_private,
			  in_buffer,
			  out_buffer,
			  amount_read,
			  TRUE);
		  tmp_buffer = out_buffer;
	       }
	       else {
		  debug_log(DEBUG_LEVEL_WARN,
                            "Data stream will not be encrypted.\n");
		  tmp_buffer = in_buffer;
	       }

	       /* Send header
                * BUGS
                *    - little endian
                */
               block_header.block_size = amount_read;

               if (block_header.block_size) {
		  block_header.block_count++;

		  if (write(socket, &block_header, sizeof(block_header))
		      == sizeof(block_header)) {
		     /* Send all encrypted data */

		     while (amount_read && result) {
			ssize_t tmp_written
			   = write(socket, tmp_buffer, amount_read);

			if (tmp_written > 0 ) {
			   amount_read -= tmp_written;
			   tmp_buffer += tmp_written;
			}
			else {
			   result = FALSE;
			   debug_perror(DEBUG_LEVEL_ERROR,
					"filesend_send_file: write()");
			}
		     }
		  }
		  else {
		     result = FALSE;
		     debug_perror(DEBUG_LEVEL_ERROR,
				  "filesend_send_file: write() header");
		  }
               }

	       /* BUGS
		*    - in case of error should tell other side, no ?
		*/
	    }
         }
      }

      if (f != -1) {
	 if (close(f)) {
	    result = FALSE;
	    debug_perror(DEBUG_LEVEL_ERROR, "filesend_send_file: close()");
	 }

	 f = -1;
      }

      if (result) {
         unsigned char hash[MD5_HASH_STRING_LENGTH];
         unsigned char out[MD5_HASH_STRING_LENGTH];
         unsigned char *buf = out;
	 
         /* Send special last block with md5 encrypted with key 2 */
	 
         md5_compute_hash(md5_context, NULL, 0, hash, TRUE);

         if (encrypt) {
             encrypt(caller_private, hash, out, MD5_HASH_STRING_LENGTH, FALSE);
         }
         else {
             buf = hash;
         }

         block_header.block_count++;
         block_header.block_size = 0; /* special */

         if (write(socket, &block_header, sizeof(block_header))
             == sizeof(block_header)) {
	    if (write(socket, buf, MD5_HASH_STRING_LENGTH)
		!= MD5_HASH_STRING_LENGTH) {
	       result = FALSE;
	       debug_perror(DEBUG_LEVEL_ERROR,
			    "filesend_send_file: write()");
	    }
         }
         else {
            result = FALSE;
            debug_perror(DEBUG_LEVEL_ERROR,
                         "filesend_send_file: write() md5 header");
         }
      }

      md5_context_free(&md5_context);
      md5_context_free(&md5_sub_file_context);
   }
   else {
      if (md5_context) { /* order counts in if */
         md5_context_free(&md5_context);
      }

      debug_log(DEBUG_LEVEL_ERROR, "Can't allocate md5 context.\n");
      result = FALSE;
   }

   return result;
}

BOOL filesend_receive_file(fd_t socket,
                           void *caller_private,
                           void decrypt(void *caller_private,
                                        unsigned char *in_buffer,
                                        unsigned char *out_buffer,
                                        unsigned int buffer_length,
                                        BOOL normal),
                           BOOL get_next_file(void *caller_private,
                                              fd_t *fd,
                                              unsigned int *file_max,
                                              md5_hash_ptr_t last_file_md5sum,
                                              BOOL last),
                           md5_hash_ptr_t md5sum) {
   BOOL result = TRUE;
   void *md5_context = md5_context_allocate();
   void *md5_sub_file_context = md5_context_allocate();

   if (md5_context && md5_sub_file_context) {
      unsigned long int block_count = 0;
      fd_t f = -1;
      BOOL finished_input = FALSE;
      unsigned char in_buffer[BUFFER_LENGTH];
      unsigned char out_buffer[BUFFER_LENGTH];
      unsigned int file_max = 0;
      unsigned int actual_stored_in_this_file = 0;
      unsigned char read_hash[MD5_HASH_STRING_LENGTH];

      md5_context_initialize(md5_context);

      while (!finished_input && result) {
         block_header_t header;

         /* Receive the header */

         if (misc_all_read(socket, (unsigned char *) &header, sizeof(header))
             == sizeof(header)) {
            block_count++;
            if (block_count == header.block_count) {
               if (header.block_size > BUFFER_LENGTH) {
                  debug_log(DEBUG_LEVEL_ERROR, "Block size too big.\n");
                  result = FALSE;
               }
               else if (header.block_size) {
                  if (misc_all_read(socket, in_buffer, header.block_size)) {
		     char *tmp_buffer;

		     /* Normal */

		     if (decrypt) {
			decrypt(caller_private,
				in_buffer,
				out_buffer,
				header.block_size,
				TRUE);
			tmp_buffer = out_buffer;
		     }
		     else {
			debug_log(DEBUG_LEVEL_WARN,
				  "Data stream will not be encrypted.\n");
			tmp_buffer = in_buffer;
		     }

		     /* Compute md5 */

		     md5_compute_hash(md5_context,
				      tmp_buffer /* in clear */, 
				      header.block_size, /* could be zero */
				      NULL,
				      FALSE);

		     /* Send data to disk */

		     while (result && header.block_size) {
			if (f == -1) {
			   unsigned char sub_file_hash[MD5_HASH_STRING_LENGTH];

			   /* Finalize the md5sum even if it is not valid
                            * (for first pass).
                            */
			   md5_compute_hash(md5_sub_file_context,
					    NULL,
					    0,
					    sub_file_hash,
					    TRUE);

                           actual_stored_in_this_file = 0;
			   if (!get_next_file(caller_private,
                                              &f,
                                              &file_max,
                                              sub_file_hash,
                                              FALSE)) {
			      debug_log(DEBUG_LEVEL_ERROR,
                                        "get_next_file().\n");
			      result = FALSE;
			      f = -1;
			   }

                           md5_context_initialize(md5_sub_file_context);

			   if (f == -1) {
			      result = FALSE;
			      debug_log(DEBUG_LEVEL_ERROR,
					"Not enough output files.\n");
			   }
			}

			if (result) {
			   size_t write_amount
                              = MIN(file_max, header.block_size);

			   if (write_amount) {
			      if (write(f, tmp_buffer, write_amount)
				  != write_amount) {
				 result = FALSE;
				 debug_perror(DEBUG_LEVEL_ERROR, "write()");
			      }
			      else {
				 md5_compute_hash(md5_sub_file_context,
						  tmp_buffer /* in clear */, 
						  write_amount,
						  NULL,
						  FALSE);
                                 actual_stored_in_this_file += write_amount;
				 file_max -= write_amount;
				 header.block_size -= write_amount;
				 tmp_buffer += write_amount;
			      }
			   }
			   else if (header.block_size) {
			      if (close(f)) {
				 debug_perror(DEBUG_LEVEL_ERROR, "close()");
				 result = FALSE;
			      }

                              f = -1;
			   }
			}
		     }
                  }
                  else {
		     debug_perror(DEBUG_LEVEL_ERROR, "misc_all_read() 2");
		     result = FALSE;
                  }
               }
               else {
                  debug_log(DEBUG_LEVEL_INFO, "Last block reached.\n");
                  if (misc_all_read(socket, read_hash, MD5_HASH_STRING_LENGTH)
                      == MD5_HASH_STRING_LENGTH) {
                     finished_input = TRUE;
                  }
                  else {
                     debug_perror(DEBUG_LEVEL_ERROR, "md5 read");
                     result = FALSE;
                  }
               }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR, "Non matching block number.\n");
               result = FALSE;
            }
         }
         else {
            debug_perror(DEBUG_LEVEL_ERROR, "misc_all_read()");
            result = FALSE;
         }
      }

      if (f != -1) {
         if (close(f)) {
            debug_perror(DEBUG_LEVEL_ERROR, "close()");
            result = FALSE;
         }
      }

      if (result) {
         unsigned char sub_file_hash[MD5_HASH_STRING_LENGTH];
	 
         /* Finalize the md5sum even if it is not valid (first) */
         md5_compute_hash(md5_sub_file_context,
                          NULL,
                          0,
                          sub_file_hash,
                          TRUE);

         /* Tell the callback the md5sum for the last subfile */

         if ((f != -1)
             && !get_next_file(caller_private,
			       &f,
			       &actual_stored_in_this_file,
			       sub_file_hash,
			       TRUE)) {
            debug_log(DEBUG_LEVEL_ERROR,
                      "get_next_file(finalize).\n");
            result = FALSE;
         }
      }

      f = -1;

      if (result) {
         unsigned char in_hash[MD5_HASH_STRING_LENGTH];
         unsigned char *buf = in_hash;

         /* Compute hash from read data */
         md5_compute_hash(md5_context, NULL, 0, md5sum, TRUE);

         /* Decrypt transmitted hash with key 2 */

         if (decrypt) {
             decrypt(caller_private,
                     read_hash,
                     in_hash,
                     MD5_HASH_STRING_LENGTH,
                     FALSE);
         }
         else {
             buf = read_hash;
         }         

         /* Verify md5sum
          * BUGS
          *    - won't be right if no data exchanged
          */

         result = !memcmp(md5sum, buf, MD5_HASH_STRING_LENGTH);
         if (!result) {
            debug_log(DEBUG_LEVEL_ERROR, "md5sum doesn't match.\n");
         }
      }

      md5_context_free(&md5_context);
      md5_context_free(&md5_sub_file_context);
   }
   else {
      if (md5_context) { /* order counts in if */
         md5_context_free(&md5_context);
      }

      debug_log(DEBUG_LEVEL_ERROR, "Can't allocate md5 context.\n");
      result = FALSE;
   }

   return result;
}
