/* Project filed -- Intelligent file cache and migration system
 * filesend_interface.h -- Interface definitions for the filesend module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the filesend module. The filesend module offers simple file
 *    sending and receiving routines (including encryption).
 * NOTES
 *    - The idea is to buffer internally a bit of a file, and then
 *      send it with a unencrypted block sequence, this block size,
 *      then the data. At end sends a md5sum. Allowing encryption too,
 *      a different encryption for md5sum, etc. Data must be confirmed
 *      at the end. Reports the error easily for the caller. Hides
 *      all the socket I/O complexity.
 * BUGS
 *    - Still a bit confusing is the relationship to the hello block size ...
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: filesend_interface.h,v 1.4 2001/02/04 19:57:15 schaefer Exp $
 */

#ifndef _FILESEND_INTERFACE_H
#define _FILESEND_INTERFACE_H
 
/* Function definitions */

/* NAME
 *    filesend_send_file
 * DESCRIPTION
 *    This will send the file file to the socket socket, including
 *    packetization, encryption of the data (if encrypt is not
 *    set to NULL), and automatic md5summing of the unencrypted file,
 *    which is sent encrypted (if encrypt is not NULL). The caller can
 *    even provide two different versions of the encryption (one for
 *    data, with normal == TRUE, and one for the md5sum).
 * RESULT
 *    TRUE if data could be successfully sent, FALSE else.
 * NOTES
 *    - The input file is generated with a call from the get_next_file()
 *      supplied callback (which sets to -1 when end of files reached,
 *      and FALSE in case of error), called at each end of file.
 *      This allows the caller to handle many small files at once.
 *      We will close it after use. Do not consider *fd at the call.
 *      The md5sum passed to that function is the one from the last subfile
 *      (== part) not the whole file (?!), undefined at the first
 *      call.
 *    - As the get_next_file() function tells when the transaction is over,
 *      we don't need to call it another time.
 *    - This function is supposedly used in a client/server application,
 *      using filesend_receive_file() as peer.
 * BUGS
 *    - If an error happens and get_next_file() didn't yet return end,
 *      internal get_next_file() will be broken.
 * TODO
 */
BOOL filesend_send_file(fd_t socket,
                        void *caller_private,
                        void encrypt(void *caller_private,
                                     unsigned char *in_buffer,
                                     unsigned char *out_buffer,
                                     unsigned int buffer_length,
                                     BOOL normal),
                        BOOL get_next_file(void *caller_private,
                                           fd_t *fd,
                                           md5_hash_ptr_t last_file_md5sum,
                                           BOOL *eof));

/* NAME
 *    filesend_receive_file
 * DESCRIPTION
 *    This will receive the file file from the socket socket, including
 *    depacketization, decryption of the data (if decrypt is not
 *    set to NULL), and automatic md5summing of the undecrypted file,
 *    which is sent decrypted (if decrypt is not NULL). The caller can
 *    even provide two different versions of the decryption (one for
 *    data, with normal == TRUE, and one for the md5sum). The
 *    processed md5sum will be verified and returned into the
 *    user's specified buffer md5sum, which must be at least
 *    MD5_HASH_STRING_LENGTH bytes long.
 * RESULT
 *    TRUE if data could be successfully received with correct
 *    md5sum, FALSE else.
 * NOTES
 *    - The output file is generated with a call from the get_next_file()
 *      supplied callback (which sets to -1 when end of files reached,
 *      and FALSE in case of error),
 *      called at each end of file. This allows the caller to handle
 *      many small files at once (e.g. if the OS doesn't support
 *      files larger than file_max, or we use the filed's paradigm.
 *      We will close it after use. Do not consider *fd at the call.
 *      The md5sum passed to that function is the one from the last file,
 *      not the whole file (?!), undefined at the first call.
 *    - At the end of the correct input stream/file, the get_next_file()
 *      will be called again, but with BOOL finished as TRUE.
 *    - This function is supposedly used in a client/server application,
 *      using filesend_receive_file() as peer.
 * BUGS
 * TODO
 */
BOOL filesend_receive_file(fd_t socket,
                           void *caller_private,
                           void decrypt(void *caller_private,
                                        unsigned char *in_buffer,
                                        unsigned char *out_buffer,
                                        unsigned int buffer_length,
                                        BOOL normal),
                           BOOL get_next_file(void *caller_private,
                                              fd_t *fd,
                                              unsigned int *file_max,
                                              md5_hash_ptr_t last_file_md5sum,
                                              const BOOL last),
                           md5_hash_ptr_t md5sum);

#endif /* !_FILESEND_INTERFACE_H */
