/* Project filed -- Intelligent file cache and migration system
 * lists.c -- Generic doubly-linked list manipulation routines.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This module implements handling functions for doubly-linked
 *    generic lists.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: lists.c,v 1.3 2001/01/24 18:57:02 schaefer Exp $
 */

#undef  DEBUG_LISTS
#ifdef DEBUG_LISTS
#include <stdio.h>
#endif /* DEBUG_LISTS */
#include <stdlib.h>
#include "global.h"
#include "lists_interface.h"

/* Private function definitions */

/* NAME
 *    list_item_count_actor
 * DESCRIPTION
 *    Actors which counts the number of items (the counter
 *    is pointed by private_1, of type list_counter_t).
 * RESULT
 *    TRUE
 * NOTES
 * BUGS
 * TODO
 */
static BOOL list_item_count_actor(list_item_header_t *item,
                                  list_counter_t counter,
                                  void *private_1,
                                  void *private_2);

#ifdef DEBUG_LISTS
/* NAME
 *    dump_list
 * DESCRIPTION
 *    Dumps the list links.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void dump_list(list_head_t *list);
#endif /* DEBUG_LISTS */

/* Function implementation */

void list_initialize(list_head_t *list) {
   list->first = NULL;
   list->last = NULL;
}

void list_append_after(list_head_t *list,
                       list_item_header_t *item,
                       list_item_header_t *new) {
   new->prev = item;
   new->next = item->next;
   item->next = new;

   if (new->next == NULL) {
      list->last = new;
   }
   else {
      new->next->prev = new;
   }
}

void list_insert_before(list_head_t *list,
			list_item_header_t *item,
			list_item_header_t *new) {
   new->next = item;
   new->prev = item->prev;
   item->prev = new;

   if (new->prev == NULL) {
      list->first = new;
   }
   else {
      new->prev->next = new;
   }
}

void list_insert_with_priority(list_head_t *list,
                               list_item_header_t *new,
                               void *private,
                               BOOL insert_now(list_head_t *list,
                                               list_item_header_t *new,
                                               list_item_header_t *current,
                                               void *private)) {
   list_item_header_t *current = NULL;

   while ((current = list_next(list, current))) {
      if (insert_now(list, new, current, private)) {
         list_insert_before(list, current, new);
         return;
      }
   }

   list_append_tail(list, new);

}

void list_append_before(list_head_t *list,
                        list_item_header_t *item,
                        list_item_header_t *new) {
   new->next = item;
   new->prev = item->prev; /* might become NULL in the second case */
   item->prev = new;

   if (new->prev == NULL) {
      list->first = new;
   }
   else {
      new->prev->next = new;
   }
}

void list_append_head(list_head_t *list,
                      list_item_header_t *new) {
   new->next = list->first;
   new->prev = NULL;
   if (list->first) {
      list->first->prev = new;
   }
   list->first = new;
   if (list->last == NULL) {
      list->last = new;
   }

#ifdef DEBUG_LISTS
   dump_list(list);
#endif /* DEBUG_LISTS */
}

void list_append_tail(list_head_t *list,
                      list_item_header_t *new) {
   new->next = NULL;
   new->prev = list->last;
   if (list->last) {
      list->last->next = new;
   }
   list->last = new;
   if (list->first == NULL) {
      list->first = new;
   }

#ifdef DEBUG_LISTS
   dump_list(list);
#endif /* DEBUG_LISTS */
}

void list_remove_item(list_head_t *list,
                      list_item_header_t *item) {
   if (item->next) {
      item->next->prev = item->prev;
   }
   else {
      list->last = item->prev;
   }

   if (item->prev) {
      item->prev->next = item->next;
   }
   else {
      list->first = item->next;
   }

   item->next = NULL;
   item->prev = NULL;
}

list_item_header_t *list_remove_head(list_head_t *list) {
   list_item_header_t *removed;

   removed = list->first;

   if (removed) {
      list->first = removed->next;

      if (removed == list->last) {
         list->last = NULL;
      }
      if (list->first) {
         list->first->prev = NULL;
      }
   }

   return removed;
}

list_item_header_t *list_remove_tail(list_head_t *list) {
   list_item_header_t *removed;

   removed = list->last;

   if (removed) {
      list->last = removed->prev;

      if (removed == list->first) {
         list->first = NULL;
      }
      if (list->last) {
         list->last->next = NULL;
      }
   }

   return removed;
}

BOOL list_empty(const list_head_t *list) {
   return (list->first == NULL);
}

void list_cat_tail(list_head_t *dest, list_head_t *to_add) {
   if (to_add->first) { /* implies ->last, too */
      if (dest->first) {
         to_add->first->prev = dest->last;
         dest->last->next = to_add->first;
         dest->last = to_add->last;
      }
      else {
        dest->first = to_add->first;
        dest->last = to_add->last;
      }

      to_add->first = NULL;
      to_add->last = NULL;
   }
   /* else NOOP */
}

void list_split(list_head_t *orig,
                list_head_t *new,
                list_item_header_t *item) {
   list_item_header_t *temp;

   temp = orig->last;
   orig->last = item->prev;

   if (item->prev) {
      item->prev->next = NULL;
   }

   new->first = item;
   new->last = temp;
   item->prev = NULL;

   /* item now belongs to new list, all items before are in
    * orig.
    */
}

list_item_header_t *list_next(const list_head_t *list,
                              const list_item_header_t *item) {
   if (item) {
      return item->next;
   }
   else {
      return list->first;
   }
}

list_item_header_t *list_prev(const list_head_t *list,
                              const list_item_header_t *item) {
   if (item) {
      return item->prev;
   }
   else {
      return list->last;
   }
}

BOOL list_iterator(list_head_t *list,
                   list_item_header_t *initial_item,
                   list_item_header_t *list_item_incr
                                          (const list_head_t *list,
                                           const list_item_header_t *item),
                   BOOL actor(list_item_header_t *item,
                              list_counter_t counter,
                              void *private_1,
                              void *private_2),
                   void *private_1,
                   void *private_2,
                   list_item_header_t **last_processed) {
   list_item_header_t *current;
   BOOL result = TRUE;
   list_counter_t counter = 0;

   current = list_item_incr(list, initial_item);

   while (result && current) {
      list_item_header_t *next = list_item_incr(list, current);

      result = actor(current,
                     counter,
                     private_1,
                     private_2);
      counter++;
      if (last_processed) {
	 *last_processed = current;
      }
      current = next;
   }

   return result;
}

list_counter_t list_size(const list_head_t *list) {
   list_counter_t size = 0;

   if (list_iterator((list_head_t *) list,
                     NULL,
                     list_next,
                     list_item_count_actor,
                     (void *) &size,
                     NULL,
                     NULL)) {
      return size;
   }
   else {
      return 0;
   }
}

/* Private function implementation */

static BOOL list_item_count_actor(list_item_header_t *item,
                                  list_counter_t counter,
                                  void *private_1,
                                  void *private_2) {
   list_counter_t *size = (list_counter_t *) private_1;

   (*size)++;

   return TRUE;
}

#ifdef DEBUG_LISTS
static void dump_list(list_head_t *list) {
   list_item_header_t *current = NULL;

   fprintf(stderr, "LIST 0x%x dump START\n", (unsigned long int) list);

   while ((current = list_next(list, current))) {
      fprintf(stderr, "   0x%lx\n", (unsigned long int) current);
   }

   fprintf(stderr, "LIST 0x%x dump END\n", (unsigned long int) list);
}
#endif /* DEBUG_LISTS */