/* Project filed -- Intelligent file cache and migration system
 * lists_interface.h -- Generic doubly-linked list manipulation routines
 *                      interface definitions.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file defines the data types and functions used to handle
 *    doubly-linked generic lists.
 * NOTES
 * BUGS
 * TODO
 *    - We could do with a find() generic function.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: lists_interface.h,v 1.2 2001/01/24 18:57:02 schaefer Exp $
 */

#ifndef _LISTS_INTERFACE_H
#define _LISTS_INTERFACE_H

/* Type definitions */

typedef unsigned long int list_counter_t;

typedef struct list_item_header {
   struct list_item_header *next;
   struct list_item_header *prev;
} list_item_header_t;

typedef struct {
   list_item_header_t *first;
   list_item_header_t *last;
} list_head_t;

/* Macros */

/* NAME
 *    LIST_ITEM_TO_OBJECT
 * DESCRIPTION
 *    Converts the list_item_header_t * to a pointer to the specified
 *    user-type t, if required, does an offset conversion so that the
 *    resulting value is at the right offset.
 * RESULT
 *    t *
 * NOTES
 *    - This is used to retrieve the englobbing structure from the item
 *      header. If the structure has the header as the first item,
 *      this will simply do a no-cost type conversion.
 *    - Assumes the list_item_header_t struct component is called list_header.
 * BUGS
 * TODO
 */
#define LIST_ITEM_TO_OBJECT(item, t) \
   ((t *) \
      (((unsigned char *) (item)) \
       - offsetof(t, \
                  list_header)))

/* Function definitions */

/* NAME
 *    list_initialize
 * DESCRIPTION
 *    Initializes a doubly-linked list structure (list_head_t *) so that
 *    it can be used to insert/remove/iterate.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void list_initialize(list_head_t *list);

/* NAME
 *   list_append_after
 * DESCRIPTION
 *    Appends a list item after another.
 * RESULT
 *    NONE
 * NOTES
 *    - Of course item must be valid, and new shall not already
 *      be in the list, and new shall not be NULL.
 * BUGS
 * TODO
 */
void list_append_after(list_head_t *list,
                       list_item_header_t *item,
                       list_item_header_t *new);

/* NAME
 *   list_insert_before
 * DESCRIPTION
 *    Inserts a list item before another.
 * RESULT
 *    NONE
 * NOTES
 *    - Of course item must be valid, and new shall not already
 *      be in the list, and new shall not be NULL.
 * BUGS
 * TODO
 */
void list_insert_before(list_head_t *list,
			list_item_header_t *item,
			list_item_header_t *new);

/* NAME
 *    list_insert_with_priority
 * DESCRIPTION
 *    Inserts into a list BEFORE the item pointed to by current when
 *    insert_now() returns TRUE, or at the end of list.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void list_insert_with_priority(list_head_t *list,
                               list_item_header_t *new,
                               void *private,
                               BOOL insert_now(list_head_t *list,
                                               list_item_header_t *new,
                                               list_item_header_t *current,
                                               void *private));

/* NAME
 *   list_append_before
 * DESCRIPTION
 *    Appends a list item before another.
 * RESULT
 *    NONE
 * NOTES
 *    - Of course item must be valid, and new shall not already
 *      be in the list, and new shall not be NULL.
 * BUGS
 * TODO
 */
void list_append_before(list_head_t *list,
			list_item_header_t *item,
			list_item_header_t *new);

/* NAME
 *    list_append_head
 * DESCRIPTION
 *    Appends a list item at the head of the specified list.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void list_append_head(list_head_t *list,
                      list_item_header_t *new);

/* NAME
 *    list_append_tail
 * DESCRIPTION
 *    Appends a list item at the tail of the specified list.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void list_append_tail(list_head_t *list,
                      list_item_header_t *new);

/* NAME
 *    list_remove_item
 * DESCRIPTION
 *    Removes the specified item from the list.
 * RESULT
 *    NONE
 * NOTES
 *    - item must belong to the list, else bad things might happen.
 * BUGS
 * TODO
 */
void list_remove_item(list_head_t *list,
                      list_item_header_t *item);

/* NAME
 *    list_remove_head
 * DESCRIPTION
 *    Removes and returns the first list item, if any.
 * RESULT
 *    list_item_header_t *, maybe NULL.
 * NOTES
 * BUGS
 * TODO
 */
list_item_header_t *list_remove_head(list_head_t *list);

/* NAME
 *    list_remove_tail
 * DESCRIPTION
 *    Removes and returns the last list item, if any.
 * RESULT
 *    list_item_header_t *, maybe NULL.
 * NOTES
 * BUGS
 * TODO
 */
list_item_header_t *list_remove_tail(list_head_t *list);

/* NAME
 *    list_empty
 * DESCRIPTION
 *    Checks if the list is empty.
 * RESULT
 *    BOOL TRUE if empty.
 * NOTES
 * BUGS
 * TODO
 */
BOOL list_empty(const list_head_t *list);

/* NAME
 *    list_cat_tail
 * DESCRIPTION
 *    Appends the list to_add at the end of the list dest.
 *    At the end of the operation, the to_add list is empty.
 * RESULT
 *    NONE
 * NOTES
 *    - Also supports the following degenerated cases:
 *         - to_add was empty
 *         - dest was empty
 *         - or both empty.
 * BUGS
 * TODO
 */
void list_cat_tail(list_head_t *dest, list_head_t *to_add);

/* NAME
 *    list_split
 * DESCRIPTION
 *    From a single list orig, creates two new lists orig and new,
 *    where new starts at the item item, and orig is the old list
 *    without item and any following items.
 * RESULT
 *    NONE
 * NOTES
 *    - item must be a valid item from that list.
 * BUGS
 * TODO
 */
void list_split(list_head_t *orig,
                list_head_t *new,
                list_item_header_t *item);

/* NAME
 *    list_next
 * DESCRIPTION
 *    Returns the next item from the specified list. If the passed
 *    item is NULL, returns the first item of the list.
 * RESULT
 *    list_item_header_t *, NULL if no more items.
 * NOTES
 * BUGS
 * TODO
 */
list_item_header_t *list_next(const list_head_t *list,
                              const list_item_header_t *item);

/* NAME
 *    list_prev
 * DESCRIPTION
 *    Returns the previous item from the specified list. If
 *    the passed item is NULL, returns the last item of the list.
 * RESULT
 *    list_item_header_t *, NULL if no more items.
 * NOTES
 * BUGS
 * TODO
 */
list_item_header_t *list_prev(const list_head_t *list,
                              const list_item_header_t *item);

/* NAME
 *    list_iterator
 * DESCRIPTION
 *    Iterates over the specified list, using the specified
 *    incremental function (e.g. list_prev, or list_next), which
 *    is passed initial_item at first call. Every valid item is passed
 *    to the actor function. Processing is interrupted and FALSE
 *    is returned if the actor function returns FALSE. Else,
 *    we proceed until the incremental function returns NULL, and
 *    we return TRUE (also if no item in the list).
 * RESULT
 *    BOOL
 * NOTES
 *    - initial_item will normally be NULL, unless you want to
 *      continue an interrupted iteration, for example, in that
 *      case you should pass list_item_incr(last_processed), unless
 *      it is NULL.
 *    - if last_processed is NULL, nothing will be stored in it.
 *    - The incremental function shall not have side effects.
 *    - The actor function will not change the structure of the list.
 *      At most, the actor function can remove the current item
 *      from the list.
 * BUGS
 * TODO
 */
BOOL list_iterator(list_head_t *list,
                   list_item_header_t *initial_item,
                   list_item_header_t *list_item_incr
                                          (const list_head_t *list,
                                           const list_item_header_t *item),
                   BOOL actor(list_item_header_t *item,
                              list_counter_t counter,
                              void *private_1,
                              void *private_2),
                   void *private_1,
                   void *private_2,
                   list_item_header_t **last_processed);

/* NAME
 *    list_size
 * DESCRIPTION
 *    Returns the size of the specified list.
 * RESULT
 *    list_counter_t
 * NOTES
 * BUGS
 * TODO
 */
list_counter_t list_size(const list_head_t *list);

#endif /* !_LISTS_INTERFACE_H */
