/* Project filed -- Intelligent file cache and migration system
 * md5_interface.h -- Interface definitions for the md5 module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    md5 module. The md5 modules encapsulates the public domain
 *    MD5sum routines.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: md5_interface.h,v 1.3 2001/01/28 20:40:30 schaefer Exp $
 */

#ifndef _MD5_INTERFACE_H
#define _MD5_INTERFACE_H

/* Type definitions */

#define MD5_HASH_STRING_LENGTH 16
typedef unsigned char *md5_hash_ptr_t;

#define MD5_HASH_TEXTUAL_STRING_LENGTH (32 + 1)
typedef char *md5_hash_textual_ptr_t;

/* Function definitions */

/* NAME
 *    md5_context_allocate
 * DESCRIPTION
 *    Allocates a MD5 context structure, to be used in later calls
 *    of md5_context_initialize(), a few md5_compute_hash() and
 *    then md5_context_free().
 * RESULT
 *    void *, NULL if allocation failed.
 * NOTES
 *    - Once allocated, the context must be either md5_context_free()d
 *      immediately, or passed first to md5_context_initialize(), and
 *      then a number of times to md5_compute_hash(), then to
 *      either md5_context_initialize() and a few md5_compute_hash(),
 *      or md5_context_free().
 * BUGS
 * TODO
 */
void *md5_context_allocate(void);

/* NAME
 *    md5_context_initialize
 * DESCRIPTION
 *    Initializes a MD5 context structure so that we can start over
 *    again with a new md5_compute_hash() cycle.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void md5_context_initialize(void *context);

/* NAME
 *    md5_context_free
 * DESCRIPTION
 *    Frees a MD5 context structure previously allocated with
 *    md5_context_allocate(). Will set context to NULL.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void md5_context_free(void **context);

/* NAME
 *    md5_compute_hash
 * DESCRIPTION
 *    Computes a partial, or continues computation of, a MD5 hash
 *    on a buffer.
 * RESULT
 *    NONE
 * NOTES
 *    - You must allocate, initialize and free the context yourself.
 *    - last call must be done with final set to TRUE.
 *    - hash should only be considered after last invocation (final == TRUE).
 *    - hash must be a buffer of size MD5_HASH_STRING_LENGTH, user allocated.
 *    - If buffer_length is zero, buffer can be NULL and the md5sum
 *      will be unchanged.
 * BUGS
 * TODO
 */
void md5_compute_hash(void *context,
                      const unsigned char *buffer,
                      const int buffer_length,
                      md5_hash_ptr_t hash,
                      BOOL final);

/* NAME
 *    md5_hash_to_textual_md5sum
 * DESCRIPTION
 *    Converts a binary hash into a md5sum-like textual representation.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void md5_hash_to_textual_md5sum(const md5_hash_ptr_t hash,
                                md5_hash_textual_ptr_t text);

/* NAME
 *    md5_textual_md5sum_to_hash
 * DESCRIPTION
 *    Converts a textual hash into a binary hash.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void md5_textual_md5sum_to_hash(const md5_hash_textual_ptr_t text,
                                md5_hash_ptr_t hash);

#endif /* !_MD5_INTERFACE_H */
