/* Project filed -- Intelligent file cache and migration system
 * process.c -- Process and IPC services.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the process module.  The process modules offers simple process
 *    services.
 * NOTES
 *    - To communicate between the father and the childs, we have the
 *      following pipes:
 *         - TO each child is a different pipe, so the child only
 *           needs to block reading on those, e.g.
 *         - all the children write to a shared fd (only one file on
 *           the father). Because of write() atomicity, there is
 *           no risk of data intermix unless the written size is bigger
 *           than PIPE_SIZE I think@@. A header is to be used specifying
 *           the message length. Maybe they can even use writev() thus
 *           being atomic (I think@@) without needing a temp copy buffer.
 *           Thus if a child dies, the father should't die because of it
 *           nor hang (but can writev() fail with a SIGSEGV for one of
 *           the buffers, if yes, use a temp buffer and write()).
 *         - the father reads the header first, then the whole data.
 *         - an additional pipe is used for child death notification,
 *           write() is ok from signal, and it won't deadlock unless there
 *           are more than PIPE_SIZE/size(notif), which should be enforced
 *           at process creation time for safety.
 *    - Children close all the file descriptors they don't use, they
 *      also free some memory.
 * BUGS
 * TODO
 *    - test
 *    - simplify: ie one low-level I/O functions.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: process.c,v 1.6 2001/01/24 18:58:12 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <signal.h>
#include <sys/wait.h>
#include <unistd.h>
#include <string.h>
#include "global.h"
#include "../debug/debug_interface.h"
#include "process_interface.h"
#include "../lists/lists_interface.h"

/* For debugging which isn't errors */
#if 1
#define DEBUG_PROCESS(format, args...) 
#else
#define DEBUG_PROCESS(format, args...) fprintf(stderr, format, ## args); 
#endif

/* Type definitions */

typedef struct {
   list_head_t children; /* list of process_child_t */

   fd_t input_fd; /* read from children */
   fd_t input_fd_write; /* write from children */
   fd_t die_fd;   /* read from2 the SIGCHLD signal handler */
   fd_t die_fd_write;   /* write from the SIGCHLD signal handler */
} process_info_t;

/* The child as seen from the father */
typedef struct {
   list_item_header_t list_header;

   pid_t child;
   fd_t output_fd;
   fd_t output_fd_write;

   process_info_t *info;

   void *user_private;
} process_child_t;

typedef unsigned short message_size_t;

typedef struct {
   pid_t identifier;

   message_size_t size;
   /* message follows */
} message_header_t;

typedef struct {
   pid_t pid;
   int status;
} child_dying_message_t;

/* Macros */

#define PROCESS_MAX_CHILDREN (4096 /* @@@ PIPE_SIZE @@@ */ / sizeof(child_dying_message_t))

#define LIST_ITEM_TO_CHILD(item) \
   LIST_ITEM_TO_OBJECT((item), process_child_t)

/* Private function definitions */

/* NAME
 *    child_signal_handler
 * DESCRIPTION
 *    Handles a dying child.
 * RESULT
 * NOTES
 * BUGS
 *    - does that work
 *    - can we risk missing a signal ?
 * TODO
 */
static void child_signal_handler(int value);

/* NAME
 *    pid_to_child
 * DESCRIPTION
 *    Converts a pid to a child.
 * RESULT
 *    process_child_t * or NULL if not found.
 * NOTES
 * BUGS
 * TODO
 */
static process_child_t *pid_to_child(process_info_t *info, pid_t pid);

/* Global variables */

static fd_t die_fd_write;

/* Function implementation */

process_context_t process_initialize(void) {
   process_info_t *info
      = (process_info_t *) malloc(sizeof(process_info_t));

   if (signal(SIGCHLD, child_signal_handler) != SIG_ERR) {
      if (info) {
	 fd_t filedes[2];

	 list_initialize(&info->children);

	 if (!pipe(filedes)) {
	    info->input_fd = filedes[0];
	    info->input_fd_write = filedes[1];

	    if (!pipe(filedes)) {
	       info->die_fd = filedes[0];
	       info->die_fd_write = filedes[1];
               /* The next line is the reason why there can only be
                * one instance of the process module.
                */
               die_fd_write = info->die_fd_write;
	    }
	    else {
	       close(info->input_fd); /* res. ign. */
	       close(info->input_fd_write); /* res. ign. */
	       free(info);
	       info = NULL;
	    }
	 }
	 else {
	    free(info);
	    info = NULL;
	 }
      }
   }
   else {
      perror("process_initialize: signal()");
   }

   return (process_context_t) info;
}

void process_deinitialize(process_context_t *context) {
   process_info_t *info = (process_info_t *) *context;
   /* BUGS
    *    - what if still some children around ?
    */

   close(info->input_fd); /* res. ign. */
   close(info->input_fd_write); /* res. ign. */
   close(info->die_fd); /* res. ign. */
   close(info->die_fd_write); /* res. ign. */

   free(info);
   *context = NULL;
}

child_context_t process_launch(process_context_t context,
                               int child_process
                                      (child_context_t child_context),
                               void *caller_private) {
   process_info_t *info = (process_info_t *) context;
   process_child_t *child = NULL;

   if (list_size(&info->children) < PROCESS_MAX_CHILDREN) {
      child = (process_child_t *) malloc(sizeof(process_child_t));

      if (child) {
	 fd_t filedes[2];

	 if (!pipe(filedes)) {
	    list_item_header_t *temp;

	    child->output_fd = filedes[0];
	    child->output_fd_write = filedes[1];
	    child->info = info;
	    child->user_private = caller_private;

	    switch ((child->child = fork())) {
	       case 0: /* child */
                  DEBUG_PROCESS("fd is: %d.\n", child->output_fd);
                  child->child = getpid();
		  /* Free everything we don't need */
		  close(child->output_fd_write); /* res. ign. */
		  /* We could use list_iterator() instead */
		  while ((temp = list_remove_head(&info->children))) {
                     process_child_t *temp_child;

                     temp_child = LIST_ITEM_TO_CHILD(temp);
		     close(temp_child->output_fd_write); /* res. ign. */
		     free(temp_child);
		  }
		  close(child->info->input_fd); /* res. ign. */
		  close(child->info->die_fd); /* res. ign. */
		  close(child->info->die_fd_write); /* res. ign. */
                  DEBUG_PROCESS("fd is: %d.\n", child->output_fd);
		  _exit(child_process((child_context_t) child));
		  break;
	       case -1: /* error */
		  perror("process_launch: fork()");
		  close(child->output_fd); /* res. ign. */
		  close(child->output_fd_write); /* res. ign. */
		  free(child);
		  child = NULL;
		  break;
	       default: /* father */
		  close(child->output_fd); /* res. ign. */
		  list_append_tail(&info->children, &(child->list_header));
		  break;
	    }
	 }
	 else {
	    perror("process_launch: pipe()");
	    free(child);
	    child = NULL;
	 }
      }
   }
   else {
      debug_log(DEBUG_LEVEL_FATAL, "process: Too many processes.\n");
   }

   if (child) {
      DEBUG_PROCESS("new child ID 0x%x.\n", (int) child);
   }
   return (child_context_t) child;
}

void *process_get_private_of(child_context_t child_context) {
   return ((process_child_t *) child_context)->user_private;
}

void process_terminate(process_context_t context,
                       child_context_t child_context) {
   debug_log(DEBUG_LEVEL_FATAL, "process_terminate: not implemented.\n");
}

void process_select_father(process_context_t context,
                           fd_set *readfds,
                           fd_t *max_fd) {
   process_info_t *info = (process_info_t *) context;

   *max_fd = MAX(*max_fd, info->input_fd);
   *max_fd = MAX(*max_fd, info->die_fd);
   FD_SET(info->input_fd, readfds);
   FD_SET(info->die_fd, readfds);
}

BOOL process_next_ready_father(process_context_t context,
                               child_context_t *child_context,
                               void **caller_private,
                               BOOL *running,
                               int *message_size,
                               fd_set *readfds,
                               fd_t max_fd) {
   process_info_t *info = (process_info_t *) context;
   BOOL result = FALSE;

   if (FD_ISSET(info->die_fd, readfds)) {
      child_dying_message_t dead;

      DEBUG_PROCESS("DIE MESSAGE READY.\n");

      if (read(info->die_fd, &dead, sizeof(dead)) != sizeof(dead)) {
         perror("process_next_ready_father: die read()");
      }
      else {
         process_child_t *child = pid_to_child(info, dead.pid);

         if (child) {
	    *running = FALSE;
	    *child_context = (child_context_t) child;
            *caller_private = child->user_private;

            list_remove_item(&info->children, &child->list_header);
            close(child->output_fd); /* res. ign. */
            close(child->output_fd_write); /* res. ign */
            free(child);
	    result = TRUE;
         }
         else {
            DEBUG_PROCESS("death notification for unknown child %d.\n",
                          dead.pid);
         }
      }

      FD_CLR(info->die_fd, readfds);
   }
   else if (FD_ISSET(info->input_fd, readfds)) {
      message_header_t message;

      DEBUG_PROCESS("NORMAL MESSAGE READY.\n");

      if (read(info->input_fd, &message, sizeof(message)) != sizeof(message)) {
         perror("process_next_ready_father: pipe read()");
      }
      else {
         process_child_t *child = pid_to_child(info, message.identifier);

         DEBUG_PROCESS("message: identifier %d size %d ev. child 0x%x\n",
                       message.identifier,
                       message.size,
                       (int) child);

         if (child) {
	    *running = TRUE;
	    *child_context = (child_context_t) child;
            *caller_private = child->user_private;
            *message_size = message.size;
	    result = TRUE;
         }
         else {
            DEBUG_PROCESS("message from unknown child %d.\n",
                          message.identifier);
         }
      }

      FD_CLR(info->input_fd, readfds);
   }

   return result;
}

void process_select_child(child_context_t child_context,
                          fd_set *readfds,
                          fd_t *max_fd) {
   process_child_t *child = (process_child_t *) child_context;

   *max_fd = MAX(*max_fd, child->output_fd_write);
   FD_SET(child->output_fd, readfds);
}

BOOL process_child_has_message(child_context_t child_context,
                               fd_set *readfds,
                               fd_t *max_fd,
                               int *message_length) {
   process_child_t *child = (process_child_t *) child_context;

   if (FD_ISSET(child->output_fd, readfds)) {
      message_header_t message;

      if (read(child->output_fd,
               &message,
               sizeof(message)) != sizeof(message)) {
         perror("process_child_has_message: pipe read()");
      }
      else {
         *message_length = message.size;
         return TRUE;
      }
   }

   return FALSE;
}

BOOL process_father_send_message(child_context_t child_context,
                                 const unsigned char *message,
                                 const int message_length) {
   process_child_t *child = (process_child_t *) child_context;

   DEBUG_PROCESS("father sends message to child 0x%x length %d.\n",
                 (int) child,
                 message_length);

   if (message_length <= PROCESS_MAX_MESSAGE_LENGTH) {
      unsigned char buffer[sizeof(message_header_t)
                           + PROCESS_MAX_MESSAGE_LENGTH];
      message_header_t *message_header = (message_header_t *) &buffer[0];

      message_header->identifier = 0; /* father */
      message_header->size = message_length;

      memcpy(buffer + sizeof(message_header_t),
	     message,
	     message_length);

      if (write(child->output_fd_write,
                buffer,
                message_length + sizeof(message_header_t))
          == (message_length + sizeof(message_header_t))) {
         return TRUE;
      }
      else {
         perror("process_father_send_message: send_message write()");
      }
   }

   return FALSE;
}

BOOL process_child_send_message(child_context_t child_context,
                                const unsigned char *message,
                                const int message_length) {
   process_child_t *child = (process_child_t *) child_context;
   process_info_t *info = (process_info_t *) child->info;

   DEBUG_PROCESS("in process_child_send_message()\n");

   DEBUG_PROCESS("child 0x%x sends message to father length %d (of which"
                 " header length %d).\n",
                 (int) child,
                 message_length,
                 sizeof(message_header_t));

   if (message_length <= PROCESS_MAX_MESSAGE_LENGTH) {
      unsigned char buffer[sizeof(message_header_t)
                           + PROCESS_MAX_MESSAGE_LENGTH];
      message_header_t *message_header = (message_header_t *) &buffer[0];

      message_header->identifier = child->child;
      message_header->size = message_length;

      DEBUG_PROCESS("message: identifier %d size %d evaluated child %x\n",
                    message_header->identifier,
                    message_header->size,
                    (int) child);

      memcpy(buffer + sizeof(message_header_t),
	     message,
	     message_length);

      /* Assumed atomic since shared */
      if (write(info->input_fd_write,
                buffer,
                message_length + sizeof(message_header_t))
           == (message_length + sizeof(message_header_t))) {
         return TRUE;
      }
      else {
         perror("process_child_send_message: send_message write()");
      }
   }

   return FALSE;
}

BOOL process_child_receive_message_blocking(child_context_t child_context,
					    unsigned char *message,
					    int *message_length,
                                            const BOOL any_length) {
   message_header_t header;
   process_child_t *child = (process_child_t *) child_context;

   if (read(child->output_fd, &header, sizeof(header)) == sizeof(header)) {
      if (header.identifier == 0) {
         if ((header.size == *message_length)
             || (any_length && (header.size < *message_length))) {
            *message_length = header.size;

	    DEBUG_PROCESS("child 0x%x receives msg from father len %d.\n",
			  (int) child,
			  *message_length);

	    *message_length = read(child->output_fd, message, *message_length);
	    if (*message_length != header.size) {
	       perror("process_child_receive_message: message read()");
	       DEBUG_PROCESS("fd was: %d.\n", child->output_fd);
	    }

	    return ((*message_length == header.size)
		    && (header.size <= PROCESS_MAX_MESSAGE_LENGTH));
	 }
	 else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "process_child_receive_message: incorrect msg size.\n");
	 }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "process_child_receive_message: incorrect identifier.\n");
      }
   }
   else {
      perror("process_child_receive_message: message header read()");
   }

   return FALSE;
}

BOOL process_child_receive_message_special(child_context_t child_context,
					   unsigned char *message,
					   int *message_length) {
   message_header_t header;
   process_child_t *child = (process_child_t *) child_context;

   DEBUG_PROCESS("child 0x%x receives NB msg from father len %d.\n",
                 (int) child,
                 *message_length);
   
   *message_length = read(child->output_fd, message, *message_length);
   if (*message_length != header.size) {
      perror("process_child_receive_message_special: message read()");
      DEBUG_PROCESS("fd was: %d.\n", child->output_fd);
   }
   
   return ((*message_length == header.size)
           && (header.size <= PROCESS_MAX_MESSAGE_LENGTH));
}

BOOL process_father_receive_message(child_context_t child_context,
                                    unsigned char *message,
                                    int *message_length) {
   process_child_t *child = (process_child_t *) child_context;
   process_info_t *info = child->info;
   int old_len = *message_length;

   DEBUG_PROCESS("father receives message from child 0x%x length %d.\n",
                 (int) child,
                 *message_length);

   *message_length = read(info->input_fd, message, *message_length);
   if (*message_length != old_len) {
      perror("process_father_receive_message: message read()");
   }

   return ((*message_length == old_len)
           && (old_len <= PROCESS_MAX_MESSAGE_LENGTH));
}

/* Private function implementation */

static void child_signal_handler(int value) {
   child_dying_message_t message;

   message.pid = wait(&message.status);

   /* No risk of deadlock because of PROCESS_MAX_CHILDREN */
   write(die_fd_write, &message, sizeof(message));

   /* We do not do much, child termination will be detected by
    * the closing of the fd.
    * BUGS
    *    - does that work
    *    - can we risk missing a signal ?  loop for all processes testing?
    */
#if 1 /* not allowed in signal handler, only enable for debugging if needed */
   debug_log(DEBUG_LEVEL_INFO,
             "child %d exited status %d.\n",
             message.pid,
             message.status);
#endif
}

static process_child_t *pid_to_child(process_info_t *info, pid_t pid) {
   list_item_header_t *current = NULL;

   /* Could use iterator, but way easier that way. */

   while ((current = list_next(&info->children, current))) {
      process_child_t *child = LIST_ITEM_TO_CHILD(current);

      if (child->child == pid) {
         return child;
      }
   }

   return NULL;
}
