/* Project filed -- Intelligent file cache and migration system
 * process_interface.h -- Interface definitions for the process module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    process module. The process modules offers simple process
 *    services.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: process_interface.h,v 1.4 2001/01/02 10:34:42 schaefer Exp $
 */

#ifndef _PROCESS_INTERFACE_H
#define _PROCESS_INTERFACE_H

/* Type definitions */

#define PROCESS_MAX_MESSAGE_LENGTH 2048 /* Must be smaller than PIPE_SIZE
                                         * plus process module's overhead.
                                         */

typedef void * child_context_t; /* child / father */

typedef void * process_context_t; /* module */

/* Function definitions */

/* NAME
 *    process_initialize
 * DESCRIPTION
 *    Sets up initialization for the process module. Returns a process
 *    module context.
 * RESULT
 *    process_context_t, NULL if initialization failed.
 * NOTES
 *    - The process module lifecycle is:
 *         process_initialize()
 *            ... any of the process control functions (repeated), except
 *                process_initialize().
 *         process_deinitialize()
 *    - Currently, only one context can be active at any time.
 * BUGS
 * TODO
 */
process_context_t process_initialize(void);

/* NAME
 *    process_deinitialize
 * DESCRIPTION
 *    Deinitializes the process module.
 * RESULT
 *    NONE
 * NOTES
 *    - *context is set to NULL.
 * BUGS
 * TODO
 */
void process_deinitialize(process_context_t *context);

/* NAME
 *    process_launch
 * DESCRIPTION
 *    Launches a new process, with its communication channel. Returns
 *    the child context to be used to communicate with/terminate the
 *    child.
 * RESULT
 *    child_context_t, NULL if child couldn't be created/initialized.
 * NOTES
 *    - void * caller_private is caller private, and will be accessible
 *      through process_get_private_of(child_context);
 *    - If must be accessed by the parent, must of course be allocated
 *      dynamically. If only used by the child, then the child must
 *      still have it in its scope (no stack allocation and exit of
 *      function!).
 * BUGS
 * TODO
 */
child_context_t process_launch(process_context_t context,
                               int child_process
                                      (child_context_t child_context),
                               void *caller_private);

/* NAME
 *    process_get_private_of
 * DESCRIPTION
 *    Returns the caller private data associated with the named child.
 * RESULT
 *    void *
 * NOTES
 * BUGS
 * TODO
 */
void *process_get_private_of(child_context_t child_context);

/* NAME
 *    process_terminate
 * DESCRIPTION
 *    Attempts to forcibly terminate a child process. The function
 *    may return before actual termination. 
 * RESULT
 *    NONE
 * NOTES
 *    - For normal termination, it's probably better to use the
 *      messaging interface with a user-defined protocol.
 * BUGS
 * TODO
 */
void process_terminate(process_context_t context,
                       child_context_t child_context);

/* NAME
 *    process_select_father
 * DESCRIPTION
 *    Adds the file descriptors for the children communication channels
 *    to the readfds array, and sets max_fd to be the higher fd.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void process_select_father(process_context_t context,
                           fd_set *readfds,
                           fd_t *max_fd);

/* NAME
 *    process_next_ready_father
 * DESCRIPTION
 *    Sets the next child context and caller private data which has data
 *    available. Will modify *readfds.
 * RESULT
 *    FALSE when no process ready (should re-do select, e.g.).
 * NOTES
 *    - If the child has died, will set running to FALSE. Thus,
 *      you don't need to do anything special except process_select_father()
 *      then process_next_ready_father().
 * BUGS
 * TODO
 */
BOOL process_next_ready_father(process_context_t context,
                               child_context_t *child_context,
                               void **caller_private,
                               BOOL *running,
                               int *message_size,
                               fd_set *readfds,
                               fd_t max_fd);

/* NAME
 *    process_select_child
 * DESCRIPTION
 *    Adds the file descriptors for the father communication channel
 *    to the readfds array, and sets max_fd to be the higher fd.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void process_select_child(child_context_t child_context,
                          fd_set *readfds,
                          fd_t *max_fd);

/* NAME
 *    process_child_has_message
 * DESCRIPTION
 *    Returns TRUE if child has a message (must be used only after
 *    a process_select_child() and select()).
 * RESULT
 *    TRUE if has message
 * NOTES
 * BUGS
 * TODO
 */
BOOL process_child_has_message(child_context_t child_context,
                               fd_set *readfds,
                               fd_t *max_fd,
                               int *message_length);

/* NAME
 *    process_child_send_message
 * DESCRIPTION
 *    Sends a message to the father.
 * RESULT
 *    TRUE unless communication error, child not present, or message
 *    too big (see constant PROCESS_MAX_MESSAGE_LENGTH).
 * NOTES
 *    - May block
 * BUGS
 * TODO
 */
BOOL process_child_send_message(child_context_t child_context,
                                const unsigned char *message,
                                const int message_length);

/* NAME
 *    process_father_send_message
 * DESCRIPTION
 *    Sends a message to a child
 * RESULT
 *    TRUE unless communication error, child not present, or message
 *    too big (see constant PROCESS_MAX_MESSAGE_LENGTH).
 * NOTES
 *    - May block
 * BUGS
 * TODO
 */
BOOL process_father_send_message(child_context_t child_context,
                                 const unsigned char *message,
                                 const int message_length);

/* NAME
 *    process_child_receive_message_blocking
 * DESCRIPTION
 *    Receives a message to a child, blocking version, any length
 *    accepted if it is below the threshold and any_length is set.
 * RESULT
 *    TRUE unless communication error, child not present, etc.
 *    or message too big (see constant PROCESS_MAX_MESSAGE_LENGTH),
 *    or unexpected length.
 * NOTES
 *    - May block.
 *    - *message_length should be initialized to the maximum receiveable
 *      message size.
 *      It will be set to the actual received length.
 *    - CANNOT be used after the process_child_has_message() function.
 *    - The blocking/special difference is more in the fact that the
 *      *internal* (private to the process module) header was already
 *      consumed or not (by process_child_has_message()).
 * BUGS
 * TODO
 */
BOOL process_child_receive_message_blocking(child_context_t child_context,
					    unsigned char *message,
					    int *message_length,
                                            const BOOL any_length);

/* NAME
 *    process_child_receive_message_special
 * DESCRIPTION
 *    Receives a message to a child, non-blocking version, length
 *    known and process_child_has_message() called.
 * RESULT
 *    TRUE unless communication error, child not present, etc.
 *    or message too big (see constant PROCESS_MAX_MESSAGE_LENGTH),
 *    or unexpected length.
 * NOTES
 *    - May block.
 *    - *message_length should be initialized to the message size
 *      as set by process_child_has_message() function.
 *      It will be set to the actual received length, which must
 *      be equal to the initial value, else we won't return TRUE.
 * BUGS
 * TODO
 */
BOOL process_child_receive_message_special(child_context_t child_context,
					   unsigned char *message,
					   int *message_length);

/* NAME
 *    process_father_receive_message
 * DESCRIPTION
 *    Receives a message to the father
 * RESULT
 *    TRUE unless communication error, child not present, etc.
 *    or message too big (see constant PROCESS_MAX_MESSAGE_LENGTH),
 *    or unexpected length.
 * NOTES
 *    - May block (especially if this is not called because of a
 *      select).
 *    - *message_length should be initialized to the message size
 *      as set by process_next_ready_father().
 *      It will be set to the actual received length, which must
 *      be equal to the initial value, else we won't return TRUE.
 * BUGS
 * TODO
 */
BOOL process_father_receive_message(child_context_t child_context,
                                    unsigned char *message,
                                    int *message_length);

#endif /* !_PROCESS_INTERFACE_H */

