/* Project filed -- Intelligent file cache and migration system
 * protocol.c -- Protocol routines
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the protocol module. The protocol module offers client/server
 *    encrypted protocol sessions.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: protocol.c,v 1.1 2000/06/14 06:50:38 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "global.h"
#include "../debug/debug_interface.h"
#include "../misc/misc_interface.h"
#include "../crypt/crypt_interface.h"
#include "protocol_interface.h"

#define SEND_DATA_BUFFER_SIZE 32768
BOOL protocol_send_data(fd_t sd,
			protocol_info_t *info,
			const unsigned char *buffer,
			int buffer_length,
			BOOL normal) {
   unsigned char out_buffer[SEND_DATA_BUFFER_SIZE];
   BOOL result = TRUE;

   while (buffer_length && result) {
      int this_time = MIN(sizeof(out_buffer), buffer_length);

      protocol_encrypt_decrypt(info,
			       (unsigned char *) buffer,
			       out_buffer,
			       this_time,
			       normal,
			       TRUE);

      result = (write(sd, out_buffer, this_time) == this_time);
      if (!result) {
         debug_perror(DEBUG_LEVEL_ERROR, "write()");
      }

      buffer += this_time;
      buffer_length -= this_time;
   }

   return result;
}

#define RECEIVE_DATA_BUFFER_SIZE 32768
BOOL protocol_receive_data(fd_t sd,
			   protocol_info_t *info,
			   unsigned char *buffer,
			   int buffer_length,
			   BOOL normal) {
   unsigned char in_buffer[RECEIVE_DATA_BUFFER_SIZE];
   BOOL result = TRUE;

   while (buffer_length && result) {
      int this_time = MIN(sizeof(in_buffer), buffer_length);

      result = (misc_all_read(sd, in_buffer, this_time) == this_time);

      protocol_encrypt_decrypt(info,
                               in_buffer,
                               buffer,
                               this_time,
                               normal,
                               FALSE);
      buffer += this_time;
      buffer_length -= this_time;
   }

   return result;
}


void protocol_encrypt_decrypt(protocol_info_t *info,
			      unsigned char *in_buffer,
			      unsigned char *out_buffer,
			      unsigned int buffer_length,
			      BOOL normal,
			      BOOL encrypt) {
   void *crypt_context;
   BOOL *crypt_initial_ptr;

   if (normal) {
      if (encrypt) {
         crypt_context = info->encrypt_context_1;
         crypt_initial_ptr = &info->encrypt_initial_1;
      }
      else {
         crypt_context = info->decrypt_context_1;
         crypt_initial_ptr = &info->decrypt_initial_1;
      }
   }
   else {
      if (encrypt) {
         crypt_context = info->encrypt_context_2;
         crypt_initial_ptr = &info->encrypt_initial_2;
      }
      else {
         crypt_context = info->decrypt_context_2;
         crypt_initial_ptr = &info->decrypt_initial_2;
      }
   }

   crypt_do(crypt_context,
            in_buffer,
            buffer_length,
            out_buffer,
            *crypt_initial_ptr,
            encrypt);

   *crypt_initial_ptr = FALSE;
}

void protocol_encrypt(void *caller_private,
		      unsigned char *in_buffer,
		      unsigned char *out_buffer,
		      unsigned int buffer_length,
		      BOOL normal) {
   protocol_encrypt_decrypt((protocol_info_t *) caller_private,
			    in_buffer,
			    out_buffer,
			    buffer_length,
			    normal,
			    TRUE);
}

void protocol_decrypt(void *caller_private,
		      unsigned char *in_buffer,
		      unsigned char *out_buffer,
		      unsigned int buffer_length,
		      BOOL normal) {
   protocol_encrypt_decrypt((protocol_info_t *) caller_private,
			    in_buffer,
			    out_buffer,
			    buffer_length,
			    normal,
			    FALSE);
}

BOOL protocol_receive_and_decrypt_session_key(fd_t sd,
					      unsigned char *key,
					      const int key_length,
					      const unsigned char *secret,
					      const int secret_length) {
   BOOL result = FALSE;
   void *crypt_context = crypt_context_allocate();

   if (crypt_context) {
      if (crypt_context_initialize(crypt_context, secret, secret_length)) {
         unsigned char buffer[key_length];

         if (misc_all_read(sd, buffer, sizeof(buffer)) == sizeof(buffer)) {
            crypt_do(crypt_context,
                     buffer,
                     sizeof(buffer),
                     key,
                     TRUE,
                     FALSE);

            result = TRUE;
         }
      }

      crypt_context_free(&crypt_context);
   }

   return result;
}

BOOL protocol_send_encrypted_session_key(fd_t sd,
                                         const unsigned char *key,
                                         const int key_length,
                                         const unsigned char *secret,
                                         const int secret_length) {
   BOOL result = FALSE;
   void *crypt_context = crypt_context_allocate();

   if (crypt_context) {
      if (crypt_context_initialize(crypt_context, secret, secret_length)) {
         unsigned char buffer[key_length];

	 crypt_do(crypt_context,
                  key,
                  sizeof(buffer),
                  buffer,
                  TRUE,
                  TRUE);

         if (write(sd, buffer, sizeof(buffer)) == sizeof(buffer)) {
	    result = TRUE;
         }
         else {
            perror("write()");
         }
      }

      crypt_context_free(&crypt_context);
   }

   return result;
}


