/* Project filed -- Intelligent file cache and migration system
 * protocol_interface.h -- Interface definitions for the protocol module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the protocol module. The protocol module offers client/server
 *    encrypted protocol sessions.
 * NOTES
 * BUGS
 *    - This is a bit messy (less than it was, but still ...)
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: protocol_interface.h,v 1.3 2000/06/14 06:50:39 schaefer Exp $
 */

#ifndef _PROTOCOL_INTERFACE_H
#define _PROTOCOL_INTERFACE_H

/* Type definitions */

typedef struct {
   void *decrypt_context_1;
   void *encrypt_context_1;
   void *decrypt_context_2;
   void *encrypt_context_2;

   BOOL decrypt_initial_1;
   BOOL encrypt_initial_1;
   BOOL decrypt_initial_2;
   BOOL encrypt_initial_2;

   void *user_defined;
} protocol_info_t;

/* Function definitions */

/* NAME
 *    protocol_send_data
 * DESCRIPTION
 *    Sends some data to the other side, which will be encrypted
 *    through protocol_encrypt(), using the appropriate output context
 *    and continuation.
 * RESULT
 *    BOOL, FALSE if output failed.
 * NOTES
 * BUGS
 * TODO
 */
BOOL protocol_send_data(fd_t sd,
			protocol_info_t *info,
			const unsigned char *buffer,
			int buffer_length,
			BOOL normal);

/* NAME
 *    protocol_receive_data
 * DESCRIPTION
 *    Receives some data from the other side, and decrypt it through
 *    protocol_decrypt() using the appropriate input context and
 *    continuation.
 * RESULT
 *    BOOL, FALSE if input failed.
 * NOTES
 * BUGS
 * TODO
 */
BOOL protocol_receive_data(fd_t sd,
			   protocol_info_t *info,
			   unsigned char *buffer,
			   int buffer_length,
			   BOOL initial);

/* NAME
 *    protocol_encrypt_decrypt
 * DESCRIPTION
 *    Encrypts or decrypts a buffer, using either normal or
 *    secondary encryption key, and doing either encryption
 *    or decryption based on encrypt. This will use the
 *    appropriate input or output context and continuation.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void protocol_encrypt_decrypt(protocol_info_t *info,
			      unsigned char *in_buffer,
			      unsigned char *out_buffer,
			      unsigned int buffer_length,
			      BOOL normal,
			      BOOL encrypt);

/* NAME
 *    protocol_encrypt
 * DESCRIPTION
 *    Wrapper for protocol_encrypt_decrypt, with encrypt value TRUE.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void protocol_encrypt(void *caller_private,
		      unsigned char *in_buffer,
		      unsigned char *out_buffer,
		      unsigned int buffer_length,
		      BOOL normal);

/* NAME
 *    protocol_decrypt
 * DESCRIPTION
 *    Wrapper for protocol_encrypt_decrypt, with encrypt value FALSE.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void protocol_decrypt(void *caller_private,
		      unsigned char *in_buffer,
		      unsigned char *out_buffer,
		      unsigned int buffer_length,
		      BOOL normal);

/* NAME
 *    protocol_receive_and_decrypt_session_key
 * DESCRIPTION
 *    Outside of any context, will receive a random key and will
 *    decrypt it with the supplied secret.
 * RESULT
 *    BOOL if input succeeded.
 * NOTES
 * BUGS
 * TODO
 */
BOOL protocol_receive_and_decrypt_session_key(fd_t sd,
					      unsigned char *key,
					      const int key_length,
					      const unsigned char *secret,
					      const int secret_length);

/* NAME
 *    protocol_send_encrypted_session_key
 * DESCRIPTION
 *    Outside of any context, will send a random key and will
 *    encrypt it first with the supplied secret.
 * RESULT
 *    BOOL if output succeeded
 * NOTES
 * BUGS
 * TODO
 */
BOOL protocol_send_encrypted_session_key(fd_t sd,
                                         const unsigned char *key,
                                         const int key_length,
                                         const unsigned char *secret,
                                         const int secret_length);


#endif /* ! _PROTOCOL_INTERFACE_H */
