/* Project filed -- Intelligent file cache and migration system
 * scsi.c -- High level changer support module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    scsi module. The scsi modules offers simple SCSI transport
 *    layer low-level features.
 * NOTES
 *    - Documentation is available at:
 *         http://www.torque.net/sg/
 *    - This is based on http://www.torque.net/sg/p/scsi-generic_long.txt
 *      (2.2.x, x >= 6).
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: scsi.c,v 1.3 2000/07/07 12:53:12 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/ioctl.h> 
#include <scsi/scsi.h>
#include <scsi/sg.h>
#include <asm/param.h> /* HZ */
#include <stdarg.h>
#include "global.h"
#include "../debug/debug_interface.h"
#include "scsi_interface.h"

/* Private type definitions */

#undef  DEBUG_SCSI_DUMP_THE_DATA
#undef  DEBUG_SCSI_DUMP_THE_COMMAND
#define DEBUG_SCSI_DUMP_THE_SENSE /* and the command if !above */

#if defined(DEBUG_SCSI_DUMP_THE_DATA) || defined(DEBUG_SCSI_DUMP_THE_COMMAND) || defined(DEBUG_SCSI_DUMP_THE_SENSE)
#define DEBUG_SCSI_ENABLE_DUMP_DATA
#endif /* any dump data */

typedef struct {
   fd_t fd;
   BOOL in_error;
   scsi_error_t error;
} scsi_info_t;

/* BUGS
 *    - Should be in sg_interface.h or similar.
 */
#define DRIVER_SENSE 8
#define DID_OK 0

/* Private function definitions */

#ifdef DEBUG_SCSI_ENABLE_DUMP_DATA
/* NAME
 *    dump_data
 * DESCRIPTION
 *    Prints a buffer content in hexadecimal.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void dump_data(const char *description,
                      const unsigned char *buffer,
                      const unsigned int len);
#endif /* DEBUG_SCSI_ENABLE_DUMP_DATA */

/* Function implementation */

scsi_context_t scsi_initialize(const char *changer_name) {
   scsi_info_t *info = (scsi_info_t *) malloc(sizeof(scsi_info_t));

   if (info) {
      info->in_error = FALSE;
      info->fd = open(changer_name, O_RDWR | O_EXCL);
      if (info->fd == -1) {
         debug_perror(DEBUG_LEVEL_ERROR, "open()");
         free(info);
         info = NULL;
      }
   }

   return (scsi_context_t) info;
}

void scsi_deinitialize(scsi_context_t *context) {
   if (close(((scsi_info_t *) *context)->fd)) {
      debug_perror(DEBUG_LEVEL_ERROR, "close()");
   }

   *context = NULL;
}

BOOL scsi_execute(scsi_context_t context,
                  unsigned char *scsi_cmd,
                  unsigned char scsi_cmd_length,
                  unsigned char *input_buffer, /* from target */
                  unsigned int *input_buffer_length,
                  unsigned char *output_buffer,
                  unsigned int *output_buffer_length,
                  scsi_timeout_t timeout) {
   scsi_info_t *info = (scsi_info_t *) context;
   BOOL result = FALSE;
   int tmo = timeout * HZ; /* BUGS: no overflow check */

#ifdef DEBUG_SCSI_DUMP_THE_COMMAND
   dump_data("SCSI command", scsi_cmd, scsi_cmd_length);
#endif /* DEBUG_SCSI_DUMP_THE_COMMAND */

   /* BUGS
    *    - We allocate the buffer on the stack.
    *    - Should we do the UNIT ATTENTION retries ?
    */

   if (ioctl(info->fd, SG_SET_TIMEOUT, &tmo)) {
      debug_perror(DEBUG_LEVEL_ERROR, "ioctl(TIMEOUT)");
   }
   else {
      unsigned int in_length = sizeof(struct sg_header)
                               + (input_buffer_length ? *input_buffer_length
                                                      : 0);
      unsigned int out_length = sizeof(struct sg_header)
                                + scsi_cmd_length
                                + (output_buffer_length ? *output_buffer_length
                                                        : 0);
      unsigned char buffer[MAX(out_length, in_length)];
      struct sg_header *header = (struct sg_header *) buffer;
   
      header->reply_len = in_length;
      header->pack_id = 0;
   
      memcpy(&buffer[sizeof(struct sg_header)],
             scsi_cmd,
             scsi_cmd_length);

      if (output_buffer_length && *output_buffer_length) {
         memcpy(&buffer[sizeof(struct sg_header) + scsi_cmd_length],
                output_buffer,
                *output_buffer_length);
      }
      
      if (write(info->fd, buffer, out_length) != out_length) {
         debug_perror(DEBUG_LEVEL_ERROR, "write()");
      }
      else {
         if (read(info->fd, buffer, in_length) != in_length) {
            debug_perror(DEBUG_LEVEL_ERROR, "read()");
         }
         else {
            if ((header->target_status == CHECK_CONDITION)
                || (header->target_status == COMMAND_TERMINATED)
                || (header->driver_status & DRIVER_SENSE)) {
#ifdef DEBUG_SCSI_DUMP_THE_SENSE
#ifndef DEBUG_SCSI_DUMP_THE_COMMAND
	       dump_data("Failed SCSI command", scsi_cmd, scsi_cmd_length);
#endif /* DEBUG_SCSI_DUMP_THE_COMMAND */
               dump_data("sense",
                         &header->sense_buffer[0],
                         sizeof(header->sense_buffer));
#endif /* DEBUG_SCSI_DUMP_THE_SENSE */

               info->error.sense = header->sense_buffer[2] & 0xF;
               info->error.asc = header->sense_buffer[12];
               info->error.ascq = header->sense_buffer[13];
               info->in_error = TRUE;

               debug_log(DEBUG_LEVEL_WARN,
                         "SENSE: 0x%x 0x%x 0x%x.\n",
                         info->error.sense,
                         info->error.asc,
                         info->error.ascq);
            }
            else if (header->host_status == DID_OK) {
#ifdef DEBUG_SCSI_DUMP_THE_DATA
               dump_data("returned data",
                         buffer + sizeof(struct sg_header), 
                         in_length - sizeof(struct sg_header));
#endif /* DEBUG_SCSI_DUMP_THE_DATA */
               info->in_error = FALSE;
               result = TRUE;

	       if (input_buffer_length && *input_buffer_length) {
		  memcpy(input_buffer,
                         &buffer[sizeof(struct sg_header)],
			 *input_buffer_length);
	       }

               /* BUGS
                *    - size is not actualized (actually sent, actually
                *      received.
                */
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR,
                         "Error: hs 0x%x ds 0x%x ts 0x%x.\n",
                         header->host_status,
                         header->driver_status,
                         header->target_status);
               /* BUGS
                *    - This causes a problem with the sense.
                */
            }
         }
      }
   }

   return result;
}

scsi_error_t scsi_get_error(scsi_context_t context) {
   scsi_info_t *info = (scsi_info_t *) context;

   if (info->in_error) {
      return info->error;
   }
   else {
      scsi_error_t error; /* zero since automatic variable */

      return error;
   }
}

const char *scsi_error_description(scsi_error_t error) {
   /* To be fixed */
   return "had a SCSI error";
}

/* Private function implementation */

#ifdef DEBUG_SCSI_ENABLE_DUMP_DATA
static void dump_data(const char *description,
                      const unsigned char *buffer,
                      const unsigned int len) {
   unsigned int i = 1;

   printf("%s: ", description);

   while (i <= len) {
      if ((i % 10) == 0) {
         printf("\n");
      }

      printf("0x%2.2x ", buffer[i - 1]);

      i++;
   }

   printf("\n");
}
#endif /* DEBUG_SCSI_ENABLE_DUMP_DATA */


