/* Project filed -- Intelligent file cache and migration system
 * scsi_interface.h -- Interface definitions for the scsi module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    scsi module. The scsi modules offers simple SCSI transport
 *    layer low-level features.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: scsi_interface.h,v 1.1 2000/06/15 17:32:00 schaefer Exp $
 */

#ifndef _SCSI_INTERFACE_H
#define _SCSI_INTERFACE_H

/* Type definitions */

typedef void * scsi_context_t;

typedef struct {
   unsigned char sense;
   unsigned char asc;
   unsigned char ascq;
} scsi_error_t;

typedef unsigned int scsi_timeout_t; /* in seconds */

/* Function definitions */

/* NAME
 *    scsi_initialize
 * DESCRIPTION
 *    Initializes a scsi module instance. Passed is the scsi_device_name,
 *    which describes a SCSI device name.
 * RESULT
 *    scsi_context_t, NULL if initialization failed.
 * NOTES
 *    - At this time the only supported SCSI device name is:
 *         "/dev/sgX"
 *      where X is the number of the SCSI generic interface
 *      on Linux.
 * BUGS
 * TODO
 */
scsi_context_t scsi_initialize(const char *changer_name);

/* NAME
 *    scsi_deinitialize
 * DESCRIPTION
 *    Deinitializes a scsi module instance.
 * RESULT
 *    NONE
 * NOTES
 *    - context will be set to NULL
 * BUGS
 * TODO
 */
void scsi_deinitialize(scsi_context_t *context);

/* NAME
 *    scsi_execute
 * DESCRIPTION
 *    Executes a SCSI command
 * RESULT
 *    TRUE if command executed correctly.
 * NOTES
 *    - Updates the length if not NULL to actually sent.
 * BUGS
 * TODO
 */
BOOL scsi_execute(scsi_context_t context,
                  unsigned char *scsi_cmd,
                  unsigned char scsi_cmd_length,
                  unsigned char *input_buffer, /* from target */
                  unsigned int *input_buffer_length,
                  unsigned char *output_buffer,
                  unsigned int *output_buffer_length,
                  scsi_timeout_t timeout);

/* NAME
 *    scsi_get_error
 * DESCRIPTION
 *    Returns an error code for the given instance. This should be
 *    called when a function returns FALSE.
 * RESULT
 *    scsi_error_t
 * NOTES
 * BUGS
 * TODO
 */
scsi_error_t scsi_get_error(scsi_context_t context);

/* NAME
 *    scsi_error_description
 * DESCRIPTION
 *    Returns an error description for the given scsi error.
 * RESULT
 *    const char *
 * NOTES
 * BUGS
 * TODO
 */
const char *scsi_error_description(scsi_error_t error);

#endif /* !_SCSI_INTERFACE_H */
