/* Project filed -- Intelligent file cache and migration system
 * tapeio.c -- Tape I/O and movement services.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the tapeio module.  The tapeio modules offers simple tape
 *    services.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: tapeio.c,v 1.5 2000/08/05 20:10:50 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <sys/mtio.h>
#include <unistd.h>
#include "global.h"
#include "debug_interface.h"
#include "tapeio_interface.h"

/* Type definitions */

#define TAPEIO_LABEL_SIZE 32768
#define TAPEIO_LABEL_HEADER "FILED TAPEIO TAPE"

/* Private function definitions */

/* Function implementation */

fd_t tapeio_open(const char *device_name, BOOL readonly) {
   return open(device_name,
               readonly ? O_RDONLY : O_RDWR);
}

BOOL tapeio_label_read(fd_t fd,
                       tapeio_label_t *label) {
   unsigned char buffer[TAPEIO_LABEL_SIZE + 1];

   if (read(fd, buffer, TAPEIO_LABEL_SIZE) == TAPEIO_LABEL_SIZE) {
      if (!strncmp(buffer,
                   TAPEIO_LABEL_HEADER,
                   sizeof(TAPEIO_LABEL_HEADER) - 1)) {
         char *name;
         char *date;

         buffer[TAPEIO_LABEL_SIZE] = '\0';

         name = &buffer[sizeof(TAPEIO_LABEL_HEADER)];
         date = strchr(name, '\n');
         if (date && (((int) date - (int) buffer) < (sizeof(buffer) - 1))) {
            char *tmp;

            *date = '\0';
            date++;

            tmp = strchr(date, '\n');
	    if (tmp && (((int) tmp - (int) buffer) < sizeof(buffer))) {
               *tmp = '\0';

               label->name = strdup(name);
               label->date = strdup(date);

               if (label->name && label->date) {
                  return TRUE;
               }
               else {
                  tapeio_label_free(label);
               }
            }
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "tapeio_read_label: not a filed tapeio tape.\n");
      }
   }
   else {
      debug_perror(DEBUG_LEVEL_ERROR, "read()");
   }

   return FALSE;
}

BOOL tapeio_label_write(fd_t fd,
                        const tapeio_label_t *label) {
   char buffer[TAPEIO_LABEL_SIZE];

   memset(buffer, ' ', sizeof(buffer));

   if (snprintf(buffer,
                sizeof(buffer),
                "%s\n%s\n%s\n",
                TAPEIO_LABEL_HEADER,
                label->name,
                label->date) > -1) {
      if (write(fd, buffer, TAPEIO_LABEL_SIZE) == TAPEIO_LABEL_SIZE) {
         return TRUE;
      }
      else {
         debug_perror(DEBUG_LEVEL_ERROR, "write()");
      }
   }
   return FALSE;
}

void tapeio_label_free(tapeio_label_t *label) {
   if (label->name) {
      free(label->name);
   }

   if (label->date) {
      free(label->date);
   }
}

BOOL tapeio_rewind(fd_t fd) {
   struct mtop mt_cmd = { MTREW, 1 };

   return ioctl(fd, MTIOCTOP, &mt_cmd) ? FALSE : TRUE;
}

BOOL tapeio_set_position(fd_t fd,
                         tapeio_position_t position) {
   tapeio_position_t current;

   if (tapeio_get_position(fd, &current)) {
      struct mtop mt_cmd;

      if (current >= position) {
         /* @@@ this needs checking @@@ */
         mt_cmd.mt_op = MTBSFM;
         mt_cmd.mt_count = (current - position) + 1;
      }
      else {
         mt_cmd.mt_op = MTFSF;
         mt_cmd.mt_count = position - current;
      }

      return ioctl(fd, MTIOCTOP, &mt_cmd) ? FALSE : TRUE;
   }

   return FALSE;
}

BOOL tapeio_get_position(fd_t fd,
                         tapeio_position_t *position) {
   struct mtop mt_cmd = { MTNOP, 1 };
   struct mtget mt_status;

   /* Recommended by manpage st(4) */
   if (ioctl(fd, MTIOCTOP, &mt_cmd)) {
      return FALSE;
   }

   if (ioctl(fd, MTIOCGET, &mt_status)) {
      return FALSE;
   }

   *position = mt_status.mt_fileno;
   return TRUE;

}

BOOL tapeio_write_file_mark(fd_t fd) {
   struct mtop mt_cmd = { MTWEOF, 1 };

   return ioctl(fd, MTIOCTOP, &mt_cmd) ? FALSE : TRUE;
}

BOOL tapeio_set_position_for_appending(fd_t fd) {
   struct mtop mt_cmd = { MTEOM, 1 };

   return ioctl(fd, MTIOCTOP, &mt_cmd) ? FALSE : TRUE;
}

BOOL tapeio_enable_writing(fd_t fd, BOOL enable) {
#if 1
   /* BUGS
    *    - see interface specification.
    */
   return TRUE;
#else
   int flags = fcntl(fd, F_GETFL);

   if (flags == -1) {
      debug_perror(DEBUG_LEVEL_ERROR, "fcntl(F_GETFL)");
   }
   else {
      if (enable) {
         flags |= O_RDWR;
         flags &= ~(O_RDONLY);
      }
      else {
         flags &= ~(O_RDWR);
         flags |= O_RDONLY;
      }

      if (fcntl(fd, F_SETFL, flags) == -1) {
         debug_perror(DEBUG_LEVEL_ERROR, "fcntl(F_SETFL)");
      }
      else {
 	 debug_log(DEBUG_LEVEL_INFO, "flags 0x%x.\n", fcntl(fd, F_GETFL));
         return TRUE;
      }
   }

   return FALSE;
#endif
}