/* Project filed -- Intelligent file cache and migration system
 * tapeio_interface.h -- Interface definitions for the tapeio module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    tapeio module. The tapeio modules offers simple tape
 *    services.
 * NOTES
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: tapeio_interface.h,v 1.4 2000/08/05 20:10:50 schaefer Exp $
 */

#ifndef _TAPEIO_INTERFACE_H
#define _TAPEIO_INTERFACE_H

/* Type definitions */

typedef struct {
   char *name;
   char *date;
} tapeio_label_t;

typedef unsigned int tapeio_position_t;

/* Function definitions */

/* NAME
 *    tapeio_open
 * DESCRIPTION
 *    Opens a tape device.
 * RESULT
 *    fd_t, -1 for error.
 * NOTES
 *    - You can close it with close().
 *    - Once open, if the tape is removed from the device,
 *      an error will occur. Also, no other opening will
 *      succeed.
 * BUGS
 * TODO
 */
fd_t tapeio_open(const char *device_name, BOOL readonly);

/* NAME
 *    tapeio_label_read
 * DESCRIPTION
 *    Reads the tape label into the label array. 
 * RESULT
 *    TRUE if no I/O error and the label is in standard
 *    format.
 * NOTES
 *    - You may need to to a tapeio_rewind() if you
 *      just tapeio_open()ed the device or if
 *      you have written to, read from, or done any
 *      movement command.
 *    - Strings are allocated and must be freed by a call to
 *      tapeio_label_free().
 *    - The tapeio_label_t structure is caller-allocated. It must be allocated
 *      so that it contains all NUL.
 * BUGS
 * TODO
 */
BOOL tapeio_label_read(fd_t fd,
                       tapeio_label_t *label);

/* NAME
 *    tapeio_label_write
 * DESCRIPTION
 *    Writes the tape label from the label array.
 * RESULT
 *    TRUE if no I/O error, and label not too big.
 * NOTES
 *    - You may need to to a tapeio_rewind() if you
 *      just tapeio_open()ed the device or if
 *      you have written to, read from, or done any
 *      movement command.
 *    - This is a destructive operation: it deletes
 *      the whole tape's data.
 * BUGS
 * TODO
 */
BOOL tapeio_label_write(fd_t fd,
                        const tapeio_label_t *label);


/* NAME
 *    tapeio_label_free
 * DESCRIPTION
 *    Frees the strings in the label structure
 * RESULT
 *    NONE
 * NOTES
 *    - The tapeio_label_t structure itself is
 *      caller-allocated. It must be allocated
 *      so that it contains all NUL.
 * BUGS
 * TODO
 */
void tapeio_label_free(tapeio_label_t *label);

/* NAME
 *    tapeio_rewind
 * DESCRIPTION
 *    Rewinds the tape.
 * RESULT
 *    TRUE if no I/O error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapeio_rewind(fd_t fd);

/* NAME
 *    tapeio_set_position_for_appending
 * DESCRIPTION
 *    Positions the tape at the end of all data, position suitable
 *    for a subsequent write (append).
 * RESULT
 *    TRUE if no I/O error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapeio_set_position_for_appending(fd_t fd);

/* NAME
 *    tapeio_set_position
 * DESCRIPTION
 *    Positions the tape at the beginning of the specified
 *    file (which is given absolute from the beginning,
 *    0 being the first file, ie the label of the tape).
 *    Files are separated by BOT or filemarks.
 * RESULT
 *    TRUE if no I/O error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapeio_set_position(fd_t fd,
                         tapeio_position_t position);

/* NAME
 *    tapeio_get_position
 * DESCRIPTION
 *    Sets the position variable to the current tape position.
 * RESULT
 *    TRUE if no I/O error.
 * NOTES
 * BUGS
 * TODO
 */
BOOL tapeio_get_position(fd_t fd,
                         tapeio_position_t *position);

/* NAME
 *    tapeio_write_file_mark
 * DESCRIPTION
 *    Writes a file mark on the tape.
 * RESULT
 *    TRUE if no I/O error.
 * NOTES
 *    - Implies a flush of the data to the medium.
 * BUGS
 * TODO
 */
BOOL tapeio_write_file_mark(fd_t fd);

/* NAME
 *    tapeio_enable_writing
 * DESCRIPTION
 *    Enable the tape fd for writing.
 * RESULT
 *    TRUE if succeeded.
 * NOTES
 * BUGS
 *    - Doesn't work (see man 2 open). Thus
 *      this is a NO-OP, and taper.c had to
 *      always open the device rw.
 * TODO
 */
BOOL tapeio_enable_writing(fd_t fd, BOOL enable);

#endif /* !_TAPEIO_INTERFACE_H */
