/* Project filed -- Intelligent file cache and migration system
 * chg.c -- High level changer support module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the chg module. The chg module offers a high level
 *    medium changer interface.
 * NOTES
 * BUGS
 *    - Implements a restrictive view on a changer: one arm, etc.
 *    - Has built-in assumptions for device names, etc.
 * TODO
 *    - Routine for checking if a label exists already, implement in
 *      changer_tell_label.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: changer.c,v 1.8 2000/07/09 12:37:04 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "global.h"
#include "debug_interface.h"
#include "chg_interface.h"
#include "changer_interface.h"

/* Private type definitions */

typedef struct {
   chg_context_t chg_context;

   chg_elt_t *elements;
   unsigned int elements_count;

   chg_elt_number_t current_slot;
} changer_info_t;

/* Private function definitions */

/* NAME
 *    dump_element_status
 * DESCRIPTION
 *    Dumps the content of the element status, including
 *    our data.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void dump_element_status(chg_elt_t *elements, unsigned int count);

/* NAME
 *    free_element_status
 * DESCRIPTION
 *    Frees the element status including our data.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void free_element_status(changer_info_t *info);

/* NAME
 *    load_slot_with_unload
 * DESCRIPTION
 *    Puts back the tape in the drive into its original slot, if any,
 *    and then loads the specified slot into the drive.
 * RESULT
 *    changer_error_t
 * NOTES
 * BUGS
 * TODO
 */
static changer_error_t load_slot_with_unload(changer_info_t *info,
					     const chg_elt_number_t slot,
					     const char *tape_device_name);

/* NAME
 *    load_slot
 * DESCRIPTION
 *    Asks the low-level chg subsystem to load the specific slot
 *    into the specified drive. Updates the emptiness of both
 *    elements and the from of the data transfer element.
 * RESULT
 *    changer_error_t
 * NOTES
 *    - There shall be no tape in the specified drive.
 *    - Low-level use only: doesn't do anything about tape labels,
 *      or any bookkeeping.
 * BUGS
 * TODO
 */
static changer_error_t load_slot(changer_info_t *info,
                                 const chg_elt_number_t slot,
                                 const chg_elt_number_t tape);

/* NAME
 *    unload_slot
 * DESCRIPTION
 *    Puts the tape in the specified drive into the specified slot.
 *    Updates the emptiness of both elements and the from of the data
 *    transfer element.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE
 *          Operation succeeded.
 *       CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 *          chg error.
 * NOTES
 *    - There shall be no tape in the specified drive.
 *    - Low-level use only: doesn't do anything about tape labels,
 *      or any bookkeeping.
 * BUGS
 * TODO
 */
static changer_error_t unload_slot(changer_info_t *info,
                                   const chg_elt_number_t slot,
                                   const chg_elt_number_t tape);

/* NAME
 *    may_unload_drive
 * DESCRIPTION
 *    If the named drive is full, empties it.
 * RESULT
 *    CHANGER_ERROR_NO_TAPE
 *       Drive is not full.
 *    CHANGER_DRIVE_NOT_KNOWN
 *       Named drive unknown.
 *    see also the error codes from unload_slot.
 * NOTES
 * BUGS
 * TODO
 */
static changer_error_t may_unload_drive(changer_info_t *info,
                                        chg_elt_t *tape_elt);

/* NAME
 *    elt_of_tape_device_name
 * DESCRIPTION
 *    Returns a mapping from tape_device_name to chg element.
 * RESULT
 *    chg_elt_number_t, possibly CHG_ELT_NUMBER_INVALID
 * NOTES
 * BUGS
 *    - Very restricted mapping. See comments in the implementation.
 * TODO
 */
static chg_elt_number_t elt_of_tape_device_name(const char *tape_device_name);

/* NAME
 *    find_next_element
 * DESCRIPTION
 *    Finds the next full or empty (storage) element, starting at current. If
 *    current is CHG_ELT_NUMBER_INVALID, returns the first. Will
 *    return CHG_ELT_NUMBER_INVALID if there are no full slots.
 * RESULT
 *    chg_elt_number_t, possibly CHG_ELT_NUMBER_INVALID.
 * NOTES
 *    - This function will loop over to the start of elements, but of course
 *      not infinitely.
 * BUGS
 * TODO
 */
static chg_elt_number_t find_next_element(changer_info_t *info,
                                          const chg_elt_number_t current,
                                          BOOL full);

/* NAME
 *    find_tape_from_label
 * DESCRIPTION
 *    Finds the named tape in the current database.
 * RESULT
 *    chg_elt_t *, or NULL if not found.
 * NOTES
 *    - The returned element can be either a slot (storage element) or a drive
 *      (data transfer element).
 * BUGS
 * TODO
 */
static chg_elt_t *find_tape_from_label(changer_info_t *info,
                                       const char *tape_label);

/* NAME
 *    update_data_xfer_from
 * DESCRIPTION
 *    Updates the fullness state of the drive and slot, and
 *    if inserts, sets the from slot in the tape.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void update_data_xfer_from(changer_info_t *info,
				  const chg_elt_number_t tape,
				  const chg_elt_number_t slot,
				  const BOOL inserts);

/* NAME
 *    elt_of_element
 * DESCRIPTION
 *    Return the element structure from the element number, NULL
 *    if inexistant.
 * RESULT
 *    chg_elt_t *, possibly NULL.
 * NOTES
 * BUGS
 * TODO
 */
static chg_elt_t *elt_of_element(changer_info_t *info,
                                 const chg_elt_number_t slot);

/* NAME
 *    foreach_elt
 * DESCRIPTION
 *    Until no matching element exist, or until a callback returns FALSE,
 *    apply the callback (actor) on the element.
 * RESULT
 *    TRUE if no callback function returned FALSE.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL foreach_elt(changer_info_t *info, 
                        const chg_elt_type_t type,
                        BOOL actor(changer_info_t *info,
                                   chg_elt_t *elt));

/* NAME
 *    empty_drive_if_full
 * DESCRIPTION
 *    If the drive (data transfer element) is full, attempt to
 *    retrieve its medium into the coming slot, as given by
 *    chg. If the coming slot was CHG_ELT_NUMBER_INVALID,
 *    put it in the first empty slot. If no empty slot is available, or
 *    if an error occurs, return FALSE.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL empty_drive_if_full(changer_info_t *info,
                                chg_elt_t *elt);

/* Function implementation */

changer_context_t changer_initialize(const char *changer_name) {
   changer_info_t *info = (changer_info_t *) malloc(sizeof(changer_info_t));

   if (info) {
      info->elements = NULL;

      info->chg_context = chg_initialize(changer_name);
      if (info->chg_context) {
         unsigned int actual_size;

         if (chg_disable_operator_access(info->chg_context)
             && chg_initialize_element_status(info->chg_context,
                                              NULL,
                                              0,
                                              &actual_size)
             && actual_size) {
            info->elements = (chg_elt_t *) malloc(sizeof(chg_elt_t)
                                                  * actual_size);
            if (info->elements) {
               if (chg_initialize_element_status(info->chg_context,
                                                 info->elements,
                                                 actual_size,
                                                 &info->elements_count)) {
                  if (info->elements_count != actual_size) {
		     debug_log(DEBUG_LEVEL_ERROR,
			       "Mismatch in elt size: %u/%u.\n",
                               info->elements_count,
                               actual_size);
		     changer_deinitialize((changer_context_t *) &info);
                  }
                  else {
                     dump_element_status(info->elements, info->elements_count);

                     info->current_slot = CHG_ELT_NUMBER_INVALID;

                     if (!foreach_elt(info,
                                      CHG_ELT_TYPE_DATA_TRANSFER,
                                      empty_drive_if_full)) {
			debug_log(DEBUG_LEVEL_ERROR,
				  "Error while emptying drives.\n");
			changer_deinitialize((changer_context_t *) &info);
                     }
                     else {
                        dump_element_status(info->elements,
                                            info->elements_count);
                     }
                  }
               }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR,
			    "Can't chg_initialize_element_status2(): %d: %s\n",
			    chg_get_error(info->chg_context),
			    chg_error_description
			       (chg_get_error(info->chg_context)));
		  changer_deinitialize((changer_context_t *) &info);
	       }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR,
                         "Can't malloc() %u elts.\n",
                         actual_size);
               changer_deinitialize((changer_context_t *) &info);
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Can't chg_initialize_element_status(): %d: %s.\n",
                      chg_get_error(info->chg_context),
                      chg_error_description
                         (chg_get_error(info->chg_context)));
            changer_deinitialize((changer_context_t *) &info);
         }
      }
      else {
         free(info);
         info = NULL;
      }
   }

   /* If deinitialized, info was NULLified */
   return (changer_context_t) info;
}

void changer_deinitialize(changer_context_t *context) {
   chg_enable_operator_access(((changer_info_t *) *context)->chg_context);
   chg_deinitialize(&(((changer_info_t *) *context)->chg_context));
   if (((changer_info_t *) *context)->elements) {
      free_element_status(((changer_info_t *) *context));
   }
   free(*context);
   *context = NULL;
}

const char *changer_error_description(const changer_error_t error) {
   switch (error) {
      case CHANGER_ERROR_NONE:
         return "NONE";
         break;
      case CHANGER_ERROR_LABEL_NOT_KNOWN:
         return "Unknown label";
         break;
      case CHANGER_ERROR_NO_TAPE:
         return "No tape";
         break;
      case CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED:
         return "Operator intervention required";
         break;
      case CHANGER_ERROR_FAILED_INTERNAL:
         return "Internal error";
         break; 
      case CHANGER_ERROR_DRIVE_NOT_KNOWN:
         return "Tape drive unknown";
         break;
      case CHANGER_ERROR_MEDIUM_BUSY:
         return "Medium busy";
         break;
      default:
         return "UNKNOWN";
         break;
   }
   /* NOT REACHED */
}

changer_error_t changer_load_tape(changer_context_t context,
				  const char *tape_label,
				  const char *tape_device_name) {
   changer_info_t *info = (changer_info_t *) context;
   chg_elt_t *slot_elt; /* or drive! */

   slot_elt = find_tape_from_label(info, tape_label);
   if (slot_elt) {
      if (slot_elt->type == CHG_ELT_TYPE_DATA_TRANSFER) {
         if (slot_elt->number == elt_of_tape_device_name(tape_device_name)) {
            return CHANGER_ERROR_NONE;
         }
         else {
            return CHANGER_ERROR_MEDIUM_BUSY;
         }
      }
      else {
         return load_slot_with_unload(info,
                                      slot_elt->number,
                                      tape_device_name);
      }
   }
   else {
      return CHANGER_ERROR_LABEL_NOT_KNOWN;
   }
   /* NOT REACHED */
}

changer_error_t changer_unload_tape(changer_context_t context,
                                    const char *tape_device_name) {
   chg_elt_number_t tape;
   changer_info_t *info = (changer_info_t *) context;

   tape = elt_of_tape_device_name(tape_device_name);
   if (tape != CHG_ELT_NUMBER_INVALID) {
      chg_elt_t *tape_elt = elt_of_element(info, tape);

      if (tape_elt) {

         return may_unload_drive(info, tape_elt);
      }
      else {
	 return CHANGER_ERROR_DRIVE_NOT_KNOWN;
      }
   }
   else {
      return CHANGER_ERROR_DRIVE_NOT_KNOWN;
   }
   /* NOT REACHED */
}

changer_error_t changer_next_tape(changer_context_t context,
                                  const char *tape_device_name,
                                  const BOOL first,
                                  unsigned int *slot_identification,
                                  unsigned int *slot_stop_at) {
   changer_info_t *info = (changer_info_t *) context;
   chg_elt_number_t next_slot;

   if (first) {
      next_slot = find_next_element(info, CHG_ELT_NUMBER_INVALID, TRUE);
   }
   else if (info->current_slot == CHG_ELT_NUMBER_INVALID) {
      return CHANGER_ERROR_NO_TAPE;
   }
   else {
      next_slot = find_next_element(info, info->current_slot, TRUE);
   }

   if ((next_slot == CHG_ELT_NUMBER_INVALID)
       || (slot_stop_at && (next_slot == *slot_stop_at))) {
      return CHANGER_ERROR_NO_TAPE;
   }

   if (slot_identification) {
      *slot_identification = next_slot;
   }

   if (first && slot_stop_at) {
      *slot_stop_at = next_slot;
   }

   return load_slot_with_unload(info, next_slot, tape_device_name);
}

changer_error_t changer_load_specific(changer_context_t context,
                                      const char *tape_device_name,
                                      const unsigned int slot) {
   return load_slot_with_unload((changer_info_t *) context,
                                slot,
                                tape_device_name);
}

changer_error_t changer_tell_label(changer_context_t context,
                                   const char *tape_label,
                                   const char *tape_device_name) {
   changer_info_t *info = (changer_info_t *) context;
   chg_elt_number_t tape;

   tape = elt_of_tape_device_name(tape_device_name);
   if (tape != CHG_ELT_NUMBER_INVALID) {
      chg_elt_t *tape_elt;

      tape_elt = elt_of_element(info, tape);
      if (tape_elt) {
         if (tape_elt->user_data != NULL) {
            free(tape_elt->user_data);
         }

         tape_elt->user_data = strdup(tape_label);
         if (tape_elt->user_data == NULL) {
            return CHANGER_ERROR_FAILED_INTERNAL;
         }
      }
      else {
         return CHANGER_ERROR_DRIVE_NOT_KNOWN;
      }
   }
   else {
      return CHANGER_ERROR_DRIVE_NOT_KNOWN;
   }

   return CHANGER_ERROR_NONE;
}

void changer_dump_element_status(changer_context_t context) {
   changer_info_t *info = (changer_info_t *) context;

   dump_element_status(info->elements, info->elements_count);
}

/* Private function implementation */

static void dump_element_status(chg_elt_t *elements, unsigned int count) {
   if (elements && count) {
      chg_elt_t *current = elements;

      debug_log(DEBUG_LEVEL_INFO,
                "ELEMENT STATUS\n");

      while (count--) {
         BOOL full = FALSE;
         BOOL display_fullness = FALSE;

         switch (current->type) {
            case CHG_ELT_TYPE_STORAGE:
               full = current->elt.storage.full;
               display_fullness = TRUE;
               break;
            case CHG_ELT_TYPE_IMPORT_EXPORT:
               full = current->elt.import_export.full;
               display_fullness = TRUE;
               break;
            case CHG_ELT_TYPE_DATA_TRANSFER:
               full = current->elt.data_transfer.full;
               display_fullness = TRUE;
               break;
            default:
               break;
         }

         debug_log(DEBUG_LEVEL_INFO,
                   "   elt %u type %s label %s %s\n",
                   current->number,
                   chg_type_description(current->type),
                   current->user_data ? (char *) current->user_data
                                      : "<NONE>",
                   display_fullness ? (full ? "full" : "empty") : "");

         current++;
      }
   }
}

static void free_element_status(changer_info_t *info) {
   if (info->elements) {
      if (info->elements_count) {
	 chg_elt_t *current = info->elements;

	 while (info->elements_count--) {
	    if (current->user_data) {
	       free(current->user_data);
	    }

	    current++;
	 }
      }

      free(info->elements);
      info->elements = NULL;
      info->elements_count = 0;
   }
}

static changer_error_t load_slot_with_unload(changer_info_t *info,
					     const chg_elt_number_t slot,
					     const char *tape_device_name) {
   chg_elt_number_t tape;

   tape = elt_of_tape_device_name(tape_device_name);
   if (tape != CHG_ELT_NUMBER_INVALID) {
      chg_elt_t *tape_elt = elt_of_element(info, tape);

      if (tape_elt) {
         changer_error_t error;

         error = may_unload_drive(info, tape_elt);
         if (error == CHANGER_ERROR_NO_TAPE) {
            error = CHANGER_ERROR_NONE;
         }

         if (error == CHANGER_ERROR_NONE) {
	    debug_log(DEBUG_LEVEL_INFO,
		      "Load: Loading drive %d with slot %d.\n",
		      tape,
		      slot);
            error = load_slot(info, slot, tape);
         }

         return error;
      }
      else {
	 return CHANGER_ERROR_DRIVE_NOT_KNOWN;
      }
   }
   else {
      return CHANGER_ERROR_DRIVE_NOT_KNOWN;
   }
   /* NOT REACHED */
}

static changer_error_t load_slot(changer_info_t *info,
                                 const chg_elt_number_t slot,
                                 const chg_elt_number_t tape) {
   if (chg_move_medium(info->chg_context,
                       0, /* assuming arm 0 */
                       slot,
                       tape)) {
      info->current_slot = slot;
      update_data_xfer_from(info,
                            tape,
                            slot,
                            TRUE);
   
      return CHANGER_ERROR_NONE;
   }
   else {
      return CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED;
   }
}

static changer_error_t unload_slot(changer_info_t *info,
                                   const chg_elt_number_t slot,
                                   const chg_elt_number_t tape) {
   if (chg_move_medium(info->chg_context,
                       0, /* assuming arm 0 */
                       tape,
                       slot)) {
      info->current_slot = slot;
   
      update_data_xfer_from(info,
                            tape,
                            slot,
                            FALSE);
   
      return CHANGER_ERROR_NONE;
   }
   else {
      return CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED;
   }
}

static chg_elt_number_t elt_of_tape_device_name(const char *tape_device_name) {
   /* BUGS
    *    - Should be based on a user-supplied mapping of Linux tape device
    *      names and element number, on a per changer basis. We could use
    *      the information in the element structure (SCSI ID), but it
    *      may not work when you have different SCSI buses, and Linux way
    *      of accessing devices is not mapped easily.
    */

   if (!strcmp(tape_device_name, "/dev/nst0")) {
      return 1;
   }
   else {
      return CHG_ELT_NUMBER_INVALID;
   }
}

static
chg_elt_number_t find_next_element(changer_info_t *info,
                                   const chg_elt_number_t current,
                                   BOOL full) {
   chg_elt_t *elt = info->elements;
   unsigned int elements = info->elements_count;
   BOOL found_one = FALSE;

   while (elements) {
      if ((elt->type == CHG_ELT_TYPE_STORAGE)
          && ((full && elt->elt.storage.full)
              || (!full && !elt->elt.storage.full))) {
         if (current == CHG_ELT_NUMBER_INVALID) {
            return elt->number;
         }

         if (elt->number > current) {
            return elt->number;
         }

         found_one = TRUE;
      }

      elt++;
      elements--;
   }

   if (found_one) {
      /* We haven't found it AFTER the named element, try before. This
       * is based on the fact that we hope that elements, within a type,
       * are ordered.
       */

      elt = info->elements;
      elements = info->elements_count;

      while (elements && (elt->number < current)) {
	 if ((elt->type == CHG_ELT_TYPE_STORAGE)
             && ((full && elt->elt.storage.full)
                 || (!full && !elt->elt.storage.full))) {
	    return elt->number;
	 }

	 elt++;
	 elements--;
      }
   }

   return CHG_ELT_NUMBER_INVALID;
}

static void update_data_xfer_from(changer_info_t *info,
				  const chg_elt_number_t tape,
				  const chg_elt_number_t slot,
				  const BOOL inserts) {
   chg_elt_t *tape_elt = elt_of_element(info, tape);
   chg_elt_t *slot_elt = elt_of_element(info, slot);

   if (tape_elt) {
      tape_elt->elt.data_transfer.full = inserts;

      if (inserts) {
	 tape_elt->elt.data_transfer.from = slot;
      }
      else {
	 tape_elt->elt.data_transfer.from = CHG_ELT_NUMBER_INVALID;
      }
   }

   if (slot_elt) {
      slot_elt->elt.storage.full = !inserts;
   }

   if (slot_elt && tape_elt) {
      if (inserts) {
         if (tape_elt->user_data) {
            debug_log(DEBUG_LEVEL_WARN, "tape_elt->user_data NOT NULL.\n");
         }
         tape_elt->user_data = slot_elt->user_data;
         slot_elt->user_data = NULL;
      }
      else {
         if (slot_elt->user_data) {
            debug_log(DEBUG_LEVEL_WARN, "slot_elt->user_data NOT NULL.\n");
         }
         slot_elt->user_data = tape_elt->user_data;
         tape_elt->user_data = NULL;
      }
   }
   else {
      if (tape_elt) {
         debug_log(DEBUG_LEVEL_WARN, "slot elt NULL.\n");
      }
      else {
         debug_log(DEBUG_LEVEL_WARN, "tape elt NULL.\n");
      }
   }
}

static chg_elt_t *elt_of_element(changer_info_t *info,
                                 const chg_elt_number_t slot) {
   chg_elt_t *current = info->elements;
   unsigned int elements = info->elements_count;

   while (elements--) {
      if (current->number == slot) {
         return current;
      }

      current++;
   }

   return NULL;
}

static BOOL foreach_elt(changer_info_t *info, 
                        const chg_elt_type_t type,
                        BOOL actor(changer_info_t *info,
                                   chg_elt_t *elt)) {
   chg_elt_t *current = info->elements;
   unsigned int elements = info->elements_count;
   BOOL result = TRUE;

   while (elements-- && result) {
      if (current->type == type) {
         result = actor(info, current);
      }

      current++;
   }

   return result;
}

static BOOL empty_drive_if_full(changer_info_t *info,
                                chg_elt_t *elt) {
   BOOL result = TRUE;

   if (elt->elt.data_transfer.full) {
      chg_elt_number_t from = elt->elt.data_transfer.from;

      if (from == CHG_ELT_NUMBER_INVALID) {
	 /* Try to find the first empty slot */
	 from = find_next_element(info, from, FALSE);
      }

      if (from == CHG_ELT_NUMBER_INVALID) {
         debug_log(DEBUG_LEVEL_ERROR,
                   "Can't unload drive %d; no empty slot.\n",
                   elt->number);
         result = FALSE;
      }
      else {
         changer_error_t error;

         debug_log(DEBUG_LEVEL_INFO,
                   "Init: Unloading drive %d to slot %d.\n",
                   elt->number,
                   from);
	 error = unload_slot(info,
			     from,
			     elt->number);     
	 if (error != CHANGER_ERROR_NONE) {
	    result = FALSE;
	    debug_log(DEBUG_LEVEL_ERROR,
		      "%d: Can't unload drive %d to %d at bootup: %s.\n",
		      error,
		      elt->number,
		      from,
		      changer_error_description
			 (error));
	 }
      }
   }

   return result;
}

static changer_error_t may_unload_drive(changer_info_t *info,
                                        chg_elt_t *tape_elt) {
   if (tape_elt) {
      if (tape_elt->elt.data_transfer.full) {
   	 debug_log(DEBUG_LEVEL_INFO,
   		   "Unloading drive %d to slot %d.\n",
   		   tape_elt->number,
   		   tape_elt->elt.data_transfer.from);
         return unload_slot(info,
                            tape_elt->elt.data_transfer.from,
                            tape_elt->number);
      }
      else {
         return CHANGER_ERROR_NO_TAPE;
      }
   }
   else {
      return CHANGER_ERROR_DRIVE_NOT_KNOWN;
   }
   /* NOT REACHED */
}

static chg_elt_t *find_tape_from_label(changer_info_t *info,
                                       const char *tape_label) {
   chg_elt_t *elt = info->elements;
   unsigned int elements = info->elements_count;

   while (elements--) {
      if (elt->user_data) {
         if (!strcmp(tape_label, elt->user_data)) {
            return elt;
         }
      }

      elt++;
   }

   return NULL;
}