/* Project filed -- Intelligent file cache and migration system
 * changer_interface.h -- Interface definitions for the changer module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    changer module. The changer modules offers high-level medium changer
 *    services.
 * NOTES
 *    - Changer operations are done by the server main task, on request
 *      by a tape drive. This ensures that once an I/O process is started
 *      on the tape, there won't be any risk of someone else unloading it.
 *      Also it means the tape process (one per tape drive) can cache
 *      status, positionning and tape label and not rewind at each I/O.
 *      This requires that the tape_device_name be only passed by matching
 *      tape, never by another one. This must be enforced by the server.
 *    - This doesn't read or write labels. It just maintains a persistant
 *      database of (slots, labels), and a non-persistant database of
 *      (labels, drives). Arrangement of the tapes in slots is done here.
 *    - If the tape process requests a tape but its label is not known,
 *      it is up to the tape process to browse through the slots,
 *      possibly also asking operator for a tape, etc.
 * BUGS
 *    - Cleaning ? Idea: changer would generate a cleaning exception
 *      when required, and scheduler will not give another job to
 *      the tape process, but instead tell it to clean when finished, which
 *      would communicate with the changer.
 *    - A bit to application specific.
 *    - Mixup of slot and elt. Initially we didn't want the elt number to
 *      go out. We require it for the labelling. So, maybe would be
 *      better to let got only a slot number from 1 to N, 0 being
 *      invalid ?  With data type, etc. And define precizely the `terms
 *      of use'.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: changer_interface.h,v 1.7 2000/07/16 09:29:19 schaefer Exp $
 */

#ifndef _CHANGER_INTERFACE_H
#define _CHANGER_INTERFACE_H

/* Type definitions */

typedef enum {
   CHANGER_ERROR_NONE,
   CHANGER_ERROR_LABEL_NOT_KNOWN,
   CHANGER_ERROR_DUPLICATE_LABEL,
   CHANGER_ERROR_NO_TAPE,
   CHANGER_ERROR_DRIVE_NOT_KNOWN,
   CHANGER_ERROR_MEDIUM_BUSY,
   CHANGER_ERROR_FAILED_INTERNAL,
   CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED
} changer_error_t;

typedef void * changer_context_t;

/* Function definitions */

/* NAME
 *    changer_initialize
 * DESCRIPTION
 *    Initializes a changer module instance. Passed is the changer_name,
 *    which describes a changer device name.
 * RESULT
 *    changer_context_t, NULL if initialization failed.
 * NOTES
 *    - Please refer to ../../lib/chg/chg_interface.h for restrictions
 *      on changer_name.
 *    - This includes unloading any full tape drives to appropriate or
 *      best slot.
 * BUGS
 * TODO
 */
changer_context_t changer_initialize(const char *changer_name);

/* NAME
 *    changer_deinitialize
 * DESCRIPTION
 *    Deinitializes a changer module instance.
 * RESULT
 *    NONE
 * NOTES
 *    - context will be set to NULL
 * BUGS
 * TODO
 */
void changer_deinitialize(changer_context_t *context);

/* NAME
 *    changer_error_description
 * DESCRIPTION
 *    Returns an error description for the given changer error.
 * RESULT
 *    const char *
 * NOTES
 * BUGS
 * TODO
 */
const char *changer_error_description(const changer_error_t error);

/* NAME
 *    changer_load_tape
 * DESCRIPTION
 *    Loads the specified tape into the specified drive. This will also
 *    unload a present tape as required.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE  The slot we think contains that tape was loaded.
 *                           The tape process should open the tape device,
 *                           and verify the label. The tape device should
 *                           remain open during the whole operation. If the
 *                           label doesn't match, changer_tell_label()
 *                           should be called.
 *       CHANGER_ERROR_LABEL_NOT_KNOWN
 *                           The requested label is not available in the
 *                           database. The tape process should do a complete
 *                           slot scan, or import the tape, or create
 *                           a tape with that label.
 *       CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 *                           A hardware error occured.
 *       CHANGER_ERROR_MEDIUM_BUSY
 *                           The medium is in use in another drive.
 *       CHANGER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive in this changer.
 * NOTES
 * BUGS
 * TODO
 */
changer_error_t changer_load_tape(changer_context_t context,
				  const char *tape_label,
				  const char *tape_device_name);

/* NAME
 *    changer_unload_tape
 * DESCRIPTION
 *    Forces the tape to be removed from the drive.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE  Operation was successfully executed.
 *       CHANGER_ERROR_NO_TAPE
 *                           There was no tape in the drive.
 *       CHANGER_ERROR_OPERATION_INTERVENTION_REQUIRED
 *                           A hardware error occured.
 *       CHANGER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive in this changer.
 * NOTES
 *    - This is normally not needed since changer_load_tape() will
 *      automatically unload the current tape. One reason to call this
 *      is for example to ensure it's not in the drive for safety
 *      reasons, e.g.
 * BUGS
 * TODO
 */
changer_error_t changer_unload_tape(changer_context_t context,
                                    const char *tape_device_name);

/* NAME
 *    changer_next_tape
 * DESCRIPTION
 *    The primary function of this is to load the next tape (or the
 *    first, if first is set to TRUE) in the specified drive.
 *    There are two pointers whose variable will be used in the
 *    following circumstances, if they are non NULL:
 *       *slot_stop_at
 *          - will be set if first is TRUE.
 *          - will be used if first is FALSE and we will return
 *            CHANGER_ERROR_NO_TAPE when the next stop would be slot_stop_at.
 *       *slot_identification
 *          - will be set to the loaded slot number.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE  Operation was successfully executed.
 *       CHANGER_ERROR_NO_TAPE Current slot invalid (missing first), or
 *                             no more full slot.
 *       CHANGER_ERROR_OPERATION_INTERVENTION_REQUIRED
 *                           A hardware error occured.
 *       CHANGER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive in this changer.
 * NOTES
 *    - This means that if you want to do a full magazine scan, starting
 *      at the current slot, you just need to call
 *         changer_next_tape(context, tape_device_name, FALSE, &slot, NULL);
 *      check the result (should be CHANGER_ERROR_NONE to proceed), then
 *      do a loop of
 *         changer_next_tape(context, tape_device_name, FALSE, NULL, &slot);
 *      until the we return CHANGER_ERROR_NO_TAPE (or any other error).
 *    - slot_identification is an information that is just for referencing
 *      the starting point, and for debugging purposes and should not be
 *      stored by calling program. Same for slot_stop_at.
 * BUGS
 * TODO
 */
changer_error_t changer_next_tape(changer_context_t context,
                                  const char *tape_device_name,
                                  const BOOL first,
                                  unsigned int *slot_identification,
                                  unsigned int *slot_stop_at);

/* NAME
 *    changer_load_specific
 * DESCRIPTION
 *    Loads the specific slot. That slot value must come from the
 *    end-user. The sole utility of this function is to label
 *    a specific medium. If the tape is full, it will automatically be
 *    unloaded.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE  Operation was successfully executed.
 *       CHANGER_ERROR_NO_TAPE That slot is not valid or is empty.
 *       CHANGER_ERROR_OPERATION_INTERVENTION_REQUIRED
 *                           A hardware error occured.
 *       CHANGER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive in this changer.
 * NOTES
 * BUGS
 *    - CHANGER_ERROR_NO_TAPE not yet implemented, will cause an error.
 * TODO
 */
changer_error_t changer_load_specific(changer_context_t context,
                                      const char *tape_device_name,
                                      const unsigned int slot);

/* NAME
 *    changer_tell_label
 * DESCRIPTION
 *    Updates the database about the label of the tape in the drive.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE  Operation was successfully executed.
 *       CHANGER_ERROR_FAILED_INTERNAL
 *                           Internal error.
 *       CHANGER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive.
 *       CHANGER_ERROR_DUPLICATE_LABEL
 *                           A tape with the same label already exists.
 * NOTES
 * BUGS
 *    - Duplicate test not yet implemented.
 *    - Shouldn't fail if tell for something we know already (same data).
 * TODO
 */
changer_error_t changer_tell_label(changer_context_t context,
                                   const char *tape_label,
                                   const char *tape_device_name);

/* NAME
 *    changer_dump_element_status
 * DESCRIPTION
 *    Dumps the content of the element status, including
 *    our data, for debugging purposes.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
void changer_dump_element_status(changer_context_t context);

#endif /* !_CHANGER_INTERFACE_H */
