/* Project filed -- Intelligent file cache and migration system
 * db.c -- Simple file/part handling functions
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the database functions to store, retrieve and
 *    do operations on the database objects. The idea behind this module
 *    is to abstract the high level operations required by the scheduler and
 *    make them database-type independant.
 * NOTES
 *    - To avoid the conversion functions, we could use BINARY CURSORs.
 *      However, I prefer for now to use the ASCII interface which is more
 *      portable.
 * BUGS
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: db.c,v 1.6 2001/02/04 19:57:16 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "global.h"
#include "debug_interface.h"
#include "libpq-fe.h"
#include "md5_interface.h"
#include "db_interface.h"

/* Private type definitions */

/* Private function definitions */

/* NAME
 *    configure_connection
 * DESCRIPTION
 *    Configure the backend, such as DATESTYLE.
 * RESULT
 *    BOOL TRUE if ok, else error.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL configure_connection(PGconn *conn);

/* NAME
 *    query_uint_value
 * DESCRIPTION
 *    Executes the query which is given in query, on the
 *    database connected to by conn. Convert the single
 *    result to *sequence_value (if everything is ok).
 * RESULT
 *    BOOL TRUE if ok
 * NOTES
 *    - The primary reason this function was written is to
 *      be able to increment atomically sequences so that
 *      the ID for SERIAL tuples can be determined efficiently.
 * BUGS
 * TODO
 */
static BOOL query_uint_value(PGconn *conn,
                             const char *query,
                             unsigned int *sequence_value,
                             db_error_t *error);

/* NAME
 *    query_uint_values
 * DESCRIPTION
 *    Executes the query which is given in query, on the
 *    database connected to by conn. Convert all the available
 *    results to the array. Sets the actual number of used array
 *    cells.
 *    The values are expected to be N row of `horizontal' columns,
 *    stored rows first.
 * RESULT
 *    BOOL TRUE if ok.
 * NOTES
 * BUGS
 * TODO
 */
static BOOL query_uint_values(PGconn *conn,
                              const char *query,
                              const unsigned int horizontal,
                              unsigned int *value_array,
                              unsigned int *value_array_count,
                              db_error_t *error);

/* NAME
 *    unsafe_character
 * DESCRIPTION
 *    Returns TRUE if the given character is considered unsafe
 *    by the definitions of the db_safe_attribute() function.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
static BOOL unsafe_character(const char c);

/* NAME
 *    insert_or_update_tuple
 * DESCRIPTION
 *    Executes a tuple insertion or update SQL query.
 * RESULT
 *    BOOL TRUE if no error,
 *              else see *error:
 *                 DB_ERROR_OBJECT_EXISTS    Object already exists
 * NOTES
 * BUGS
 * TODO
 */
static BOOL insert_or_update_tuple(PGconn *conn,
                                   const BOOL insert,
				   const char *query,
				   db_error_t *error);

/* Function implementation */

db_context_t db_initialize(const db_init_t db_init) {
   PGconn *conn;

   /* make a connection to the database */
   conn = PQsetdb(db_init.host,
                  db_init.port,
                  db_init.options,
                  db_init.tty,
                  db_init.db_name);

   /* check to see that the backend connection was successfully made */
   if (PQstatus(conn) == CONNECTION_BAD) {
     debug_log(DEBUG_LEVEL_FATAL,
               "connection to database %s failed: %s",
               db_init.db_name,
               PQerrorMessage(conn));
   }
   else {
      if (configure_connection(conn)) {
         return (db_context_t) conn;
      }
      else {
         debug_log(DEBUG_LEVEL_FATAL, "Can't configure connection.\n");
      }
   }

   return NULL;
}

void db_deinitialize(db_context_t *context) {
   PQfinish((PGconn *) *context);
   *context = NULL;
}

#define BUFFER_LENGTH 1024 /* Is boundary-checked */
BOOL db_create_file_tuple(db_context_t context,
                          const char *file_name,
                          const char *client_name,
                          const char *server_name,
                          unsigned int *file_id,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   /* NOTES
    *    - We need to set the *file_id. This value is assigned by the
    *      database. We could use the OID of the created tuple
    *      (through PQoidValue()) to do a SELECT on the ID, in something
    *      similar to:
    *
    *         INSERT INTO some_table(name, surname) VALUES('marc', 'moua');
    *         -> creates OID 37492374
    *         SELECT id FROM some_table WHERE oid = 37492374;
    *
    *      This was dismissed as unefficient and complicated, notably
    *      by Tom Lane <tgl@sss.pgh.pa.us>, in Message-ID:
    *         <22003.978627865@sss.pgh.pa.us>.
    *      (on the psql-general mailing-list).
    *
    *      He gave the correct solution:
    *         SELECT nextval('some_table_id_sequence');
    *         INSERT INTO some_table(id, name, surname)
    *            VALUES(currval('some_table_id_sequence'), 'marc', 'moua');
    *      A transaction is not required, because currval is backend
    *      private. Alternatively one can use the value returned by the
    *      SELECT and feed it back into the INSERT, sparing a currval().
    *      His words were: ``Simple, reliable, fast.''.
    *
    *    - Additionnally the hack for the duplicate check was proposed by
    *      Peter Eisentraut <peter_e@gmx.net>, in Message-ID:
    *         <Pine.LNX.4.30.0101041706440.1245-100000@peter.localdomain>
    *      (on the psql-general mailing-list).
    */

   if (query_uint_value(conn,
                        "SELECT nextval('file_id_seq')",
                        file_id,
                        error)) {
      *error = DB_ERROR_INTERNAL;

      if (snprintf(buffer,
		   sizeof(buffer),
		   "INSERT INTO file (id, file_name, client_name, server_name)"
		   " VALUES(%u, '%s', '%s', '%s')",
		   *file_id,
		   file_name,
		   client_name,
		   server_name)
	  == -1) {
	 *error = DB_ERROR_WOULD_OVERFLOW;
      }
      else {
         result = insert_or_update_tuple(conn, TRUE, buffer, error);
      }
   }
   /* else error already set */

   return result;
}

BOOL db_lookup_file_tuple(db_context_t context,
                          const char *file_name,
                          const char *client_name,
                          const char *server_name,
                          unsigned int *file_id,
                          unsigned int *part_count,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];

   if (snprintf(buffer,
                sizeof(buffer),
                "SELECT id, part_count FROM file"
                " WHERE (file_name = '%s')"
                " AND (client_name = '%s')"
                " AND (server_name = '%s')", /* distinct */
                file_name,
                client_name,
                server_name)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;

      return FALSE;
   }
   else {
      unsigned int array[2];
      unsigned int array_count = 2;

      if (query_uint_values((PGconn *) context,
                            buffer,
                            2,
                            array,
                            &array_count,
                            error)) {
         if (array_count == 2) {
	    *file_id = array[0];
	    *part_count = array[1];
	    return TRUE;
         }
         else {
            return FALSE;
         }
      }
      else {
         return FALSE;
      }
   }
   /* NOT REACHED */
}

BOOL db_update_file_tuple(db_context_t context,
                          const unsigned int file_id,
                          const BOOL complete,
                          const unsigned int part_count,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   *error = DB_ERROR_INTERNAL;
   
   if (snprintf(buffer,
                sizeof(buffer),
                "UPDATE file"
                " SET complete = '%c', part_count = %u"
                " WHERE id = %u",
                complete ? 't'
                         : 'f',
                part_count,
                file_id)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
   }
   else {
      result = insert_or_update_tuple(conn, FALSE, buffer, error);
   }
   /* else error already set */

   return result;
}

BOOL db_create_part_tuple(db_context_t context,
                          const unsigned int file_id,
                          const unsigned int part_number,
                          unsigned int *part_id,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   if (query_uint_value(conn,
                        "SELECT nextval('part_id_seq')",
                        part_id,
                        error)) {
      *error = DB_ERROR_INTERNAL;

      if (snprintf(buffer,
		   sizeof(buffer),
		   "INSERT INTO part (id, file_id, part_number)"
		   " VALUES(%u, %u, %u)",
		   *part_id,
		   file_id,
		   part_number)
	  == -1) {
	 *error = DB_ERROR_WOULD_OVERFLOW;
      }
      else {
         result = insert_or_update_tuple(conn, TRUE, buffer, error);
      }
   }
   /* else error already set */

   return result;
}

BOOL db_lookup_part_tuple(db_context_t context,
                          const unsigned int file_id,
                          const unsigned int part_number,
                          unsigned int *part_id,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];

   if (snprintf(buffer,
                sizeof(buffer),
                "SELECT id FROM part"
                " WHERE (file_id = %u)"
                " AND (part_number = %u)", /* distinct */
                file_id,
                part_number)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
      return FALSE;
   }
   else {
      return query_uint_value((PGconn *) context,
			      buffer,
			      part_id,
			      error);
   }
   /* NOT REACHED */
}

BOOL db_update_part_tuple(db_context_t context,
                          const unsigned int part_id,
                          const BOOL complete,
                          const unsigned char *md5sum,
                          const unsigned int size,
                          db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;
   char textual_md5sum[MD5_HASH_TEXTUAL_STRING_LENGTH];

   md5_hash_to_textual_md5sum((const md5_hash_ptr_t) md5sum, textual_md5sum);
   debug_log(DEBUG_LEVEL_INFO, "md5sum: %s.\n", textual_md5sum);

   *error = DB_ERROR_INTERNAL;
   
   if (snprintf(buffer,
                sizeof(buffer),
                "UPDATE part"
                " SET complete = '%c', md5sum = '%s', size = %u"
                " WHERE id = %u",
                complete ? 't'
                         : 'f',
                textual_md5sum,
                size,
                part_id)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
   }
   else {
      result = insert_or_update_tuple(conn, FALSE, buffer, error);
   }
   /* else error already set */

   return result;
}

BOOL db_create_part_spooled_tuple(db_context_t context,
				  const unsigned int part_id,
                                  const char *file_name,
				  unsigned int *part_spooled_id,
				  db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   if (query_uint_value(conn,
                        "SELECT nextval('part_spooled_id_seq')",
                        part_spooled_id,
                        error)) {
      *error = DB_ERROR_INTERNAL;

      if (snprintf(buffer,
		   sizeof(buffer),
		   "INSERT INTO part_spooled (id, part_id, file_name)"
		   " VALUES(%u, %u, '%s')",
		   *part_spooled_id,
		   part_id,
		   file_name)
	  == -1) {
	 *error = DB_ERROR_WOULD_OVERFLOW;
      }
      else {
         result = insert_or_update_tuple(conn, TRUE, buffer, error);
      }
   }
   /* else error already set */

   return result;
}

BOOL db_lookup_part_spooled_tuple(db_context_t context,
				  const unsigned int part_id,
				  unsigned int *part_spooled_id,
				  db_error_t *error) {
   char buffer[BUFFER_LENGTH];

   if (snprintf(buffer,
                sizeof(buffer),
                "SELECT id FROM part_spooled"
                " WHERE (part_id = %u)", /* distinct */
                part_id)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
      return FALSE;
   }
   else {
      return query_uint_value((PGconn *) context,
			      buffer,
			      part_spooled_id,
			      error);
   }
   /* NOT REACHED */
}

BOOL db_update_part_spooled_tuple(db_context_t context,
                                  const unsigned int part_spooled_id,
                                  const BOOL complete,
                                  db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   *error = DB_ERROR_INTERNAL;
   
   if (snprintf(buffer,
                sizeof(buffer),
                "UPDATE part_spooled"
                " SET complete = '%c'"
                " WHERE id = %u",
                complete ? 't'
                         : 'f',
                part_spooled_id)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
   }
   else {
      result = insert_or_update_tuple(conn, FALSE, buffer, error);
   }
   /* else error already set */

   return result;
}

BOOL db_create_part_stored_tuple(db_context_t context,
				 const unsigned int part_id,
				 unsigned int *part_stored_id,
				 db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   if (query_uint_value(conn,
                        "SELECT nextval('part_stored_id_seq')",
                        part_stored_id,
                        error)) {
      *error = DB_ERROR_INTERNAL;

      if (snprintf(buffer,
		   sizeof(buffer),
		   "INSERT INTO part_stored (id, part_id)"
		   " VALUES(%u, %u)",
		   *part_stored_id,
		   part_id)
	  == -1) {
	 *error = DB_ERROR_WOULD_OVERFLOW;
      }
      else {
         result = insert_or_update_tuple(conn, TRUE, buffer, error);
      }
   }
   /* else error already set */

   return result;
}

BOOL db_lookup_part_stored_tuple(db_context_t context,
				 const unsigned int part_id,
                                 unsigned int *part_stored_id_array_count,
				 unsigned int *part_stored_id_array,
				 db_error_t *error) {
   char buffer[BUFFER_LENGTH];

   if (snprintf(buffer,
                sizeof(buffer),
                "SELECT id FROM part_stored"
                " WHERE (part_id = %u)"
                " LIMIT %u", /* NOT distinct */
                part_id,
                *part_stored_id_array_count)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
      return FALSE;
   }
   else {
      return query_uint_values((PGconn *) context,
			       buffer,
                               1,
			       part_stored_id_array,
                               part_stored_id_array_count,
			       error);
   }
   /* NOT REACHED */
}

BOOL db_update_part_stored_tuple(db_context_t context,
                                 const unsigned int part_stored_id,
                                 const BOOL complete,
                                 const unsigned int medium_id,
                                 const unsigned int locator,
                                 db_error_t *error) {
   char buffer[BUFFER_LENGTH];
   BOOL result = FALSE; /* failed */
   PGconn *conn = (PGconn *) context;

   *error = DB_ERROR_INTERNAL;
   
   if (snprintf(buffer,
                sizeof(buffer),
                "UPDATE part_stored"
                " SET complete = '%c', medium_id = %u, locator = %u"
                " WHERE id = %u",
                complete ? 't'
                         : 'f',
                medium_id,
                locator,
                part_stored_id)
       == -1) {
      *error = DB_ERROR_WOULD_OVERFLOW;
   }
   else {
      result = insert_or_update_tuple(conn, FALSE, buffer, error);
   }
   /* else error already set */

   return result;
}

BOOL db_safe_attribute(char *attribute) {
   if (*attribute) {
      while (*attribute) {
         if (unsafe_character(*attribute)) {
            return FALSE;
         }
         attribute++;
      }

      return TRUE;
   }

   return FALSE; /* empty string */
}

/* Private function implementation */

static BOOL configure_connection(PGconn *conn) {
   PGresult *res;
   
   res = PQexec(conn, "SET DATESTYLE = 'ISO'");
   if (!(res && (PQresultStatus(res) == PGRES_COMMAND_OK))) {
      debug_log(DEBUG_LEVEL_FATAL,
                "SET DATESTYLE query failed: %s",
                PQerrorMessage(conn));
      return FALSE;
   }
   
   PQclear(res); /* avoid memory leaks */

   return TRUE;
}

static BOOL query_uint_value(PGconn *conn,
                             const char *query,
                             unsigned int *sequence_value,
                             db_error_t *error) {
   PGresult *res;
   BOOL result = FALSE; /* failed */

   *error = DB_ERROR_INTERNAL;
   
   /* BUGS
    *    - Should protect or bind like in Perl.
    *    - Who protects against ', etc ?
    */

   fprintf(stderr, "UINT QUERY: %s.\n", query);
   
   res = PQexec(conn, query);
   if (res) {
      if (PQresultStatus(res) == PGRES_TUPLES_OK) {
         if (PQntuples(res) == 1) {
            unsigned long int temp;
            char *endptr;

            temp = strtoul(PQgetvalue(res, 0, 0),
                           &endptr,
                           10);

            if ((PQgetvalue(res, 0, 0)[0] != '\0') && (*endptr == '\0')) {
               *sequence_value = temp;
	       result = TRUE;
               /* not really useful, shouldn't be read */
	       *error = DB_ERROR_NONE;
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR,
                         "Value unparsable: '%s'.\n",
                         PQgetvalue(res, 0, 0));
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Expected one tuple, got %d.\n",
                      PQntuples(res));
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "SELECT query failed: %s",
                   PQerrorMessage(conn));
      }
   
      PQclear(res); /* avoid memory leaks */
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "SELECT query failed (no result): %s",
                PQerrorMessage(conn));
   }

   return result;
}

static BOOL query_uint_values(PGconn *conn,
                              const char *query,
                              const unsigned int horizontal,
                              unsigned int *value_array,
                              unsigned int *value_array_count,
                              db_error_t *error) {
   PGresult *res;
   BOOL result = FALSE; /* failed */

   *error = DB_ERROR_INTERNAL;
   
   /* BUGS
    *    - Should protect or bind like in Perl.
    *    - Who protects against ', etc ?
    */

   fprintf(stderr, "UINTS QUERY: %s.\n", query);
   
   res = PQexec(conn, query);
   if (res) {
      if (PQresultStatus(res) == PGRES_TUPLES_OK) {
         unsigned int count = MIN(PQntuples(res) * PQnfields(res),
                                  *value_array_count);
         unsigned int i = 0;
         unsigned int column = 0;
         unsigned int row = 0;

         result = TRUE;         
         while ((i < count) && result) {
            unsigned long int temp;
            char *endptr;

            temp = strtoul(PQgetvalue(res, row, column),
                           &endptr,
                           10);

            if ((PQgetvalue(res, row, column)[0] != '\0')
                && (*endptr == '\0')) {
               value_array[i] = temp;
               /* not really useful, shouldn't be read */
	       *error = DB_ERROR_NONE;
            }
            else {
	       result = FALSE;
               debug_log(DEBUG_LEVEL_ERROR,
                         "Value unparsable: '%s'.\n",
                         PQgetvalue(res, row, column));
            }

            column++;
            if (column >= PQnfields(res)) {
               column = 0;
               row++;
            }

            i++;
         }

         if (result) {
            *value_array_count = i;
         }
         else {
            result = FALSE;
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "SELECT query failed: %s",
                   PQerrorMessage(conn));
         *value_array_count = 0;
      }
   
      PQclear(res); /* avoid memory leaks */
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "SELECT query failed (no result): %s",
                PQerrorMessage(conn));
      *value_array_count = 0;
   }

   return result;
}

static BOOL unsafe_character(const char c) {
   /* NOTES
    *    - We play safe. Using isXXXX() functions has problems
    *      depending on the locale.
    */
   if ((c >= 'a') && (c <= 'z')) {
      return FALSE; /* safe: [a-z] */
   }

   if ((c >= 'A') && (c <= 'Z')) {
      return FALSE; /* safe: [A-Z] */
   }

   if ((c >= '0') && (c <= '9')) {
      return FALSE; /* safe: [0-9] */
   }

   if ((c == '-') || (c == '_') || (c == '.')) {
      return FALSE; /* safe: [\-_\.] */
   }

   return TRUE; /* unsafe */
}

static BOOL insert_or_update_tuple(PGconn *conn,
                                   const BOOL insert,
				   const char *query,
				   db_error_t *error) {
   PGresult *res;
   BOOL result = FALSE;
   char *what = "UPDATE";

   if (insert) {
      what = "INSERT";
   }

   *error = DB_ERROR_INTERNAL;
   
   /* BUGS
    *    - Should protect or bind like in Perl.
    *    - Who protects against ', etc ?
    */

   fprintf(stderr, "QUERY: %s.\n", query);
   
   res = PQexec(conn, query);
   if (res) {
      if (PQresultStatus(res) == PGRES_COMMAND_OK) {
         result = TRUE;
         /* not really useful, shouldn't be read */
         *error = DB_ERROR_NONE;
      }
      else {
         /* BUGS
          *    - This is about the best you can do.
          */
         if (insert
             && !strcmp(PQerrorMessage(conn),
                        "ERROR:  Cannot insert a duplicate key into a unique index\n")) {
            *error = DB_ERROR_OBJECT_EXISTS;
            debug_log(DEBUG_LEVEL_ERROR,
                      "%s query failed because duplicate.\n",
                      what);
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "%s query failed: %s",
                      what,
                      PQerrorMessage(conn));
         }            
      }
   
      PQclear(res); /* avoid memory leaks */
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "%s query failed (no result): %s",
                what,
                PQerrorMessage(conn));
   }

   return result;
}
