/* Project filed -- Intelligent file cache and migration system
 * db_interface.h -- Interface definitions for the db module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Mon Jan 1 2001
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    db module. The db modules offers simple database services for
 *    handling file and part objects.
 * NOTES
 * BUGS
 *    - Missing delete functions for file, part, part_spooled, part_stored.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: db_interface.h,v 1.6 2001/02/04 19:57:16 schaefer Exp $
 */

#ifndef _DB_INTERFACE_H
#define _DB_INTERFACE_H

/* Type definitions */

typedef void * db_context_t;

typedef struct {
   char *host; /* if NULL uses default */
   char *port; /* if NULL uses default */
   char *options; /* if NULL uses default */
   char *tty; /* if NULL uses default */

   char *db_name; /* cannot be NULL */
} db_init_t;

typedef enum {
   DB_ERROR_NONE,
   DB_ERROR_OBJECT_EXISTS,
   DB_ERROR_OBJECT_DOES_NOT_EXISTS,
   DB_ERROR_WOULD_OVERFLOW,
   DB_ERROR_INTERNAL
} db_error_t;

/* Function definitions */

/* NAME
 *    db_initialize
 * DESCRIPTION
 *    Sets up initialization for the db module. Returns a db
 *    context.
 * RESULT
 *    db_context_t, NULL if initialization failed.
 * NOTES
 *    - The db module lifecycle is:
 *         db_initialize()
 *            ... any of the db control functions (repeated), except
 *                db_initialize().
 *         db_deinitialize()
 *    - This involves connecting to the database and basic initialization.
 * BUGS
 * TODO
 */
db_context_t db_initialize(const db_init_t db_init);

/* NAME
 *    db_deinitialize
 * DESCRIPTION
 *    Deinitializes the db module.
 * RESULT
 *    NONE
 * NOTES
 *    - *context is set to NULL.
 * BUGS
 * TODO
 */
void db_deinitialize(db_context_t *context);

/* NAME
 *    db_create_file_tuple
 * DESCRIPTION
 *    Creates a file tuple in the database, setting *file_id to its ID
 *    if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 *    - Who does the file_name and friends filtering ?
 * TODO
 */
BOOL db_create_file_tuple(db_context_t context,
                          const char *file_name,
                          const char *client_name,
                          const char *server_name,
                          unsigned int *file_id,
                          db_error_t *error);

/* NAME
 *    db_lookup_file_tuple
 * DESCRIPTION
 *    Look ups a file tuple in the database, setting *file_id to its ID
 *    if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 *    - Who does the file_name and friends filtering ?
 * TODO
 */
BOOL db_lookup_file_tuple(db_context_t context,
                          const char *file_name,
                          const char *client_name,
                          const char *server_name,
                          unsigned int *file_id,
                          unsigned int *part_count,
                          db_error_t *error);

/* NAME
 *    db_update_file_tuple
 * DESCRIPTION
 *    Update the file information in the database.
 * RESULT
 *    BOOL TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 *    - Who does the file_name and friends filtering ?
 * TODO
 */
BOOL db_update_file_tuple(db_context_t context,
                          const unsigned int file_id,
                          const BOOL complete,
                          const unsigned int part_count,
                          db_error_t *error);

/* NAME
 *    db_create_part_tuple
 * DESCRIPTION
 *    Creates a part tuple in the database, setting *part_id to its ID
 *    if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 *    - This creates an uncompleted part. See db_update_part_tuple().
 * BUGS
 * TODO
 */
BOOL db_create_part_tuple(db_context_t context,
                          const unsigned int file_id,
                          const unsigned int part_number,
                          unsigned int *part_id,
                          db_error_t *error);

/* NAME
 *    db_lookup_part_tuple
 * DESCRIPTION
 *    Look ups a part tuple in the database, setting *part_id to its ID
 *    if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 * TODO
 */
BOOL db_lookup_part_tuple(db_context_t context,
                          const unsigned int file_id,
                          const unsigned int part_number,
                          unsigned int *part_id,
                          db_error_t *error);

/* NAME
 *    db_update_part_tuple
 * DESCRIPTION
 *    Update the part information in the database.
 * RESULT
 *    BOOL TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 *    - md5sum must be 16 bytes long (unsigned char).
 * BUGS
 * TODO
 */
BOOL db_update_part_tuple(db_context_t context,
                          const unsigned int part_id,
                          const BOOL complete,
                          const unsigned char *md5sum,
                          const unsigned int size,
                          db_error_t *error);

/* NAME
 *    db_create_part_spooled_tuple
 * DESCRIPTION
 *    Creates a part_spooled tuple in the database, setting
 *    *part_spooled_id to its ID if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 *    - This creates an uncompleted part_spooled. See
 *      db_update_part_spooled_tuple().
 * BUGS
 * TODO
 */
BOOL db_create_part_spooled_tuple(db_context_t context,
				  const unsigned int part_id,
                                  const char *file_name,
				  unsigned int *part_spooled_id,
				  db_error_t *error);

/* NAME
 *    db_lookup_part_spooled_tuple
 * DESCRIPTION
 *    Look ups a part_spooled tuple in the database, setting
 *    *part_spooled_id to its ID if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 *    - As explained in the doc/DATABASE, we have upto one spooled part
 *      for each part (no multiple here).
 * BUGS
 * TODO
 */
BOOL db_lookup_part_spooled_tuple(db_context_t context,
				  const unsigned int part_id,
				  unsigned int *part_spooled_id,
				  db_error_t *error);

/* NAME
 *    db_update_part_spooled_tuple
 * DESCRIPTION
 *    Update the part_spooled information in the database.
 * RESULT
 *    BOOL TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 *    - Caller responsible for making the part table/class consistent
 *      respective of complete.
 * TODO
 */
BOOL db_update_part_spooled_tuple(db_context_t context,
                                  const unsigned int part_spooled_id,
                                  const BOOL complete,
                                  db_error_t *error);

/* NAME
 *    db_create_part_stored_tuple
 * DESCRIPTION
 *    Creates a part_stored tuple in the database, setting
 *    *part_stored_id to its ID if the function succeeds.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 *    - This creates an uncompleted part_stored. See
 *      db_update_part_stored_tuple().
 * BUGS
 * TODO
 */
BOOL db_create_part_stored_tuple(db_context_t context,
				 const unsigned int part_id,
				 unsigned int *part_stored_id,
				 db_error_t *error);

/* NAME
 *    db_lookup_part_stored_tuple
 * DESCRIPTION
 *    Sets all the matching stored parts in the pointed array.
 *    Updates the count to actual afterwards.
 * RESULT
 *    BOOL, TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 * TODO
 *    - Later should be able to select by e.g. medium name, etc.
 */
BOOL db_lookup_part_stored_tuple(db_context_t context,
				 const unsigned int part_id,
                                 unsigned int *part_stored_id_array_count,
				 unsigned int *part_stored_id_array,
				 db_error_t *error);

/* NAME
 *    db_update_part_stored_tuple
 * DESCRIPTION
 *    Update the part_stored information in the database.
 * RESULT
 *    BOOL TRUE if succeeded.
 * NOTES
 *    - If didn't succeed, error will be set.
 * BUGS
 *    - Caller responsible for making the part table/class consistent
 *      respective of complete.
 * TODO
 */
BOOL db_update_part_stored_tuple(db_context_t context,
                                 const unsigned int part_stored_id,
                                 const BOOL complete,
                                 const unsigned int medium_id,
                                 const unsigned int locator,
                                 db_error_t *error);
/* NAME
 *    db_safe_attribute
 * DESCRIPTION
 *    Checks that the pointed string is a valid attribute.
 * RESULT
 *    BOOL
 * NOTES
 *    - Currently will reject any string containing characters
 *      outside of [a-zA-Z\-0-9\._], or when the string is empty.
 * BUGS
 * TODO
 *    - Infact, the db module should really encapsulate (like e.g. in
 *      Perl binding) the data, and this function should be moved to
 *      the scheduler (if specific constraints are required).
 */
BOOL db_safe_attribute(char *attribute);

#endif /* !_DB_INTERFACE_H */
