/* Project filed -- Intelligent file cache and migration system
 * scheduler.c -- I/O scheduler.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the data I/O scheduler, which decides of the
 *    ordering and size of I/O operations to/from the cache and to/from
 *    media.
 * NOTES
 * BUGS
 *    - Does it need to be so complex?
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: scheduler.c,v 1.10 2001/02/04 19:57:15 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/vfs.h>
#include <unistd.h>
#include <string.h>
#include "global.h"
#include "md5_interface.h"
#include "lists_interface.h"
#include "debug_interface.h"
#include "server_defs.h"
#include "db_interface.h"
#include "scheduler_interface.h"

/* Private type definitions */

typedef struct {
   unsigned int min_free_files;
   unsigned int min_free_blocks;

   /* For the ratios, 20 means that (free / total) > 20 */
   unsigned int min_free_files_ratio;
   unsigned int min_free_blocks_ratio;
} constraint_t;

typedef struct {
   const server_config_t *config;
   unsigned long int session_count; /* for allocate/free tracking */

   /* This is the list of the sessions that are waiting on some parts,
    * for writing or reading, be it dumper or medium.
    */
   list_head_t part_wait_list;

   running_count_t running_dumpers;

   constraint_t constraints;

   db_context_t db_context;

   BOOL spool_full;
} scheduler_info_t;

typedef struct {
   list_item_header_t list_header;

   part_stored_id_t part_stored_id; /* PART_STORED_ID_INVALID if creation
                                     * would be required.
                                     */
   age_t age;
   priority_t priority;
   locator_t locator;
   medium_id_t medium_id;
   part_id_t part_id;
   BOOL for_write;
   BOOL will_be_freed_by_writing;

   scheduler_session_t *medium_process;
} part_stored_memory_t; /* for the candidate list */

/* Macros */

#define LIST_ITEM_TO_SESSION(item) \
   LIST_ITEM_TO_OBJECT((item), scheduler_session_t)

#define LIST_ITEM_TO_PART_STORED_MEMORY(item) \
   LIST_ITEM_TO_OBJECT((item), part_stored_memory_t)

/* Private function definitions */

/* NAME
 *    get_next_schedulable_session
 * DESCRIPTION
 *    This function returns the next schedulable session, if any.
 * RESULT
 *    scheduler_session_t *, NULL if none is schedulable.
 * NOTES
 * BUGS
 * TODO
 */
static scheduler_session_t *get_next_schedulable_session
   (scheduler_info_t *info);

/* NAME
 *    ok_to_io
 * DESCRIPTION
 *    This function returns TRUE if the filesystem which contains the
 *    named directory complies with the passed constraints (when
 *    creating or not a file depending on create_file). The constraints
 *    are checked after applying a chunk_size virtual allocation.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 *    - We assume that all values are given in the filesystem block size
 *      for now.
 *    - Some border cases in the ratio are not right, think again.
 * TODO
 */
static BOOL ok_to_io(const char *directory,
                     const BOOL create_file,
                     const unsigned int chunk_size,
                     const constraint_t constraints);

/* NAME
 *    is_a_schedulable_session
 * DESCRIPTION
 *    Returns TRUE if the named session is a candidate for scheduling, at
 *    this time.
 * RESULT
 *    BOOL TRUE if schedulable
 * NOTES
 * BUGS
 * TODO
 */
static BOOL is_a_schedulable_session(scheduler_info_t *info,
                                     scheduler_session_t *session);

/* NAME
 *    initiate_spool_purge
 * DESCRIPTION
 *    Frees the spool of some unused parts that haven't been used in a while.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void initiate_spool_purge(scheduler_info_t *info);

/* NAME
 *    is_part_in_spool
 * DESCRIPTION
 *    Returns TRUE if the requested part is in spool.
 * RESULT
 *    BOOL
 * NOTES
 *    - Sets the appropriate ID's (file, part, part_spooled) in the session.
 * BUGS
 * TODO
 *    - Document this much better.
 */
static BOOL is_part_in_spool(scheduler_info_t *info,
                             scheduler_session_t *session,
                             BOOL dumper,
                             BOOL complete);

/* NAME
 *    compute_medium_part_priority
 * DESCRIPTION
 *    Computes the priority of a part that is to be stored to final
 *    medium. The priority depends on diverse factors, such as the
 *    dumper session priority (as set by the client), the fullness
 *    of the disk, the direction, the waiting time (age) of this
 *    request and the locality of the request.
 * RESULT
 *    priority_t, higher is better
 * NOTES
 *    - We maybe have to choose a priority default, positive greater than
 *      zero so that it is useful (because we substract sometimes ...)
 *      Or choose the values accordingly, else we may have a lot of
 *      zero-priority tasks.
 * BUGS
 * TODO
 *    - In RAID1, if the other part is going to be written in the same run,
 *      the second and next should have decreased priority. However, if
 *      we want to clean the spool to get the space back, it shouldn't.
 *      This should make so that if all drives are idle, parallel I/O is
 *      possible, but else, we don't consume all resources.
 */
static priority_t compute_medium_part_priority
   (scheduler_info_t *info,
    part_stored_memory_t *part_stored_memory);

/* NAME
 *    priority_substract
 * DESCRIPTION
 *    Lowers a priority with a lower bound at zero.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void priority_substract(priority_t *priority,
                               const priority_t value);

/* NAME
 *    medium_scheduler
 * DESCRIPTION
 *    Scheduler for the medium processes.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 *    - How do we talk with it ?
 *    - Should be merged (or separated) from get_next_schedulable_session().
 * TODO
 */
static void medium_scheduler(scheduler_info_t *info);

/* NAME
 *    find_parts_stored_memory
 * DESCRIPTION
 *    This function builds a prioritized list of virtual (in memory)
 *    part_stored. This means that all parts that are requested by a
 *    dumper, exist as stored parts but aren't in the spool, or
 *    all parts that are in the spool but haven't been stored yet
 *    on final storage will be returned if they are on a medium which
 *    is reachable by medium_process.
 * RESULT
 * NOTES
 *    - For writes (storing), all possibilities are returned. This might
 *      be quite large, so that's why it's done at the end. Maybe we
 *      should implement limits on this ?  Or limits on the number of
 *      open-for-writing-currently media?  Or consider meta-types, such
 *      as partly written, and none written ?
 *    - Will allocate dynamic structures that must be deallocated by caller.
 * BUGS
 * TODO
 *    - ensure we expand write properly, set all fields correctly, including
 *      locator for writes to tapes. Note that locator for write to tapes
 *      is an estimate to be used for the priority evaluation; and it
 *      should be set to LOCATOR_INVALID just after (added at end).
 *      Some medium types may not be so complex, though (CD-Rs). However
 *      medium_id must stick since it's part of the decision.
 */
static void find_parts_stored_memory(scheduler_info_t *info,
                                     list_head_t *list,
                                     scheduler_session_t *medium_process);

/* NAME
 *    medium_device_still_idle
 * DESCRIPTION
 *    Returns TRUE if the pointed medium process (device) is still idle,
 *    ie it is still on the info->part_wait_list. Else returns FALSE,
 *    since it will be busy.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
static BOOL medium_device_still_idle(scheduler_info_t *info,
                                     scheduler_session_t *medium_process);

/* NAME
 *    still_required_to_do_op
 * DESCRIPTION
 *    Returns TRUE if the pointed virtual stored part is still required
 *    to be scheduled because there isn't already a part to be scheduled
 *    fullfilling the needs for that part.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 *    - This function will be extended to support RAID1 policies, for example.
 */
static BOOL still_required_to_do_op(scheduler_info_t *info,
				    list_head_t *scheduled_list,
				    part_stored_memory_t *part_stored_memory);

/* NAME
 *    dumper_session_is_lower_priority
 * DESCRIPTION
 *    Returns TRUE if the new item has a greater priority than the current.
 * RESULT
 *    BOOL
 * NOTES
 * BUGS
 * TODO
 */
static BOOL dumper_session_is_lower_priority(list_head_t *list,
					     list_item_header_t *new,
					     list_item_header_t *current,
					     void *private);

/* Private constants */

const db_init_t db_init = { NULL, NULL, NULL, NULL, "filed" };

const constraint_t default_constraints
   = { 100, 10000, 20, 20 };

const priority_factors_t default_priority_factors
   = {   150, /* age */
       10000, /* same_medium */
        5000, /* rewind_pos */
        5000, /* could_free_space_by_writing */
        5000, /* eject */
        5000, /* contiguous_file */
         500, /* fsf */
        1000, /* bsf */
         100  /* fullness */
     };

/* Function implementation */

scheduler_context_t scheduler_initialize
   (const server_config_t *server_config) {
   scheduler_info_t *info
      = (scheduler_info_t *) malloc(sizeof(scheduler_info_t));

   if (info) {
      info->config = server_config;
      info->session_count = 0;
      info->running_dumpers = 0;
      info->spool_full = FALSE;

      memcpy(&info->constraints,
             &default_constraints,
             sizeof(default_constraints));

      list_initialize(&info->part_wait_list);

      info->db_context = db_initialize(db_init);

      if (info->db_context == NULL) {
         free(info);
         info = NULL;
      }
   }

   return (scheduler_context_t) info;
}

void scheduler_deinitialize(scheduler_context_t *scheduler_context) {
   scheduler_info_t *info = *scheduler_context;

   if (info->session_count) {
      debug_log(DEBUG_LEVEL_FATAL,
                "INTERNAL ERROR: still %ld open session at deinitialize"
                " time.\n",
                info->session_count);
   }

   if (list_next(&info->part_wait_list,
                 NULL)) {
      debug_log(DEBUG_LEVEL_FATAL,
                "INTERNAL ERROR: part wait list not empty.\n");
   }

   db_deinitialize(&info->db_context);

   free(info);
   *scheduler_context = NULL;
}

scheduler_session_t *scheduler_session_allocate
   (scheduler_context_t scheduler_context,
    scheduler_session_union_t *session_info,
    void *child,
    scheduler_session_type_t type,
    scheduler_error_t *error) {
   scheduler_session_t *session = malloc(sizeof(scheduler_session_t));

   *error = SCHEDULER_ERROR_INTERNAL;

   /* For a session of type `dumper', we know now the file name, client
    * name and direction: we can thus allocate the appropriate file in the
    * database.
    * For a session of type `medium', there is not much to do except
    * allocate the session.
    */

   if (session) {
      BOOL succeeded = TRUE;
      db_error_t db_error;
      scheduler_info_t *info = (scheduler_info_t *) scheduler_context;

      info->session_count++; /* we don't handle the 4GB - 1 case, we assume
                              * there are much less processes.
                              */
      session->type = type;
      session->part_id = PART_ID_INVALID;
      session->file_id = FILE_ID_INVALID;
      session->error = SCHEDULER_ERROR_NONE;
      session->part_count = 0;
      session->current_part = 0;
      session->child = child;
      session->direction_to_cache = TRUE; /* undetermined for medium, yet,
                                           * will be corrected for dumper.
                                           */

      switch (type) {
         case SCHEDULER_SESSION_TYPE_DUMPER:
            succeeded = FALSE;

            memcpy(&session->object.dumper,
                   &session_info->dumper,
                   sizeof(session_dumper_info_t));

            if (!db_safe_attribute(session->object.dumper.client_name)) {
               *error = SCHEDULER_ERROR_BAD_CLIENT_NAME;
            }
            else if (!db_safe_attribute(session->object.dumper.server_name)) {
               *error = SCHEDULER_ERROR_BAD_SERVER_NAME;
            }
            else if (!db_safe_attribute(session->object.dumper.file_name)) {
               *error = SCHEDULER_ERROR_BAD_FILE_NAME;
            }
            else {
               succeeded = TRUE; /* for now */

	       debug_log(DEBUG_LEVEL_GARBAGE,
			 "session 0x%x direction: 0x%x client: %s server: %s "
			 "file: %s flags: 0x%x block_size: %u blocks: %u "
			 "offset: %u.\n",
			 (unsigned int) session,
			 (unsigned int) session->object.dumper.direction,
			 session->object.dumper.client_name,
			 session->object.dumper.server_name,
			 session->object.dumper.file_name,
			 (unsigned int) session->object.dumper.flags,
			 (unsigned int) session->object.dumper.block_size,
			 (unsigned int) session->object.dumper.blocks,
			 (unsigned int) session->object.dumper.offset);

	       if (session->object.dumper.direction == DIRECTION_TO_CACHE) {
		  /* BUGS
		   *    - what level does file_name filtering ?
		   */
		  succeeded = db_create_file_tuple
				 (info->db_context,
				  session->object.dumper.file_name,
				  session->object.dumper.client_name,
				  session->object.dumper.server_name,
				  &session->file_id,
				  &db_error);
		  if (!succeeded) {
		     if (db_error == DB_ERROR_OBJECT_EXISTS) {
			*error = SCHEDULER_ERROR_DATABASE_FILE_EXISTS;
			debug_log(DEBUG_LEVEL_WARN,
				  "File already exists in database.\n");
		     }
		     else {
			*error = SCHEDULER_ERROR_DATABASE_GENERIC;
			debug_log(DEBUG_LEVEL_ERROR,
				  "Database error: 0x%x.\n",
				  db_error);
		     }
		  }
	       }
	       else if (session->object.dumper.direction
			== DIRECTION_FROM_CACHE) {
		  session->direction_to_cache = FALSE;

		  succeeded = db_lookup_file_tuple
				 (info->db_context,
				  session->object.dumper.file_name,
				  session->object.dumper.client_name,
				  session->object.dumper.server_name,
                                  &session->part_count,
				  &session->file_id,
				  &db_error);

		  if (!succeeded) {
		     if (db_error == DB_ERROR_OBJECT_DOES_NOT_EXISTS) {
			*error = SCHEDULER_ERROR_NOT_IN_DATABASE;
			debug_log(DEBUG_LEVEL_WARN,
				  "File does not exist in database.\n");
		     }
		     else {
			*error = SCHEDULER_ERROR_DATABASE_GENERIC;
			debug_log(DEBUG_LEVEL_ERROR,
				  "Database error: 0x%x.\n",
				  db_error);
		     }
		  }
		  /* BUGS
		   *    - It's not because we can get hold of file_id that
		   *      we can get hold of parts: and they might disappear
		   *      under us. Maybe add some kind of reference count not
		   *      in the database, at least for parts, so that when
		   *      the dumper/medium opens it it still exists. Before
		   *      we fix that, there will be a problem window.
		   */
	       }
	       else {
		  succeeded = FALSE;
		  debug_log(DEBUG_LEVEL_FATAL,
			    "Incorrect direction 0x%x.\n",
			    session->object.dumper.direction);
		  /* *error already set */
	       }
            }

            debug_log(succeeded ? DEBUG_LEVEL_INFO
                                : DEBUG_LEVEL_ERROR,
                      "%s/%s: %sStarting DUMPER %s session 0x%x "
                      "on file %s ID %u.\n",
                      session->object.dumper.client_name,
                      session->object.dumper.server_name,
                      succeeded ? ""
                                : "FAILED ",
                      session->direction_to_cache ? "DUMP" : "RESTORE",
                      session,
                      session->object.dumper.file_name,
                      session->file_id);
            break;
         case SCHEDULER_SESSION_TYPE_MEDIUM:
            /* Nothing special todo for now: we will set the file name and
             * IDs at the correct times (namely, in the scheduler).
             */
            session->object.medium.current_locator = LOCATOR_INVALID;
            session->object.medium.current_medium_id = MEDIUM_ID_INVALID;
            memcpy(&session->object.medium.priority_factors,
                   &default_priority_factors,
                   sizeof(session->object.medium.priority_factors));
            break;
         default:
            debug_log(DEBUG_LEVEL_FATAL,
                      "Unknown session type 0x%x.\n",
                      type);
            succeeded = FALSE;
            break;
      }

      if (!succeeded) {
         scheduler_session_free(scheduler_context, &session);
         /* sets session to NULL */
      }
   }
   else {
      debug_log(DEBUG_LEVEL_FATAL, "Can't allocate session.\n");
   }

   return session;
}

void scheduler_session_free(scheduler_context_t scheduler_context,
                            scheduler_session_t **session) {
   scheduler_info_t *info = (scheduler_info_t *) scheduler_context;

   /* TODO
    *    - See if any unterminated file or part associated with this session,
    *      and if yes, complain loudly.
    */

   if (info->session_count) {
      info->session_count--;
   }
   else {
      debug_log(DEBUG_LEVEL_FATAL,
                "INTERNAL ERROR: one scheduler_session_free() too many.\n");
   }

   free(*session);
   *session = NULL;
}

BOOL scheduler_queue_request(scheduler_context_t scheduler_context,
                             scheduler_session_t *session) {
   scheduler_info_t *info = (scheduler_info_t *) scheduler_context;

   /* If this session is a `dumper' session, this will be either a
    * request for creating a part in the database and in the temporary
    * directory, or to retrieve a part from temporary or final storage.
    * If this session is a `medium' session, it will be put to sleep
    * until we need to do I/Os.
    * NOTES
    *    - For now, we queue all requests, and the actual decisions are
    *      done within scheduler_schedule().
    */
   
   /* Very simple implementation for now */
   list_append_tail(&info->part_wait_list,
                    &session->list_header);

   return TRUE; /* Something to schedule */
}

BOOL scheduler_schedule(scheduler_context_t scheduler_context,
                        scheduler_session_t **session,
                        size_t *part_size,
                        BOOL *from_cache,
                        locator_t *locator,
                        char *part_file_name,
                        int part_file_name_length) {
   scheduler_info_t *info = (scheduler_info_t *) scheduler_context;

   /* Currently the medium scheduled is separated from the dumper scheduler */
   medium_scheduler(info);

   /* This is the work-horse which needs to awake specific sessions,
    * handing them temporary file names, updating the database, etc,
    * whether the session is a dumper or a medium.
    */

   *session = get_next_schedulable_session(info);
   if (*session) {
      debug_log(DEBUG_LEVEL_INFO,
                "scheduler_schedule(): candidate %c(%s/%s/%s, %u %u %u)%s.\n",
                ((*session)->type == SCHEDULER_SESSION_TYPE_DUMPER) ? 'D'
                                                                    : 'M',
                (*session)->object.dumper.client_name,
                (*session)->object.dumper.server_name,
                (*session)->object.dumper.file_name,
                (*session)->part_count,
                (*session)->current_part,
                (*session)->object.dumper.offset,
                ((*session)->error == SCHEDULER_ERROR_NONE) ? ""
                                                            : " ERROR");
      if ((*session)->error == SCHEDULER_ERROR_NONE) {
         if ((*session)->type == SCHEDULER_SESSION_TYPE_DUMPER) {
	    if (snprintf(part_file_name,
			 part_file_name_length,
			 "%s/%s-%s-%s-%u",
			 info->config->spool_directory,
			 (*session)->object.dumper.client_name,
			 (*session)->object.dumper.server_name,
			 (*session)->object.dumper.file_name,
			 (*session)->current_part) > -1) {
	       *part_size
		  = info->config->max_part_size;
	       if (*part_size > ((*session)->object.dumper.blocks)
                                 * (*session)->object.dumper.block_size) {
		  *part_size = (*session)->object.dumper.blocks
                               * (*session)->object.dumper.block_size;
	       }

               debug_log(DEBUG_LEVEL_INFO,
                         "part_size: %u; max: %u; dbs: %u; blocks: %u.\n",
                         *part_size,
                         info->config->max_part_size,
                         (*session)->object.dumper.block_size,
                         (*session)->object.dumper.blocks);

               if ((*session)->direction_to_cache) {
                  db_error_t db_error;

		  if (db_create_part_tuple(info->db_context,
					   (*session)->file_id,
                                           (*session)->current_part,
					   &(*session)->part_id,
					   &db_error)) {
		     if (db_create_part_spooled_tuple
                            (info->db_context,
                             (*session)->part_id,
                             part_file_name,
                             &((*session)->part_spooled_id),
                             &db_error)) {
                        debug_log(DEBUG_LEVEL_INFO,
                                  "scheduler: part and spooled C in db.\n");
                     }
                     else {
			debug_log(DEBUG_LEVEL_ERROR,
				  "scheduler_schedule(): can't create part: "
				  "database error 0x%x.\n",
				  db_error);
			(*session)->error = SCHEDULER_ERROR_DATABASE_GENERIC;
		     }
		  }
                  else {
		     debug_log(DEBUG_LEVEL_ERROR,
			       "scheduler_schedule(): can't create part: "
                               "database error 0x%x.\n",
                               db_error);
		     (*session)->error = SCHEDULER_ERROR_DATABASE_GENERIC;
                  }
               }
               else {
                  /* BUGS
                   *    - Should somewhere set session->current_part_count
                   *      to the actual one. And someone needs to do
                   *      the shifting (e.g. dumper) to actual offset.
                   *      (when the guy doesn't want to read the whole file).
                   */
                  debug_log(DEBUG_LEVEL_ERROR, "not yet impl.\n"); /* @@@ */
               }

               if ((*session)->error == SCHEDULER_ERROR_NONE) {
		  debug_log(DEBUG_LEVEL_INFO,
			    "part %s of size %ld started.\n",
			    part_file_name,
                            *part_size);
               }
	    }
	    else {
	       debug_log(DEBUG_LEVEL_FATAL,
			 "scheduler_schedule(): file name overflow.\n");
	       (*session)->error = SCHEDULER_ERROR_INTERNAL;
	    }
         }
         else if ((*session)->type == SCHEDULER_SESSION_TYPE_MEDIUM) {
	    debug_log(DEBUG_LEVEL_FATAL,
                      "Can't schedule MEDIUM from normal scheduler.\n");
         }         
         else {
	    debug_log(DEBUG_LEVEL_FATAL,
                      "Scheduling an object with unknown type 0%x.\n",
                      (*session)->type);
	    (*session)->error = SCHEDULER_ERROR_INTERNAL;
         }
      }
      else {
         debug_log(DEBUG_LEVEL_WARN, "Scheduling an object in error.\n");
      }

      return TRUE;
   }
   else {
      return FALSE;
   }
}

BOOL scheduler_finished_part(scheduler_context_t scheduler_context,
                             scheduler_session_t *session,
                             size_t part_size,
                             locator_t locator) {
   scheduler_info_t *info = (scheduler_info_t *) scheduler_context;
   BOOL result = FALSE; /* don't rerun scheduler */
   db_error_t db_error;

   if (session->error != SCHEDULER_ERROR_NONE) {
      debug_log(DEBUG_LEVEL_FATAL,
                "scheduler_finished_part: error set already: 0x%x.\n",
                session->error);
   }

   if (session->part_id == PART_ID_INVALID) {
      debug_log(DEBUG_LEVEL_FATAL,
                "scheduler_finished_part: invalid part ID.\n");
      session->error = SCHEDULER_ERROR_INTERNAL;
      return FALSE;
   }

   debug_log(DEBUG_LEVEL_INFO,
             "%s: [0x%x]: part %s of size %ld finished.\n",
             (session->type == SCHEDULER_SESSION_TYPE_MEDIUM) ? "MEDIUM"
                                                              : "DUMPER",
             session,
             session->part_file_name,
             part_size);

   if (session->type == SCHEDULER_SESSION_TYPE_DUMPER) {
      session->current_part++;

      if (session->direction_to_cache) {
         /* TODO
          *    - Maybe in a transaction.
          */
         if (db_update_part_tuple(info->db_context,
                                  session->part_id,
                                  TRUE,
                                  session->md5sum,
                                  part_size,
                                  &db_error)) {
	    if (!db_update_part_spooled_tuple(info->db_context,
					      session->part_spooled_id,
                                              TRUE,
					      &db_error)) {
	       session->error = SCHEDULER_ERROR_DATABASE_GENERIC;
	       debug_log(DEBUG_LEVEL_ERROR,
			 "Can't update spooled part info: 0x%x.\n",
			 db_error);
	    }
         }
         else {
            /* BUGS
             *    - spooled left dangling
             */
            session->error = SCHEDULER_ERROR_DATABASE_GENERIC;
            debug_log(DEBUG_LEVEL_ERROR,
                      "Can't update part info: 0x%x.\n",
                      db_error);
         }
      }
      /* else reading from cache
       * TODO
       *    - Verify md5sum
       */

      result = TRUE; /* run scheduler */
   }
   else if (session->type == SCHEDULER_SESSION_TYPE_MEDIUM) {
      if (session->direction_to_cache) {
         /* TODO
          *    - Verify md5sum
          */
	 if (!db_update_part_spooled_tuple(info->db_context,
					   session->part_spooled_id,
                                           TRUE,
					   &db_error)) {
            session->error = SCHEDULER_ERROR_DATABASE_GENERIC;
            debug_log(DEBUG_LEVEL_ERROR,
                      "Can't update spooled part info: 0x%x.\n",
                      db_error);
         }
      }
      else { /* from cache */
         if (locator == LOCATOR_INVALID) {
            debug_log(DEBUG_LEVEL_WARN, "Invalid locator!\n");
         }

         session->locator = locator;
         /* TODO
          *    - Verify md5sum
          *    - Missing verify case where no update required.
          */
	 if (!db_update_part_stored_tuple(info->db_context,
					  session->part_stored_id,
                                          TRUE,
                                          session->medium_id,
                                          session->locator,
					  &db_error)) {
            session->error = SCHEDULER_ERROR_DATABASE_GENERIC;
            debug_log(DEBUG_LEVEL_ERROR,
                      "Can't update stored part info: 0x%x.\n",
                      db_error);
         }
      }                              
                                
      result = TRUE; /* run scheduler */
   }
   else {
      debug_log(DEBUG_LEVEL_FATAL,
                "Finishing an object with unknown type 0%x.\n",
                session->type);
      session->error = SCHEDULER_ERROR_INTERNAL;
   }

   return result;
}

BOOL scheduler_finished_file(scheduler_context_t scheduler_context,
                             scheduler_session_t *session,
                             BOOL succeeded) {
   scheduler_info_t *info = (scheduler_info_t *) scheduler_context;
   db_error_t db_error;

   /* Probably called ONLY by dumpers */

   if (session->direction_to_cache) {
      /* NOTES
       *    - Maybe could do integrity checks.
       */
      session->part_count = session->current_part;
      if (!db_update_file_tuple(info->db_context,
				session->file_id,
                                TRUE,
                                session->part_count,
				&db_error)) {
	 session->error = SCHEDULER_ERROR_DATABASE_GENERIC;
	 debug_log(DEBUG_LEVEL_ERROR,
		   "Can't update stored part info: 0x%x.\n",
		   db_error);
      }
   }

   debug_log(DEBUG_LEVEL_INFO,
             "FILE (%s/%s/%s) finished %s.\n",
             session->object.dumper.client_name,
             session->object.dumper.server_name,
             session->object.dumper.file_name,
             succeeded ? "SUCCESSFULLY" : "UNSUCCESSFULLY");

   return TRUE; /* run scheduler */
}

/* Private function implementation */

static scheduler_session_t *get_next_schedulable_session
   (scheduler_info_t *info) {
   list_head_t temp_list;
   list_item_header_t *current = list_next(&info->part_wait_list, NULL);

   list_initialize(&temp_list);

   /* The idea is to pick an item from the part_wait_list (which is the
    * list of dumpers or medium waiting for a part), and to return it
    * if it matches some constraints. See the is_a_schedulable_session()
    * function implementation for more details.
    */

   /* First, let's extract candidates from the session list, and put them, in
    * a specific order, in a temporary list.
    */

   while (current) {
      list_item_header_t *next = list_next(&info->part_wait_list, current);

      if (is_a_schedulable_session(info,
                                   LIST_ITEM_TO_SESSION(current))) {
	 list_remove_item(&info->part_wait_list, current);
         list_insert_with_priority(&temp_list,
                                   current,
                                   dumper_session_is_lower_priority,
                                   NULL);
      }

      current = next;
   }

   /* Then, select the first candidate, if any @@@ none supported ? @@@ */
   current = list_remove_head(&temp_list);

   /* NOTES
    *    - If we return one, the scheduler will make sure we are called
    *      again quickly: this will allow us to start multiple
    *      sessions if possible/required.
    */

   /* Then move back all candidates to the normal list, also supporting
    * the degenerated cases.
    */
   list_cat_tail(&info->part_wait_list, &temp_list);

   return current ? LIST_ITEM_TO_SESSION(current)
                  : NULL;
}

/* NOTES
 *    - All values expressed in fs's units, see f_bsize.
 */
static BOOL ok_to_io(const char *directory,
		     const BOOL create_file,
                     const unsigned int chunk_size,
		     const constraint_t constraints) {
   struct statfs buf;

   if (statfs(directory, &buf)) {
      perror("stat()");
   }
   else {
      BOOL result_for_files;

      printf("optimal block size: %d\n", buf.f_bsize);
      printf("      total blocks: %ld\n", buf.f_blocks);
      printf("       free blocks: %ld\n", buf.f_bfree);
      printf("  user free blocks: %ld\n", buf.f_bavail);
      printf("      total  files: %ld\n", buf.f_files);
      printf("       free  files: %ld\n", buf.f_ffree);

      result_for_files
         = (constraints.min_free_files < buf.f_ffree)
	   && (buf.f_ffree && buf.f_files
	       && (buf.f_ffree
		   > (buf.f_files * constraints.min_free_files_ratio)
                     / 100));
      if ((!create_file || result_for_files)
          && (buf.f_bavail >= chunk_size)
          && (constraints.min_free_blocks
              < (buf.f_bavail - chunk_size)) /* user */
          && (buf.f_bavail && buf.f_blocks
              && ((buf.f_bavail - chunk_size)
                  > (buf.f_blocks * (constraints.min_free_blocks_ratio
                                     /100))))) {
         return TRUE;
      }
   }

#if 0
   return FALSE;
#else
   return TRUE;
#endif
}

static BOOL dumper_session_is_lower_priority(list_head_t *list,
					     list_item_header_t *new,
					     list_item_header_t *current,
					     void *private) {
#if 0
   scheduler_session_t *current_session
      = LIST_ITEM_TO_SESSION(current);
   scheduler_session_t *new_session
      = LIST_ITEM_TO_SESSION(item);

   return new_session->priority_factors.client
          > current_session->priority_factors.client;

   /* BUGS
    *    - priority_factors is a misnomer
    */
#else
   return FALSE; /* insert as soon as possible, for now */
#endif
}

static BOOL is_a_schedulable_session(scheduler_info_t *info,
                                     scheduler_session_t *session) {

   /* NOTES
    *    - This function shall not modify anything in the session, because
    *      this is just used to evaluate the *candidates*, not to
    *      effectively start operations.
    * BUGS
    *    - In that case, some constraints should be evaluated statically,
    *      or we should change to a schedule list approach (runnable, etc).
    */

   if (session->error != SCHEDULER_ERROR_NONE) {
      return FALSE; /* Can't schedule when in error */
   }

   /* Later, we may add constraints such as: only activate medium for
    * writing or reading if the right volume is loaded, etc. We might
    * also schedule some medium reads in advance, e.g., when the dumper
    * tells us about it (not yet possible). We might also proritize the
    * requests: write to tape when data in spool or send to dumper when data
    * on spool might be inherently priority operations. Also, the dumper
    * might set priorities. Also we need to implement the complete set
    * of priorities (later). Some of the prioritization has to be
    * done in the actor function.
    *
    * Finally, some could be scheduled for error (timeout on waiting for
    * a volume, for example).
    *
    * When there isn't enough room in the spool for a dumper write or
    * medium read, we will call a maybe_free_some_space_in_spool() that
    * might drop data in spool which is also stored, maybe with a
    * good last usage pattern policy.
    */

   if (session->type == SCHEDULER_SESSION_TYPE_DUMPER) {
      /* For a dumper which wants to write a part to the
       * spool, we will only return (== activate) it if:
       *    - the spool is not too full
       *    - there aren't too many dumper operations going on already
       *
       * For a dumper which wants to read a part form the spool, we will only
       * return if if the spool already contains the required data.
       */

      if ((info->running_dumpers /* the ones having parts */
           < info->config->max_running_dumpers)) {
         if (session->direction_to_cache) {
            /* Does it LOOK schedulable ? */
            if (ok_to_io(info->config->spool_directory,
                         TRUE,
                         info->config->max_part_size,
                         info->constraints)) {
               return TRUE; /* we do not instanciate, either */
            }
            else {
               initiate_spool_purge(info);
            }
         }
         else {
            /* We do not lock, that must be done by the caller
             * in an atomic operation, after deciding which.
             * BUGS
             *    - and this is bad since we already have set the IDs ...
             */
            return is_part_in_spool(info,
                                    session,
                                    TRUE, /* dumper */
                                    TRUE /* complete */);
            /* this means: is there a part in the spool which was completely
             * read from tape and required by this dumper ?
             */
         }
      }
   }
   else if (session->type == SCHEDULER_SESSION_TYPE_MEDIUM) {
      /* For a medium, we activate if:
       *    - there is anything in the spool that is ready for being written.
       *    - there is a need of some data stored on the medium, and there is
       *      room for 
       * (in the medium case, _we_ set the data direction).
       */

      /* For now, saving has priority over reading, since saving may
       * drop some data off spool. However hasn't a great incidence here.
       */

      if (is_part_in_spool(info,
                           session,
                           FALSE, /* medium */
                           TRUE /* complete */)) {
         /* this means: there is a part in the spool which was completely
          * written to by a dumper, and is ready for sending to this medium.
          */
         return TRUE;
      }

      if (is_part_in_spool(info, session, TRUE, FALSE)) {
         /* this means: is there a part in the spool which is requested by
          * a dumper and would be available on our medium but isn't complete.
          */
         if (ok_to_io(info->config->spool_directory,
                      TRUE,
                      info->config->max_part_size,
                      info->constraints)) {
            return TRUE;
         }
         else {
            initiate_spool_purge(info);
         }
      }
   }
   /* else ignore */

   return FALSE; /* Do not schedule this one */
}

static void initiate_spool_purge(scheduler_info_t *info) {
   debug_log(DEBUG_LEVEL_INFO, "Spool purge started.\n");

   info->spool_full = TRUE;

   /* @@@ Delete one unused file from the spool, choosing the one that
    * was created a long time ago and not accessed recently. Unused 
    * means that it's not going to be used (ie we didn't give it as a
    * part for some session; either we have a list or we browse through
    * the sessions).
    */

   /* The idea is to remove some parts of the spool when they are
    * no longer used, and haven't been used for some time. It is
    * perfectly legal not to delete any file. This could also use
    * the policies.
    */

   debug_log(DEBUG_LEVEL_INFO, "Spool purge finished.\n");

   info->spool_full = FALSE; /* @@@ well ... */
}

static BOOL is_part_in_spool(scheduler_info_t *info,
                             scheduler_session_t *session,
                             BOOL dumper,
                             BOOL complete) {
   session->file_id = FILE_ID_INVALID;
   session->part_id = PART_ID_INVALID;
   session->part_spooled_id = PART_SPOOLED_ID_INVALID;

   /* NOTES
    *    - We could use part_id directly, in the session, if
    *      anyone could create a part. For now, only dumper
    *      creates parts.
    *    - We assume this is for read access only.
    *    - We set the appropriate values in the session.
    * BUGS
    *    - This should be re-worked, really.
    */

   if (dumper && session->direction_to_cache) {
      debug_log(DEBUG_LEVEL_FATAL, "isis: incoherency.\n");
      return FALSE;
   }

   if (dumper) {
      db_error_t db_error;

      /* So, first check if a file as named is available */
      if (!db_lookup_file_tuple(info->db_context,
				session->object.dumper.file_name,
				session->object.dumper.client_name,
				session->object.dumper.server_name,
				&session->file_id,
                                &session->part_count,
				&db_error)) {
         debug_log(DEBUG_LEVEL_ERROR, "isis: F db_error 0x%x.\n", db_error);
         session->error = SCHEDULER_ERROR_EOF; /* not really necessary */
      }
      else if (!db_lookup_part_tuple(info->db_context,
				     session->file_id,
                                     session->current_part,
                                     &session->part_id,
  				     &db_error)) {
         debug_log(DEBUG_LEVEL_ERROR, "isis: P db_error 0x%x.\n", db_error);
         session->error = SCHEDULER_ERROR_EOF; /* not really correct */
      }
      else if (!db_lookup_part_spooled_tuple(info->db_context,
				             session->part_id,
                                             &session->part_spooled_id,
  				             &db_error)) {
         debug_log(DEBUG_LEVEL_ERROR, "isis: PS db_error 0x%x.\n", db_error);
         session->error = SCHEDULER_ERROR_EOF; /* not really correct */
      }

      return TRUE;
   }
   else {
      debug_log(DEBUG_LEVEL_INFO, "taper file in spool not yet impl.\n");
   }
                            
   return FALSE;
}

static void medium_scheduler(scheduler_info_t *info) {
   list_head_t temp_list;
   list_head_t scheduled_list;
   list_item_header_t *current = NULL;

   list_initialize(&temp_list);
   list_initialize(&scheduled_list);

   /* Busy/error medium processes won't be in that (badly named) list,
    * only idle are.
    */

   while ((current = list_next(&info->part_wait_list, current))) {
      scheduler_session_t *current_medium_session
         = LIST_ITEM_TO_SESSION(current);

      /* Well it's not really a part_wait_list ... */

      if (current_medium_session->type == SCHEDULER_SESSION_TYPE_MEDIUM) {
         /* Build a prioritized list of all parts that are
          * being required for medium I/O and that have a stored candidate
          * that is reachable through this device, or are for writing.
          * NOTES
          *    - This allocates memory for part_stored_memory_t items.
          *    - If memory is short, won't be able to create a list,
          *      and may not schedule some which would require to be.
          */
         find_parts_stored_memory(info,
                                  &temp_list,
                                  current_medium_session);
      }
   }

   /* Extract in order, find matching medium process (device). Don't start
    * one on a busy device, don't start if one filling the requirement was
    * started already.
    */
   while ((current = list_remove_head(&temp_list))) {
      part_stored_memory_t *part_stored_memory
         = LIST_ITEM_TO_PART_STORED_MEMORY(current);

      if (medium_device_still_idle
             (info, part_stored_memory->medium_process)
          && still_required_to_do_op(info,
                                     &scheduled_list,
                                     part_stored_memory)) {
         /* Mark virtual part_stored for scheduling */
         list_append_tail(&scheduled_list,
                          &part_stored_memory->list_header);
         /* Remove medium process device from part_wait_list
          * since will be scheduled.
          */
         list_remove_item(&info->part_wait_list,
                          &part_stored_memory->medium_process->list_header);
      }
      else {
         free(part_stored_memory);
      }
   }

   /* TODO
    *    - actual scheduling of all items in scheduled_list.
    */

   /* Free the remaining items */
   while ((current = list_remove_head(&scheduled_list))) {
      free(LIST_ITEM_TO_PART_STORED_MEMORY(current));
   }
}                                                        

static priority_t compute_medium_part_priority
   (scheduler_info_t *info,
    part_stored_memory_t *part_stored_memory) {
   priority_t priority = 0;
   session_medium_info_t *object
      = &part_stored_memory->medium_process->object.medium;
   priority_factors_t *priority_factors = &object->priority_factors;

#if 0
   /* client-set priority
    * BUGS
    *    - We do NOT get the client (dumper) priority that way.
    */
   priority += (part_stored_memory->priority * priority_factors->client);
#endif

   /* the age of this part, ie the time it has waited for scheduling */
   priority += (part_stored_memory->age * priority_factors->age);

   if (info->spool_full) {
      if (part_stored_memory->will_be_freed_by_writing) {
         priority += priority_factors->could_free_space_by_writing;
      }
   }

   /* NOTES
    *    - Ideally the locator will be blocks instead of files for more
    *      precize penalties.
    */

   if (object->current_medium_id != MEDIUM_ID_INVALID) {
      if (object->current_medium_id == part_stored_memory->medium_id) {
	 /* Never executes when medium_id invalid, e.g. write */
         priority += priority_factors->same_medium;
      }
      else {
         /* Not same medium (but medium loaded) */

         /* Penalty of rewinding */
         if (object->current_locator != LOCATOR_INVALID) {
            priority_substract
               (&priority,
                object->current_locator
                * priority_factors->rewind_pos);
         }
         /* else assume start of tape */

         /* Penalty of ejecting */

         priority_substract
            (&priority, priority_factors->eject);
      }
   }

   if (object->current_locator != LOCATOR_INVALID) {
      if (part_stored_memory->locator != LOCATOR_INVALID) {
         if (part_stored_memory->locator == (object->current_locator + 1)) {
            priority += priority_factors->contiguous_file;
         }
         else if (part_stored_memory->locator > object->current_locator) {
            priority_substract(&priority,
                               (part_stored_memory->locator
                                - object->current_locator)
                               * priority_factors->fsf);
         }
         else {
            /* We assume when value 1, means to read it again we
             * need to bsf.
             */
            priority_substract(&priority,
                               (1
                                + object->current_locator
                                - part_stored_memory->locator)
                               * priority_factors->bsf);
         }
      }
      /* else undetermined penalty */
   }
   else {
      priority_substract(&priority,
                         part_stored_memory->locator
                         * priority_factors->fsf);
   }

   if (part_stored_memory->locator != LOCATOR_INVALID) {
      if (part_stored_memory->for_write) {
         /* Please write first on already quite used volumes.
          * TODO
          *    - colocation and no-colocation strategies as with the
          *      rest of this function.
          */
         priority += priority_factors->fullness * part_stored_memory->locator;
      }
   }

   return priority;
}

static void priority_substract(priority_t *priority,
                               const priority_t value) {
   if (*priority > value) {
      *priority -= value;
   }
   else {
      *priority = 0;
   }
}

static void find_parts_stored_memory(scheduler_info_t *info,
                                     list_head_t *list,
                                     scheduler_session_t *medium_process) {
   /* All stored parts that could be created from spooled parts
    *    - Because they are spooled but not yet stored
    */

   /* Parts that could be read from medium
    *    - Because they are requested by a dumper.
    *    - And they exist as stored parts.
    */
}

BOOL medium_device_still_idle(scheduler_info_t *info,
                              scheduler_session_t *medium_process) {
   list_item_header_t *current = NULL;

   /* Basically a medium process (device) is idle when it is on
    * the part_wait list.
    */

   while ((current = list_next(&info->part_wait_list,
                               current))) {
      if (LIST_ITEM_TO_SESSION(current) == medium_process) {
         return TRUE;
      }
   }

   return FALSE;
}

BOOL still_required_to_do_op(scheduler_info_t *info,
                             list_head_t *scheduled_list,
                             part_stored_memory_t *part_stored_memory) {
   list_item_header_t *current = NULL;

   /* Should return TRUE when there isn't any part on the scheduled
    * list which has the same part ID.
    */

   while ((current = list_next(scheduled_list,
                               current))) {
      if (LIST_ITEM_TO_PART_STORED_MEMORY(current)->part_id
           == part_stored_memory->part_id) {
         return FALSE;
      }
   }

   return TRUE;
}