/* Project filed -- Intelligent file cache and migration system
 * schededuler_interface.h -- I/O scheduler interface definitions.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the data I/O scheduler data types and function
 *    definitions, as called from the main (father) process.
 * NOTES
 * BUGS
 *    - Might need some rethinking. (data structures, polling, etc).
 *      E.g.: session should only contain non specific data, also
 *      for dumpers and medium (tapers), etc.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: scheduler_interface.h,v 1.10 2001/02/04 19:57:15 schaefer Exp $
 */

#ifndef _SCHEDULER_INTERFACE_H
#define SCHEDULER_INTERFACE_H

/* Type definitions */

#define MAX_PATHNAME_LENGTH 1024

typedef enum {
   SCHEDULER_SESSION_TYPE_DUMPER, /* client (TCP) <-> server */
   SCHEDULER_SESSION_TYPE_MEDIUM  /* medium I/O */
} scheduler_session_type_t;

typedef enum {
   SCHEDULER_ERROR_NONE,
   SCHEDULER_ERROR_NOT_IN_DATABASE,
   SCHEDULER_ERROR_DATABASE_FILE_EXISTS,
   SCHEDULER_ERROR_DATABASE_GENERIC,
   SCHEDULER_ERROR_IO,
   SCHEDULER_ERROR_BAD_CLIENT_NAME, /* contains illegal characters */
   SCHEDULER_ERROR_BAD_SERVER_NAME, /* contains illegal characters */
   SCHEDULER_ERROR_BAD_FILE_NAME,   /* contains illegal characters */
   SCHEDULER_ERROR_EOF, /* Not really an error. Needed for readers */
   SCHEDULER_ERROR_INTERNAL
} scheduler_error_t;

typedef unsigned int part_count_t;
typedef unsigned int part_id_t;
typedef unsigned int file_id_t;
typedef unsigned int part_spooled_id_t;
typedef unsigned int part_stored_id_t;
typedef unsigned int medium_id_t;

#define FILE_ID_INVALID 0xFFFFFFFF
#define PART_ID_INVALID 0xFFFFFFFF
#define PART_SPOOLED_ID_INVALID 0xFFFFFFFF
#define PART_STORED_ID_INVALID 0xFFFFFFFF
#define MEDIUM_ID_INVALID 0xFFFFFFFF

#define LOCATOR_INVALID 0
typedef unsigned int locator_t;

#define PRIORITY_LOWER 0
#define PRIORITY_HIGHER 0xFFFFFFFF
typedef unsigned int priority_t;

typedef unsigned int age_t;

typedef unsigned int priority_factor_t;

typedef struct {
   /* Some of these are bonuses, some are penalties */
   priority_factor_t age;
   priority_factor_t same_medium;
   priority_factor_t rewind_pos;
   priority_factor_t could_free_space_by_writing;
   priority_factor_t eject;
   priority_factor_t contiguous_file;
   priority_factor_t fsf;
   priority_factor_t bsf;
   priority_factor_t fullness;
} priority_factors_t;

typedef struct {
   priority_factors_t priority_factors;

   locator_t current_locator;
   medium_id_t current_medium_id;
} session_medium_info_t;

typedef union {
   session_dumper_info_t dumper;
   session_medium_info_t medium;
} scheduler_session_union_t;

typedef struct {
   list_item_header_t list_header;

   scheduler_session_type_t type;

   scheduler_session_union_t object;

   part_id_t part_id;
   file_id_t file_id;
   BOOL direction_to_cache;
   part_count_t current_part;
   part_count_t part_count;

   char part_file_name[MAX_PATHNAME_LENGTH];
   scheduler_error_t error;

   /* This md5sum is either set by the scheduler when the part is already
    * in the database, or by the dumper when creating.
    */
   unsigned char md5sum[MD5_HASH_STRING_LENGTH];

   part_spooled_id_t part_spooled_id;
   part_stored_id_t part_stored_id;

   medium_id_t medium_id;
   locator_t locator;

   void *child;
} scheduler_session_t;

typedef void * scheduler_context_t;

typedef unsigned int running_count_t;

typedef struct {
   char *server_name; /* must be freed since allocated dynamically by
                       * server.c's load_configuration()
                       */
   char *spool_directory; /* must be freed */
   char *medium_configurations; /* must be freed if not NULL */
   int port;

   size_t max_part_size;

   running_count_t max_running_dumpers;
} server_config_t;

/* Function definitions */

/* NAME
 *    scheduler_initialize
 * DESCRIPTION
 *    Initializes the scheduler.
 * RESULT
 *    scheduler_context_t, or NULL if failed.
 * NOTES
 *    - The data pointed to by server_config shall not be freed until
 *      after scheduler_deinitialize().
 * BUGS
 * TODO
 */
scheduler_context_t scheduler_initialize(const server_config_t *server_config);

/* NAME
 *    scheduler_deinitialize
 * DESCRIPTION
 *    Deinitializes the scheduler.
 * RESULT
 *    NONE
 * NOTES
 *    - Will set *scheduler_context to NULL.
 * BUGS
 * TODO
 */
void scheduler_deinitialize(scheduler_context_t *scheduler_context);

/* NAME
 *    scheduler_session_allocate
 * DESCRIPTION
 *    Allocates a new session.
 * RESULT
 *    scheduler_session_t *, or NULL if couldn't be allocated.
 * NOTES
 *    - child will be set in the session (child).
 *    - If returns NULL, look at *error for the detailed error.
 * BUGS
 * TODO
 */
scheduler_session_t *scheduler_session_allocate
   (scheduler_context_t scheduler_context,
    scheduler_session_union_t *session_info,
    void *child,
    scheduler_session_type_t type,
    scheduler_error_t *error);

/* NAME
 *    scheduler_session_free
 * DESCRIPTION
 *    Free a session.
 * RESULT
 *    NONE
 * NOTES
 *    - session will be NULLified.
 * BUGS
 * TODO
 */
void scheduler_session_free(scheduler_context_t scheduler_context,
                            scheduler_session_t **session);

/* NAME
 *    scheduler_queue_request
 * DESCRIPTION
 *    Queues a request to the request queue.
 * RESULT
 *    BOOL, if TRUE, caller must call scheduler_schedule().
 * NOTES
 *    - We systematically queue the part requests, and we possibly return
 *      TRUE immediately if a scheduler_schedule() is needed.
 * BUGS
 * TODO
 */
BOOL scheduler_queue_request(scheduler_context_t scheduler_context,
                             scheduler_session_t *session);

/* NAME
 *    scheduler_schedule
 * DESCRIPTION
 *    Based on internal rules, the scheduler will return part of files
 *    for the specified process (session).
 * RESULT
 *    TRUE if scheduler gave some information, false otherwise.
 *    If the scheduler returns TRUE, this function must be
 *    re-called again, until it returns FALSE (no more entries).
 * NOTES
 *    - Error is stored in session->error
 * BUGS
 * TODO
 */
BOOL scheduler_schedule(scheduler_context_t scheduler_context,
                        scheduler_session_t **session,
                        size_t *part_size,
                        BOOL *from_cache,
                        locator_t *locator,
                        char *part_file_name,
                        int part_file_name_length);

/* NAME
 *    scheduler_finished_part
 * DESCRIPTION
 *    Tells the scheduler when a part was finished successfully.
 * RESULT
 *    TRUE if scheduler_schedule() should be run.
 * NOTES
 *    - This means that the scheduler can schedule writing of the
 *      part to final medium (write case), or delete the part
 *      (read case when part is on final medium, and scheduler needs
 *      the space).
 * BUGS
 * TODO
 */
BOOL scheduler_finished_part(scheduler_context_t scheduler_context,
                             scheduler_session_t *session,
                             size_t part_size,
                             locator_t locator);

/* NAME
 *    scheduler_finished_file
 * DESCRIPTION
 *    Tells the scheduler when a file was finished successfully, or not.
 * RESULT
 *    TRUE if scheduler_schedule() should be run.
 * NOTES
 *    - In all the case, the finished file message should be replied by
 *      the caller.
 *    - This means that the scheduler updates the state of the
 *      file in its database, and possibly deletes it if
 *      it was unsuccessful, and specific modes where chosen.
 * BUGS
 * TODO
 */
BOOL scheduler_finished_file(scheduler_context_t scheduler_context,
                             scheduler_session_t *session,
                             BOOL succeeded);
#endif /* !_SCHEDULER_INTERFACE_H */
