/* Project filed -- Intelligent file cache and migration system
 * server.c -- Implementation of a server.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the actual server program, scheduler and dumper.
 * NOTES
 *    - If size of paths, etc, augment a lot, this means problems with
 *      our pipe message system, probably. Or maybe handled by
 *      process module anyway ?
 * BUGS
 *    - something wrong with the init of unions.
 * TODO
 *    - Clean up, priv function definitions doc, often used routines to
 *      lib or server-lib, modularize, etc. Maybe use external programs ?
 *    - do the child to father encoding, decoding in father, and
 *      calling of scheduler functions.
 *    - add handler for SIGTERM and SIGINT, setting BOOL and
 *      used in the select() loop in EINTR (with correct cleanup).
 *    - maybe modularize dumper/medium so that can be external programs,
 *      with a very well defined API and support libraries, and rewrite
 *      so it can maybe register, and be similar.
 *    - check that we free the child_data_t whenever a process terminates.
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: server.c,v 1.15 2001/02/04 19:57:15 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h> 
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "global.h"
#include "inet_interface.h"
#include "auth_interface.h"
#include "crypt_interface.h"
#include "misc_interface.h"
#include "process_interface.h"
#include "lists_interface.h"
#include "configuration_interface.h"
#include "debug_interface.h"
#include "md5_interface.h"
#include "filesend_interface.h"
#include "protocol_interface.h"
#include "server_defs.h" /* General definitions */
#include "server_client_defs.h" /* TCP client/server protocol */
#include "scheduler_interface.h"
#include "changer_interface.h"
#include "taper_interface.h"

/* Private type definitions */

typedef struct {
   server_config_t config;

   char *program_name;

   list_head_t secret_list; /* client/server passwords */

   scheduler_context_t scheduler_context;

   process_context_t process_context;
} server_info_t;

typedef struct {
   const char *medium_device_name;
   const char *changer_device_name; /* @@@ should be a changer ID instead, and
				     * medium should ask main about changes.
				     * through that ID @@@
				     */
} medium_data_t;

typedef struct {
   fd_t cd; /* socket connection descriptor */

   char *server_name;
   list_head_t *secret_list;

   size_t max_part_size; /* from config */
} dumper_data_t;

typedef struct {
   list_item_header_t list_header;

   char *client;
   char *server_password;
   char *client_password;
} secret_item_t;

typedef struct {
   scheduler_session_t *session; /* only for father */

   union {
      dumper_data_t dumper;
      medium_data_t medium;
   } object;
} child_data_t;

typedef struct {
   session_dumper_info_t hello;

   child_data_t *child_data; /* can be gotten from child_context */
   child_context_t child_context;

   protocol_info_t protocol_info;

   size_t current_part_size;

   BOOL had_a_file; /* not first file in get_next() */
} session_t;

typedef enum {
   MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER,
   MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_MEDIUM,
   MESSAGE_TYPE_TO_SCHEDULER_GET_PART,
   MESSAGE_TYPE_TO_SCHEDULER_FINISHED_PART,
   MESSAGE_TYPE_TO_SCHEDULER_FINISHED_FILE, /* session */

   MESSAGE_TYPE_FROM_SCHEDULER_SET_SESSION,
   MESSAGE_TYPE_FROM_SCHEDULER_GET_PART,
   MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_PART,
   MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_FILE, /* session */

   MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP,

   MESSAGE_TYPE_FROM_MEDIUM_OP_RESULT
} message_type_t;

typedef struct {
   message_type_t type;
} message_header_t;

typedef struct {
   message_header_t header;

   session_medium_info_t medium_info;
} message_to_scheduler_set_session_medium_t;

typedef struct {
   message_header_t header;

   session_dumper_info_t dumper_info;
} message_to_scheduler_set_session_dumper_t;

typedef struct {
   message_header_t header;
} message_to_scheduler_get_part_t;

typedef struct {
   message_header_t header;

   size_t part_size;

   char part_file_name[MAX_PATHNAME_LENGTH]; /* can be a label if medium */

   BOOL from_cache; /* Used as double-check for dumper; and
                     * required for medium.
                     */
   locator_t locator; /* required for medium only, when !from_cache */

   scheduler_error_t error;
} message_from_scheduler_get_part_t;

typedef struct {
   message_header_t header;

   size_t part_size;

   unsigned char md5sum[MD5_HASH_STRING_LENGTH]; /* only for dumper with
                                                  * to cache.
                                                  */
   locator_t locator; /* required for medium only, when from_cache */
} message_to_scheduler_finished_part_t;

typedef struct {
   message_header_t header;

   scheduler_error_t error;
} message_from_scheduler_finished_part_t;

typedef struct {
   message_header_t header;

   BOOL succeeded;
} message_to_scheduler_finished_file_t;

typedef message_from_scheduler_finished_part_t
   message_from_scheduler_finished_file_t;

typedef message_from_scheduler_finished_part_t
   message_from_scheduler_set_session_t;

typedef enum {
   MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP_TYPE_MOUNT,
   MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP_TYPE_EJECT,
   MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP_TYPE_LABEL,
   MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP_TYPE_SHUTDOWN
} message_type_from_scheduler_medium_op_type_t;

typedef struct {
   message_header_t header; /* contains type */

   message_type_from_scheduler_medium_op_type_t op;
   char label[MAX_LABEL_NAME_SIZE];
} message_type_from_scheduler_medium_op_t;

typedef struct {
   message_header_t header;

   BOOL succeeded; /* could be more detailed ... */
} message_type_from_medium_op_result_t;

/* Macros */

#define LIST_ITEM_TO_SECRET(item) \
   LIST_ITEM_TO_OBJECT((item), secret_item_t)

#define SESSION_OF_CHILD(child) \
   (((child_data_t *) process_get_private_of((child)))->session)

/* Private function definitions */

/* BUGS
 *    - Missing function definitions, modularization, etc.
 */
static int do_processing(server_info_t *info,
                         fd_t sd);
static int do_operation(session_t *session,
                        fd_t sd,
                        client_initial_message_t *initial_message);
static int handle_dumper(child_context_t child_context);
static int handle_medium(child_context_t child_context);

static BOOL load_configuration(const char *config_file_name,
                               server_config_t *config);
static BOOL load_secret_file(list_head_t *list,
                             const char *secret_file_name);
static void free_secret_list(list_head_t *list);
static BOOL add_secret_item(list_head_t *list,
                            const char *client,
                            const char *server_password,
                            const char *client_password);
static secret_item_t *find_secret_item(list_head_t *list, const char *client);
static void free_secret_item(secret_item_t *item);
static BOOL parse_line(list_head_t *list,
                       unsigned char *buffer,
                       int buffer_length);
static BOOL encode_scheduler_set_session_dumper(session_t *session);
static BOOL encode_scheduler_get_part(session_t *session,
                                      size_t *part_size,
                                      char *part_file_name,
                                      int part_file_name_length,
                                      BOOL *eof);
static BOOL encode_scheduler_finished_part(session_t *session,
                                           size_t part_size,
                                           locator_t locator,
                                           unsigned char *md5sum);
static BOOL encode_scheduler_finished_file(session_t *session,
                                           BOOL succeeded);
static BOOL get_next_file_read(void *caller_private,
                               fd_t *fd,
                               md5_hash_ptr_t last_file_md5sum,
                               BOOL *eof);
static BOOL get_next_file_write(void *caller_private,
                               fd_t *fd,
                               unsigned int *file_max,
                               md5_hash_ptr_t last_file_md5sum,
                               const BOOL last);

/* NAME
 *    launch_medium_if_any
 * DESCRIPTION
 *    Launches all the required medium processes, depending on
 *    configuration.
 * RESULT
 *    TRUE if succeeded
 * NOTES
 * BUGS
 * TODO
 */
static BOOL launch_medium_if_any(server_info_t *info);

/* NAME
 *    medium_load_config
 * DESCRIPTION
 *    Loads configuration for the specified medium and allocates
 *    the space for medium_device_name and changer_device_name (which
 *    will have to be freed by caller).
 * RESULT
 *    BOOL; TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
static BOOL medium_load_config(server_info_t *info,
                               const char *current,
                               char **medium_device_name,
                               char **changer_device_name);

/* NAME
 *    medium_launch
 * DESCRIPTION
 *    Launches the specified medium process.
 * RESULT
 *    BOOL; TRUE if ok
 * NOTES
 * BUGS
 * TODO
 */
static BOOL medium_launch(server_info_t *info,
                          const char *medium_device_name,
                          const char *changer_device_name);

/* NAME
 *    stop_all_launched_children
 * DESCRIPTION
 *    Stop all launched medium (medium) or dumpers.
 * RESULT
 *    NONE
 * NOTES
 * BUGS
 * TODO
 */
static void stop_all_launched_children(server_info_t *info);

/* Private variables */

/* Function implementation */

int main(int argc, char **argv) {
   int result;

   if (argc != 3) {
      debug_log(DEBUG_LEVEL_FATAL,
                "%s config-file secret-file\n%s: bad args.\n",
                argv[0], argv[0]);
      result = 2;
   }
   else if (signal(SIGPIPE, SIG_IGN) == SIG_ERR) {
      perror("signal()");
      result = 1;
   }
   else {
      server_info_t server_info;

      if (load_configuration(argv[1], &server_info.config)) {
         if (load_secret_file(&server_info.secret_list, argv[2])) {

            fd_t sd = inet_create_endpoint(server_info.config.port, TRUE);
            if (sd == -1) {
               debug_log(DEBUG_LEVEL_FATAL,
                         "%s: endpoint creation failed.\n", argv[0]);
               result = 1;
            }
            else {
               server_info.program_name = argv[0];
               if ((server_info.scheduler_context
                    = scheduler_initialize(&server_info.config))) {
                  server_info.process_context = process_initialize();
                  if (server_info.process_context) {
                     if (launch_medium_if_any(&server_info)) {
                        result = do_processing(&server_info,
                                               sd);
                     }
                     else {
                        debug_log(DEBUG_LEVEL_FATAL,
                                  "can't launch medium.\n");
                     }

                     stop_all_launched_children(&server_info);

                     process_deinitialize(&server_info.process_context);
                  }
                  else {
                     debug_log(DEBUG_LEVEL_FATAL,
                               "process_initialize() failed.\n");
                  }

                  scheduler_deinitialize(&server_info.scheduler_context);
               }
               else {
                  debug_log(DEBUG_LEVEL_FATAL,
                            "scheduler_initialize() failed.\n");
               }

               if (close(sd)) {
                  debug_perror(DEBUG_LEVEL_ERROR, "close()");
                  if (result == 0) {
                     result = 1;
                  }       
               }
            }

            free_secret_list(&server_info.secret_list);
         }
         else {
            debug_log(DEBUG_LEVEL_FATAL,
                      "%s: can't load secret list %s.\n",
                      argv[0],
                      argv[2]);
         }

         free(server_info.config.server_name);
         free(server_info.config.spool_directory);
         if (server_info.config.medium_configurations) {
            free(server_info.config.medium_configurations);
         }
      }
      else {
         debug_log(DEBUG_LEVEL_FATAL,
                   "%s: error loading configuration %s.\n",
                   argv[0],
                   argv[1]);
         result = 1;
      }
   }

   return result;
}

static int do_processing(server_info_t *info,
                         fd_t sd) {
   fd_t cd;
   struct sockaddr_in peer_addr;
   socklen_t length;
   int result = 0;

   /* BUGS
    *    - we assume that the child must be a session (dumper), it
    *      can also be a medium with a different structure ...
    *    - or we should have two different process object instances.
    */
   while (result == 0) {
      fd_t max_fd = sd;
      fd_set readfds;
      int retval;
   
      FD_ZERO(&readfds);
      FD_SET(sd, &readfds);
   
      process_select_father(info->process_context, &readfds, &max_fd);
   
      retval = select(max_fd + 1, &readfds, NULL, NULL, NULL);
      if (retval > 0) {
         child_context_t child;
         void *private;
         BOOL running;
         BOOL should_reschedule = FALSE;
         int message_size;
   
         if (FD_ISSET(sd, &readfds)) {
            length = sizeof(peer_addr);
            cd = accept(sd, &peer_addr, &length);
            if (cd == -1) {
               debug_perror(DEBUG_LEVEL_ERROR, "accept()");
               result = 1;
            }
            else {
               child_data_t *child_data
                  = (child_data_t *) malloc(sizeof(child_data_t));

               if (child_data) {
                  child_context_t child;

		  child_data->session = NULL;
		  child_data->object.dumper.cd = cd;
		  child_data->object.dumper.server_name
                     = info->config.server_name;
		  child_data->object.dumper.secret_list = &info->secret_list;
                  child_data->object.dumper.max_part_size
                     = info->config.max_part_size; /* from config */

		  debug_log(DEBUG_LEVEL_INFO,
			    "peer: %s port %d.\n",
			    inet_ntoa(peer_addr.sin_addr),
			    ntohs(peer_addr.sin_port));

		  child = process_launch(info->process_context,
					 handle_dumper,
					 (void *) child_data);
		  if (child) {
		     debug_log(DEBUG_LEVEL_INFO,
			       "new child 0x%x.\n",
			       (int) child);
		  }
		  else {
		     debug_log(DEBUG_LEVEL_ERROR,
			       "failed creating a child.\n");
		  }
               }
               else {
                  debug_log(DEBUG_LEVEL_ERROR, "malloc() child.\n");
               }
   
               /* Either close because of error, or close
                * father's copy. Will be a bit abrupt for the other size
                * if error.
                */
               close(cd); /* res. ign. */
            }
         }
   
         debug_log(DEBUG_LEVEL_INFO, "BEFORE\n");
         while (process_next_ready_father(info->process_context,
                                          &child,
                                          &private,
                                          &running,
                                          &message_size,
                                          &readfds,
                                          max_fd)) {
            debug_log(DEBUG_LEVEL_INFO, "AFTER (0x%x)\n", (int) child);
            if (running) {
               unsigned char buffer[message_size];
   
               debug_log(DEBUG_LEVEL_INFO,
                         "child 0x%x sends message len=%d.\n",
                         (int) child,
                          message_size);
   
               if (message_size < sizeof(message_header_t)) {
                  debug_log(DEBUG_LEVEL_ERROR,
                            "Message would be too small.\n");
                  /* @@@ */
               }
               else if (process_father_receive_message(child,
                                                       buffer,
                                                       &message_size)) {
                  scheduler_session_t *session;
                  scheduler_error_t error;
   
                  debug_log(DEBUG_LEVEL_INFO, "DONE reading message.\n");
   
                  /* BUGS
                   *    - We don't verify the message length. This is
                   *      no security risk (no network involved), but
                   *      should be done anyway.
                   */
   
                  switch (((message_header_t *) buffer)->type) {
                     case MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_MEDIUM:
                     case MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER:
                        debug_log(DEBUG_LEVEL_INFO,
                                  "Replying SET_SESSION (%s).\n",
                                  (((message_header_t *) buffer)->type
                               == MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER)
                                  ? "DUMPER"
                                  : "MEDIUM");
                        session
                           = scheduler_session_allocate
                                (info->scheduler_context,
                                 ((((message_header_t *) buffer)->type
                               == MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER)
                               ? (scheduler_session_union_t *)
                                 &((message_to_scheduler_set_session_dumper_t
                                    *) buffer)->dumper_info
                               : (scheduler_session_union_t *)
                                 &((message_to_scheduler_set_session_medium_t
                                    *) buffer)->medium_info),
                                 child,
                                 (((message_header_t *) buffer)->type
                               == MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER)
                                  ? SCHEDULER_SESSION_TYPE_DUMPER
                                  : SCHEDULER_SESSION_TYPE_MEDIUM,
                                 &error);

                        if (session) {
                           SESSION_OF_CHILD(child) = session;
                        }
                        else {
                           debug_log(DEBUG_LEVEL_WARN,
                                     "session allocation failed: 0x%x.\n",
                                     error);
                        }
   
                        ((message_from_scheduler_set_session_t *)
                            buffer)->header.type
                           = MESSAGE_TYPE_FROM_SCHEDULER_SET_SESSION;
                        ((message_from_scheduler_set_session_t *)
                            buffer)->error
                               = session ? session->error
                                         : error;
   
                        if (!process_father_send_message
                                (child,
                                 buffer,
                                 sizeof
                                    (message_from_scheduler_set_session_t))) {
                           debug_log(DEBUG_LEVEL_ERROR,
                                     "Can't send reply to 0x%x.\n",
                                     (int) child);
                        }
                        break;
                     case MESSAGE_TYPE_TO_SCHEDULER_GET_PART:
                        debug_log(DEBUG_LEVEL_INFO,
                                  "father GET_PART (child 0x%x, sess 0x%x).\n",
                                  child,
                                  SESSION_OF_CHILD(child));
                        should_reschedule
                           = should_reschedule
                             || scheduler_queue_request
                                   (info->scheduler_context,
                                    SESSION_OF_CHILD(child));
                        break;
                     case MESSAGE_TYPE_TO_SCHEDULER_FINISHED_PART:
                        session = SESSION_OF_CHILD(child);

                        if ((session->type == SCHEDULER_SESSION_TYPE_DUMPER)
                            && session->direction_to_cache) {
                           debug_log(DEBUG_LEVEL_INFO,
                                     "Dumper has created new part.\n");
                           memcpy(&session->md5sum,
                                  ((message_to_scheduler_finished_part_t *)
                                   buffer)->md5sum,
                                  sizeof(session->md5sum));
                        }

                        debug_log(DEBUG_LEVEL_INFO, "Set Finish part.\n");
                        should_reschedule
                           = scheduler_finished_part
                                (info->scheduler_context,
                                 SESSION_OF_CHILD(child),
                                 ((message_to_scheduler_finished_part_t *)
                                     (buffer))->part_size,
                                 ((message_to_scheduler_finished_part_t *)
                                     (buffer))->locator)
                             || should_reschedule; /* order matters */
   
                        /* For now reply */
                        {
                           message_from_scheduler_finished_part_t reply;
   
                           reply.header.type
                              = MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_PART;
                           reply.error = SESSION_OF_CHILD(child)
                                         ->error;
   
                           debug_log(DEBUG_LEVEL_INFO,
                                     "Replying FINISHED_PART.\n");
   
                           if (!process_father_send_message
                                   (child,
                                    (char *) &reply,
                                    sizeof(reply))) {
                              debug_log(DEBUG_LEVEL_ERROR,
                                        "Can't send FINISHED_PART reply "
                                        "to 0x%x.\n",
                                        (int) child);
                              /* @@@ */
                           }
                        }
                        break;
                     case MESSAGE_TYPE_TO_SCHEDULER_FINISHED_FILE:
                        should_reschedule
                           = scheduler_finished_file
                                (info->scheduler_context,
                                 SESSION_OF_CHILD(child),
                                 ((message_to_scheduler_finished_file_t *)
                                     (buffer))->succeeded)
                             || should_reschedule; /* order matters */
                        /* In any case */
                        {
                           message_from_scheduler_finished_file_t reply;

			   /* Send a successful completion message */
			   reply.header.type
                              = MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_FILE;
			   reply.error = SESSION_OF_CHILD(child)->error;

			   debug_log(DEBUG_LEVEL_INFO,
                                     "Replying FINISHED_FILE.\n");

			   if (!process_father_send_message
				   (SESSION_OF_CHILD(child)->child,
				    (char *) &reply,
				    sizeof(reply))) {
			      debug_log(DEBUG_LEVEL_ERROR,
					"Can't send FINISHED_FILE reply to "
                                        "0x%x.\n",
					(int) SESSION_OF_CHILD(child)->child);
			   }
                        }
                        break;
                     default:
                        debug_log(DEBUG_LEVEL_ERROR,
                                  "Unknown message type 0x%x",
                                  ((message_header_t *) buffer)->type);
                        /* @@@ abort child @@@ */
                        break;
                  }
               }
               else {
                   debug_log(DEBUG_LEVEL_ERROR,
                             "Error receiving msg from child 0x%x.\n",
                             (int) child);
               }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR,
                         "child 0x%x has died.\n",
                         (int) child);
               /* @@@ abort and then free child_data_t and session */
            }
         }
   
         if (should_reschedule) {
            scheduler_session_t *session = NULL;
            message_from_scheduler_get_part_t reply;
   
            while (scheduler_schedule(info->scheduler_context,
                                      &session,
                                      &reply.part_size,
                                      &reply.from_cache,
                                      &reply.locator,
                                      reply.part_file_name,
                                      sizeof(reply.part_file_name))) {
               /* This part is now available to the named child */
   
               reply.header.type = MESSAGE_TYPE_FROM_SCHEDULER_GET_PART;
               reply.error = session->error;
   
               memcpy(session->part_file_name,
                      reply.part_file_name,
                      sizeof(session->part_file_name));
   
               debug_log(DEBUG_LEVEL_INFO, "Replying GET_PART.\n");
   
               if (!process_father_send_message
                       (session->child,
                        (char *) &reply,
                        sizeof(reply))) {
                  debug_log(DEBUG_LEVEL_ERROR,
                            "Can't send GET_PART reply to 0x%x.\n",
                            (int) child);
                  /* @@@ mark part as unused ? @@@ */
               }
            }
         }
      }
      else if (retval == -1) {
         if (errno != EINTR) {
             debug_perror(DEBUG_LEVEL_ERROR, "select()");
             result = 1;
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "strange select() says 0.\n");
      }
   }
   
   /* @@@ kill all dumper children ? @@@ */

   return result;
}

static int handle_dumper(child_context_t child_context) {
   fd_t cd;
   int result = 1; /* failure */
   session_t session;
   client_initial_message_t initial_message;
   secret_item_t *item;

   session.child_context = child_context;
   session.child_data = (child_data_t *) process_get_private_of(child_context);
   cd = session.child_data->object.dumper.cd;

   /* should also close the sd @@@ */

   if (misc_all_read(cd,
                     (unsigned char *) &initial_message,
                     sizeof(initial_message))
       != sizeof(initial_message)) {
      return 1;
   }

   initial_message.client_name
      [sizeof(initial_message.client_name) - 1] = '\0';
   initial_message.server_name
      [sizeof(initial_message.server_name) - 1] = '\0';

   if (initial_message.protocol_version
       != CLIENT_HELLO_PROTOCOL_VERSION) {
      debug_log(DEBUG_LEVEL_ERROR, "unknown protocol %d.\n",
                initial_message.protocol_version);
      return 1;
   }

   if (strcmp(initial_message.server_name,
              session.child_data->object.dumper.server_name)) {
      debug_log(DEBUG_LEVEL_ERROR, "unknown server %s.\n",
                initial_message.server_name);
      return 1;
   }

   debug_log(DEBUG_LEVEL_INFO, "PROTOCOL: %d\nclient: %s\nserver: %s\n",
             initial_message.protocol_version,
             initial_message.client_name,
             initial_message.server_name);

   item = find_secret_item(session.child_data->object.dumper.secret_list,
                           initial_message.client_name);
   if (item == NULL) {
      debug_log(DEBUG_LEVEL_ERROR, "no such client.\n");
      return 1;
   }

   /* Verify this is a legitimate client */
   if (!auth_random_cleartext_challenge_issue_and_verify
           (cd,
            item->client_password,
            strlen(item->client_password))) {
      debug_log(DEBUG_LEVEL_ERROR, "client doesn't know its key.\n");
      return 1;
   }

   if (!auth_challenge_receive_and_respond
        (cd,
         item->server_password,
         strlen(item->server_password))) {
      debug_log(DEBUG_LEVEL_ERROR, "client doesn't know my key, or error.\n");
      return 1;
   }

   debug_log(DEBUG_LEVEL_INFO, "CLIENT-SERVER authentification succeeded.\n");

   /* real starts here */
   session.protocol_info.decrypt_context_1 = NULL;
   session.protocol_info.encrypt_context_1 = NULL;
   session.protocol_info.decrypt_context_2 = NULL;
   session.protocol_info.encrypt_context_2 = NULL;
   
   session.protocol_info.decrypt_initial_1 = TRUE;
   session.protocol_info.decrypt_initial_2 = TRUE;
   session.protocol_info.encrypt_initial_1 = TRUE;
   session.protocol_info.encrypt_initial_2 = TRUE;
   
   if ((session.protocol_info.decrypt_context_1 = crypt_context_allocate())
       && (session.protocol_info.encrypt_context_1 = crypt_context_allocate())
       && (session.protocol_info.decrypt_context_2 = crypt_context_allocate())
       && (session.protocol_info.encrypt_context_2
           = crypt_context_allocate())) {
      unsigned char random_crypt_key[RANDOM_CRYPT_KEY_SIZE + 1]; /* \0 */

      if (misc_get_random_data(random_crypt_key,
                               sizeof(random_crypt_key) - 1)) {
         int hack = 0;

         random_crypt_key[sizeof(random_crypt_key) - 1] = '\0';

         /* @@@ beurk: key can't be nulterm @@@ */
         for (hack = 0; hack < (sizeof(random_crypt_key) - 1); hack++) {
            if (!random_crypt_key[hack]) {
               random_crypt_key[hack] = 1;
            }
         }

         if (crypt_context_initialize(session.protocol_info.decrypt_context_1,
                                      random_crypt_key,
                                      sizeof(random_crypt_key) - 1)
             && crypt_context_initialize(session.protocol_info
                                            .encrypt_context_1,
                                         random_crypt_key,
                                         sizeof(random_crypt_key) - 1)) {
            if (protocol_send_encrypted_session_key
                   (cd,
                    random_crypt_key,
                    sizeof(random_crypt_key) - 1,
                    item->server_password,
                    strlen(item->server_password))) {
	       if (misc_get_random_data(random_crypt_key,
					sizeof(random_crypt_key) - 1)) {
		  random_crypt_key[sizeof(random_crypt_key) - 1] = '\0';

		  /* @@@ beurk: key can't be nulterm @@@ */
		  for (hack = 0;
                       hack < (sizeof(random_crypt_key) - 1);
                       hack++) {
		     if (!random_crypt_key[hack]) {
			random_crypt_key[hack] = 1;
		     }
		  }

		  if (crypt_context_initialize(session.protocol_info
                                                  .decrypt_context_2,
					       random_crypt_key,
					       sizeof(random_crypt_key) - 1)
		      && crypt_context_initialize(session.protocol_info
						     .encrypt_context_2,
						  random_crypt_key,
						  sizeof(random_crypt_key)
                                                  - 1)) {
		     if (protocol_send_encrypted_session_key
			    (cd,
			     random_crypt_key,
			     sizeof(random_crypt_key) - 1,
			     item->client_password,
			     strlen(item->client_password))) {
			debug_log(DEBUG_LEVEL_INFO,
                                  "SESSION ENCRYPTION KEYS SET.\n");

			free_secret_list
                           (session.child_data->object.dumper.secret_list);
			session.child_data->object.dumper.secret_list = NULL;

			result = do_operation(&session, cd, &initial_message);
		     }
		     else {
			debug_log(DEBUG_LEVEL_ERROR,
                                  "can't send session key 2.\n");
		     }
		  }
		  else {
		     debug_log(DEBUG_LEVEL_ERROR, "weak key 2.\n");
		  }
	       }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR, "can't get random data 2.\n");
	       }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR, "can't send session key.\n");
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR, "weak key.\n");
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "can't get random data.\n");
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "no crypt context.\n");
   }  

   if (close(cd)) {
      debug_perror(DEBUG_LEVEL_ERROR, "close()");
   }

   if (session.protocol_info.decrypt_context_1) {
      crypt_context_free(&(session.protocol_info.decrypt_context_1));
      if (session.protocol_info.encrypt_context_1) {
         crypt_context_free(&(session.protocol_info.encrypt_context_1));
         if (session.protocol_info.decrypt_context_2) {
            crypt_context_free(&(session.protocol_info.decrypt_context_2));
            if (session.protocol_info.encrypt_context_2) {
               crypt_context_free(&(session.protocol_info.encrypt_context_2));
            }
         }
      }
   }

   return result;
}

static int do_operation(session_t *session,
                        fd_t sd,
                        client_initial_message_t *initial_message) {
   if (protocol_receive_data(sd,
			     &session->protocol_info,
			     (unsigned char *) &session->hello,
			     sizeof(session->hello),
			     TRUE)) {
      int result = 0; /* succeeded */

      session->hello.file_name[sizeof(session->hello.file_name) - 1] = '\0';
      session->hello.reference_file_name
         [sizeof(session->hello.reference_file_name) - 1] = '\0';

      /* Transfer information to session info */
      
      memcpy(&session->hello.client_name,
             &initial_message->client_name,
             sizeof(session->hello.client_name));
      
      memcpy(&session->hello.server_name,
             &initial_message->server_name,
             sizeof(session->hello.server_name));

      debug_log(DEBUG_LEVEL_INFO,
		"file: %s\nref: %s\n"
		"flags: 0x%x\ndirection: %s\n"
		"block_size: %ld\nblocks: %ld\noffset: %ld\n",
		session->hello.file_name,
		session->hello.reference_file_name,
		session->hello.flags,
		(session->hello.direction == DIRECTION_TO_CACHE)
		   ? "FROM_CLIENT"
		   : "TO_CLIENT",
		session->hello.block_size,
		session->hello.blocks,
		session->hello.offset);

      if (!encode_scheduler_set_session_dumper(session)) {
         debug_log(DEBUG_LEVEL_ERROR, "error setting scheduler session.\n");
         return 1;
      }

      debug_log(DEBUG_LEVEL_INFO, "session: 0x%x.\n", session);

      session->protocol_info.user_defined = (void *) session;
      session->had_a_file = FALSE;

      if (session->hello.block_size
          && (session->hello.block_size <= MAX_BLOCK_SIZE)) {
         if (session->hello.direction == DIRECTION_TO_CACHE) {
            unsigned char md5_hash[MD5_HASH_STRING_LENGTH];
            /* This is the whole file checksum
	     * TODO
	     *    - do something useful with it, such as store in
	     *      the file structure.
	     */

            debug_log(DEBUG_LEVEL_INFO, "to server ...\n");	
	 
            if (!filesend_receive_file(sd,
                                       (void *) &session->protocol_info,
                                       protocol_decrypt,
                                       get_next_file_write,
                                       md5_hash)) {
               result = 1;
               /* Could be e.g. bad md5sum. Scheduler should normally discard
                * the whole file.
                */
               debug_log(DEBUG_LEVEL_ERROR,
                         "filesend_receive_file() failed.\n");
            }
            else {
               debug_log(DEBUG_LEVEL_INFO, "SUCCEEDED\n");
            }
         }
         else {
            debug_log(DEBUG_LEVEL_INFO, "from server\n");

            /* TODO
             *    - the stored db md5sum should really be verified somehow,
             *      by filesend_send_file().
             */	 
            if (!filesend_send_file(sd,
                                    (void *) &session->protocol_info,
                                    protocol_encrypt,
                                    get_next_file_read)) {
               result = 1;
               debug_log(DEBUG_LEVEL_ERROR,
                         "filesend_send_file() failed.\n");
            }
            else {
               debug_log(DEBUG_LEVEL_INFO, "SUCCEEDED\n");
            }
         }

         if (!encode_scheduler_finished_file(session, (result == 0))) {
            debug_log(DEBUG_LEVEL_ERROR,
                      "scheduler_finished_file() failed.\n");
            result = 1;
         }

         return result;
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "block size %ld too big.\n",
                   session->hello.block_size);
         result = 1;
      }

      return result;
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "error receiving session->hello structure.\n");
      return 1;
   }
   /* NOT REACHED */
}

static BOOL load_configuration(const char *config_file_name,
                               server_config_t *config) {
   static const configuration_description_t configuration_array[]
      = {
	   {
	      FALSE,
	      {
		 {
		    "server_port",
		    CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER
		 },
		 { 0 }
	      }
	   },
	   {
	      FALSE,
	      {
		 {
		    "max_part_size",
		    CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER
		 },
		 { 0 }
	      }
	   },
	   {
	      FALSE,
	      {
		 {
		    "max_running_dumpers",
		    CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER
		 },
		 { 0 }
	      }
	   },
	   {
	      FALSE,
	      {
		 {
		    "server_name",
		    CONFIGURATION_ITEM_TYPE_STRING
		 },
		 { "" }
	      }
	   },
	   {
	      FALSE,
	      {
		 {
		    "spool_directory",
		    CONFIGURATION_ITEM_TYPE_STRING
		 },
		 { "" }
	      }
	   },
	   {
	      TRUE,
	      {
		 {
		    "medium_configurations",
		    CONFIGURATION_ITEM_TYPE_STRING
		 },
		 { "" }
	      }
	   }
	};
   configuration_context_t context;

   context = configuration_initialize(&configuration_array[0],
                                      sizeof(configuration_array)
                                      / sizeof(configuration_array[0]));

   if (context) {
      if (configuration_load(context, config_file_name)) {
         configuration_item_t item;

         item.type.type = CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER;
         item.type.name = "server_port";
         if (configuration_get(context, &item)) {
            config->port = item.object.unsigned_integer;

	    item.type.type = CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER;
	    item.type.name = "max_part_size";
	    if (configuration_get(context, &item)) {
	       config->max_part_size = item.object.unsigned_integer;

	       item.type.type = CONFIGURATION_ITEM_TYPE_UNSIGNED_INTEGER;
	       item.type.name = "max_running_dumpers";
	       if (configuration_get(context, &item)) {
		  config->max_running_dumpers = item.object.unsigned_integer;

                  item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                  item.type.name = "server_name";
                  if (configuration_get(context, &item)) {
                     config->server_name = strdup(item.object.string);
                     if (config->server_name) {
                        item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                        item.type.name = "spool_directory";
                        if (configuration_get(context, &item)) {
                           config->spool_directory
                              = strdup(item.object.string);
                           if (config->spool_directory) {
                              item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
                              item.type.name = "medium_configurations";
                              if (configuration_get(context, &item)) {
                                 config->medium_configurations
                                    = strdup(item.object.string);
                                 if (config->medium_configurations) {
                                    return TRUE;
                                 }
                                 else {
                                    free(config->server_name);
                                    free(config->spool_directory);
                                    debug_log(DEBUG_LEVEL_ERROR,
                                              "no memory for "
                                              "medium_configs.\n");
                                 }
                              }
                              else {
                                 config->medium_configurations = NULL;
                              }
                           }
                           else {
                              free(config->server_name);
                              debug_log(DEBUG_LEVEL_ERROR,
                                        "no memory for spool_directory.\n");
                           }
                        }
                        else {
                           free(config->server_name);
                           debug_log(DEBUG_LEVEL_ERROR,
                                     "missing spool_directory in %s.\n",
                                     config_file_name);
                        }
                     }
                     else {
                        debug_log(DEBUG_LEVEL_ERROR,
                                  "no memory for server_name.\n");
                     }
                  }
		  else {
		     debug_log(DEBUG_LEVEL_ERROR,
                               "missing server_name in %s.\n",
			       config_file_name);
		  }
	       }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR,
                            "missing max_running_dumpers in %s.\n",
			    config_file_name);
	       }
            }
            else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "missing max_part_size in %s.\n",
			 config_file_name);
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "missing server_port in %s.\n",
                      config_file_name);
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "can't load configuration from file %s.\n",
                   config_file_name);
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "configuration module initialization failed.\n");
   }

   return FALSE;
}

static BOOL load_secret_file(list_head_t *list,
                             const char *secret_file_name) {
   BOOL result = TRUE;
   FILE *f;

   list_initialize(list);

   f = fopen(secret_file_name, "r");
   if (f) {
      unsigned char buffer[MAX_LINE_LENGTH];

      do {
         if (fgets(buffer, sizeof(buffer), f) != NULL) {
            if (!parse_line(list, buffer, sizeof(buffer))) {
               result = FALSE;
            }
         }
      } while (!feof(f) && !ferror(f) && result);

      if (ferror(f)) {
         debug_perror(DEBUG_LEVEL_ERROR, "stdio error");
         result = FALSE;
      }

      if (fclose(f)) {
         debug_perror(DEBUG_LEVEL_ERROR, "fclose()");
         result = FALSE;   
      }

      f = NULL;
   }
   else {
      debug_perror(DEBUG_LEVEL_ERROR, "fopen()");
      result = FALSE;
   }

   if (!result) {
      free_secret_list(list);
   }

   return result;   
}

static void free_secret_list(list_head_t *list) {
   list_item_header_t *current = list_next(list, NULL);

   /* Could use iterator, but way easier that way. */

   while (current) {
      secret_item_t *item = LIST_ITEM_TO_SECRET(current);
      list_item_header_t *next = list_next(list, current);

      list_remove_item(list, current);
      current = next;

      free_secret_item(item);
   }
}

static BOOL parse_line(list_head_t *list,
                       unsigned char *buffer,
                       int buffer_length) {
   BOOL result = TRUE;

   misc_delete_spaces(&buffer);

   /* Only supports comments at start of line (possibly after spaces) */
   if (buffer[0] == '#') {
      buffer[0] = '\0';
   }

   if (buffer[0]) {
      unsigned char *client
         = strsep((char **) &buffer, " "); /* only accepts 0x20 */   

      result = FALSE;

      if (client) {
         misc_delete_spaces(&client);
         if (client[0]) {
            unsigned char *client_password
               = strsep((char **) &buffer, " "); /* only accepts 0x20 */   

            if (client_password) {
               misc_delete_spaces(&client_password);

	       if (client_password[0]) {
		  unsigned char *server_password
		     = strsep((char **) &buffer,
                              " "); /* only accepts 0x20 */   

		  if (server_password) {
		     misc_delete_spaces(&server_password);

                     if (server_password[0]) {
                        result
                           = add_secret_item(list,
                                             (const char *) client,
                                             (const char *) server_password,
                                             (const char *) client_password);
                     }
		  }
	       }
            }
         }
      }
   }

   return result;
}

static BOOL add_secret_item(list_head_t *list,
                            const char *client,
                            const char *server_password,
                            const char *client_password) {
   if (find_secret_item(list, client) == NULL) {
      secret_item_t *item = (secret_item_t *) malloc(sizeof(secret_item_t));

      item->client = strdup(client);
      item->server_password = strdup(server_password);
      item->client_password = strdup(client_password);

      if (item->client && item->server_password && item->client_password) {
	 list_append_tail(list, &item->list_header);

	 return TRUE;
      }
      else {
	 free_secret_item(item);

	 return FALSE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "duplicate secret for %s.\n", client);
      return FALSE;
   }
}

static void free_secret_item(secret_item_t *item) {
   if (item->client) {
      free(item->client);
   }

   if (item->server_password) {
      free(item->server_password);
   }

   if (item->client_password) {
      free(item->client_password);
   }

   free(item);
}

static secret_item_t *find_secret_item(list_head_t *list, const char *client) {
   list_item_header_t *current = NULL;

   /* Could use iterator, but way easier that way. */

   while ((current = list_next(list, current))) {
      secret_item_t *item = LIST_ITEM_TO_SECRET(current);

      if (!strcmp(client, item->client)) {
         return item;
      }
   }

   return NULL;
}

static BOOL encode_scheduler_set_session_dumper(session_t *session) {
   message_to_scheduler_set_session_dumper_t message;

   debug_log(DEBUG_LEVEL_INFO, "Setting session (DUMPER).\n");

   message.header.type = MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_DUMPER;
   memcpy(&message.dumper_info, &session->hello, sizeof(message.dumper_info));

   if (process_child_send_message(session->child_context,
                                  (unsigned char *) &message,
                                  sizeof(message))) {
      message_from_scheduler_set_session_t message;
      int message_length = sizeof(message);

      if (process_child_receive_message_blocking(session->child_context,
						 (unsigned char *) &message,
						 &message_length,
                                                 FALSE)) {
         if (message.header.type
             == MESSAGE_TYPE_FROM_SCHEDULER_SET_SESSION) {
            if (message.error == SCHEDULER_ERROR_NONE) {
               debug_log(DEBUG_LEVEL_INFO, "SET_SESSION ok.\n");
               return TRUE;
            }
 	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "Scheduler error 0x%x.\n",
                         message.error);
	    }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Incorrect type 0x%x.\n",
                      message.header.type);
         }
      }
      else {
	 debug_log(DEBUG_LEVEL_ERROR,
                   "Setting session: message couldn't be received.\n");
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "Setting session: message couldn't be sent.\n");
   }

   return FALSE;
}

static BOOL encode_scheduler_get_part(session_t *session,
                                      size_t *part_size,
                                      char *part_file_name,
                                      int part_file_name_length,
                                      BOOL *eof) {
   message_to_scheduler_get_part_t message;
   message_from_scheduler_get_part_t reply;

   if (eof) {
      *eof = FALSE;
   }

   debug_log(DEBUG_LEVEL_INFO, "in encode_scheduler_get_part()\n");

   if (part_file_name_length < sizeof(reply.part_file_name)) {
      debug_log(DEBUG_LEVEL_ERROR,
                "encode_scheduler_get_part: too small length.\n");
      return FALSE;
   }
   else {
      message.header.type = MESSAGE_TYPE_TO_SCHEDULER_GET_PART;

      debug_log(DEBUG_LEVEL_INFO, "in encode_scheduler_get_part 2()\n");

      debug_log(DEBUG_LEVEL_INFO, "context: 0x%x.\n", session->child_context);

      if (process_child_send_message(session->child_context,
				     (unsigned char *) &message,
				     sizeof(message))) {
	 int message_length = sizeof(reply);

         debug_log(DEBUG_LEVEL_INFO, "in encode_scheduler_get_part 3()\n");

	 if (process_child_receive_message_blocking(session->child_context,
						    (unsigned char *) &reply,
						    &message_length,
                                                    FALSE)) {
            debug_log(DEBUG_LEVEL_INFO, "in encode_scheduler_get_part 4()\n");
	    if (reply.header.type == MESSAGE_TYPE_FROM_SCHEDULER_GET_PART) {
	       if (reply.error == SCHEDULER_ERROR_NONE) {
		  memcpy(part_file_name,
			 reply.part_file_name,
			 sizeof(reply.part_file_name));
		  part_file_name[sizeof(reply.part_file_name) - 1] = '\0';
		  *part_size = reply.part_size;

		  debug_log(DEBUG_LEVEL_INFO, "GET_PART ok.\n");

		  return TRUE;
	       }
               else if (eof && (reply.error == SCHEDULER_ERROR_EOF)) {
                  debug_log(DEBUG_LEVEL_INFO,
                            "End of file.\n");
                  /* Assuming read, else it's not right */
                  *eof = TRUE;
                  return FALSE;
               }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR,
                            "Scheduler error 0x%x.\n",
                            reply.error);
	       }
	    }
	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "Incorrect type 0x%x.\n",
                         reply.header.type);
	    }
            debug_log(DEBUG_LEVEL_INFO, "in encode_scheduler_get_part 5()\n");
	 }
      }
   }

   debug_log(DEBUG_LEVEL_INFO, "out encode_scheduler_get_part()\n");   

   return FALSE;
}

static BOOL encode_scheduler_finished_part(session_t *session,
                                           size_t part_size,
                                           locator_t locator,
                                           unsigned char *md5sum) {
   message_to_scheduler_finished_part_t message;

   debug_log(DEBUG_LEVEL_INFO, "Finishing part.\n");

   message.header.type = MESSAGE_TYPE_TO_SCHEDULER_FINISHED_PART;
   message.part_size = part_size;
   memcpy(&message.md5sum,
          md5sum,
          sizeof(message.md5sum));

   if (process_child_send_message(session->child_context,
                                  (unsigned char *) &message,
                                  sizeof(message))) {
      message_from_scheduler_finished_part_t message;
      int message_length = sizeof(message);

      if (process_child_receive_message_blocking(session->child_context,
						 (unsigned char *) &message,
						 &message_length,
                                                 FALSE)) {
         if (message.header.type
             == MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_PART) {
            if (message.error == SCHEDULER_ERROR_NONE) {
               debug_log(DEBUG_LEVEL_INFO, "FINISH_PART ok.\n");
               return TRUE;
            }
 	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "Scheduler error 0x%x.\n",
                          message.error);
	    }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Incorrect type 0x%x.\n",
                      message.header.type);
         }
      }
   }

   return FALSE;
}

static BOOL encode_scheduler_finished_file(session_t *session,
                                           BOOL succeeded) {
   message_to_scheduler_finished_file_t message;

   message.header.type = MESSAGE_TYPE_TO_SCHEDULER_FINISHED_FILE;
   message.succeeded = succeeded;

   debug_log(DEBUG_LEVEL_INFO, "finishing file %c.\n", succeeded ? 'T' : 'F');

   if (process_child_send_message(session->child_context,
                                  (unsigned char *) &message,
                                  sizeof(message))) {
      message_from_scheduler_finished_file_t message;
      int message_length = sizeof(message);

      if (process_child_receive_message_blocking(session->child_context,
						 (unsigned char *) &message,
						 &message_length,
                                                 FALSE)) {
         if (message.header.type
             == MESSAGE_TYPE_FROM_SCHEDULER_FINISHED_FILE) {
            if (message.error == SCHEDULER_ERROR_NONE) {
               debug_log(DEBUG_LEVEL_INFO, "FINISH_FILE ok.\n");
               return TRUE;
            }
 	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "Scheduler error 0x%x.\n",
                         message.error);
	    }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Incorrect type 0x%x.\n",
                      message.header.type);
         }
      }
   }

   return FALSE;
}

static BOOL get_next_file_read(void *caller_private,
                               fd_t *fd,
                               md5_hash_ptr_t last_file_md5sum,
                               BOOL *eof) {
   char part_file_name[MAX_PATHNAME_LENGTH];
   protocol_info_t *protocol_info = (protocol_info_t *) caller_private;
   session_t *session = (session_t *) (protocol_info->user_defined);

   *eof = FALSE;

   debug_log(DEBUG_LEVEL_INFO, "in get_next_file_read()\n");
   debug_log(DEBUG_LEVEL_INFO, "session: 0x%x.\n", session);

   if (session->had_a_file) {
      if (!encode_scheduler_finished_part(session,
                                          session->current_part_size,
                                          LOCATOR_INVALID,
                                          last_file_md5sum)) {
         debug_log(DEBUG_LEVEL_ERROR,
                   "scheduler_finished_part() failed.\n");
         *fd = -1;
         return FALSE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_INFO, "No file to commit\n");
   }

   debug_log(DEBUG_LEVEL_INFO, "in get_next_file_read() 2\n");

   *fd = -1;
   if (encode_scheduler_get_part(session,
                                 &session->current_part_size,
                                 part_file_name,
                                 sizeof(part_file_name),
                                 eof)) {
      debug_log(DEBUG_LEVEL_INFO, "in get_next_file_read() 2a\n");

      *fd = open(part_file_name, O_RDONLY);
      if (*fd == -1) {
         debug_perror(DEBUG_LEVEL_ERROR, "open(RO)");
      }
      else {
         debug_log(DEBUG_LEVEL_INFO, "(file %s open).\n", part_file_name);
         session->had_a_file = TRUE;
      }
      debug_log(DEBUG_LEVEL_INFO, "in get_next_file_read() 2b\n");
   }
   else if (!*eof) {
      debug_log(DEBUG_LEVEL_ERROR,
		"scheduler_get_part() failed.\n");
   }

   debug_log(DEBUG_LEVEL_INFO, "out get_next_file_read()\n");

   return (*fd != -1);
}

static BOOL get_next_file_write(void *caller_private,
				fd_t *fd,
				unsigned int *file_max,
				md5_hash_ptr_t last_file_md5sum,
				const BOOL last) {
   char part_file_name[MAX_PATHNAME_LENGTH];
   protocol_info_t *protocol_info = (protocol_info_t *) caller_private;
   session_t *session = (session_t *) (protocol_info->user_defined);

   debug_log(DEBUG_LEVEL_INFO, "in get_next_file_write()\n");
   debug_log(DEBUG_LEVEL_INFO, "session: 0x%x.\n", session);

   if (last) {
      session->current_part_size = *file_max;
   }

   if (session->had_a_file) {
      if (!encode_scheduler_finished_part(session,
                                          session->current_part_size,
                                          LOCATOR_INVALID,
                                          last_file_md5sum)) {
         debug_log(DEBUG_LEVEL_ERROR,
                   "scheduler_finished_part() failed.\n");
         *fd = -1;
         return FALSE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_INFO, "No file to commit\n");
   }
       
   *fd = -1;

   if (last) {
      printf("---LAST CALLED---\n");
      return TRUE; /* special */
   }
   else {
      if (encode_scheduler_get_part(session,
				    &session->current_part_size,
				    part_file_name,
				    sizeof(part_file_name),
                                    NULL)) {
	 *fd = open(part_file_name, O_CREAT | O_EXCL | O_WRONLY, 0600);
	 if (*fd == -1) {
	    debug_perror(DEBUG_LEVEL_ERROR, "open(CR)");
	 }
	 else {
	    session->had_a_file = TRUE;
	 }
      }
      else {
	 debug_log(DEBUG_LEVEL_ERROR,
		   "scheduler_get_part() failed.\n");
      }

      debug_log(DEBUG_LEVEL_INFO, "out get_next_file_write()\n");

      *file_max = session->current_part_size;
      return (*fd != -1);
   }
}

static BOOL launch_medium_if_any(server_info_t *info) {
   if (info->config.medium_configurations[0]) {
      char *medium_configurations
         = strdup(info->config.medium_configurations);

      if (medium_configurations) {
         BOOL result = TRUE;
         char *current;
         char *next = medium_configurations;

	 debug_log(DEBUG_LEVEL_INFO,
		   "Medium configurations: %s.\n",
		   medium_configurations);

         while (result
                && (current = strsep(&next, " "))) { /* only accepts 0x20 */
            char *medium_device_name = NULL;
            char *changer_device_name = NULL;

            result = medium_load_config(info,
                                        current,
                                        &medium_device_name,
                                        &changer_device_name);
            if (result) {
               result = medium_launch(info,
                                      medium_device_name,
                                      changer_device_name);
               if (result) {
                  debug_log(DEBUG_LEVEL_INFO,
                            "Medium launch succeeded for %s.\n",
                            current);
               }
               else {
                  debug_log(DEBUG_LEVEL_ERROR,
                            "Medium launch failed for %s.\n",
                            current);
               }

               free(medium_device_name);
               free(changer_device_name);
            }
            else {
	       debug_log(DEBUG_LEVEL_ERROR,
                         "Can't load configuration %s.\n",
                         current);
	       result = FALSE;
            }
         }

         free(medium_configurations);

         return result;
      }
      else {
         debug_log(DEBUG_LEVEL_FATAL, "Can't strdup() medium_configurations\n");
         return FALSE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_INFO, "No medium configuration.\n");
   }

   return TRUE;
}

static BOOL medium_load_config(server_info_t *info,
                               const char *current,
                               char **medium_device_name,
                               char **changer_device_name) {
   static const configuration_description_t configuration_array[]
      = {
	   {
	      FALSE,
	      {
		 {
		    "medium_device_name",
		    CONFIGURATION_ITEM_TYPE_STRING
		 },
		 { "" }
	      }
	   },
	   {
	      TRUE, /* not compulsory */
	      {
		 {
		    "changer_device_name",
		    CONFIGURATION_ITEM_TYPE_STRING
		 },
		 { "" }
	      }
	   }
	};
   configuration_context_t context;

   context = configuration_initialize(&configuration_array[0],
                                      sizeof(configuration_array)
                                      / sizeof(configuration_array[0]));

   if (context) {
      if (configuration_load(context, current)) {
         configuration_item_t item;

         item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
         item.type.name = "medium_device_name";
         if (configuration_get(context, &item)) {
            *medium_device_name = strdup(item.object.string);
            if (*medium_device_name) {
               item.type.type = CONFIGURATION_ITEM_TYPE_STRING;
               item.type.name = "changer_device_name";
               if (configuration_get(context, &item)) {
                  *changer_device_name = strdup(item.object.string);
                  if (*changer_device_name) {
                     return TRUE;
                  }
                  else {
                     free(*medium_device_name);
                     *medium_device_name = NULL;
                     debug_log(DEBUG_LEVEL_ERROR,
                               "no memory for changer_device_name.\n");
                  }
               }
               else {
                  free(*medium_device_name);
                  *medium_device_name = NULL;
                  debug_log(DEBUG_LEVEL_ERROR,
                            "missing changer_device_name in %s.\n",
                            current);
               }
            }
            else {
               debug_log(DEBUG_LEVEL_ERROR, "no memory for tapedevname.\n");
            }
	 }
	 else {
	    debug_log(DEBUG_LEVEL_ERROR, "missing tapedevname in %s.\n",
	 		 current);
	 }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "can't load configuration from file %s.\n",
                   current);
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "configuration module initialization failed.\n");
   }

   return FALSE;
}

static BOOL medium_launch(server_info_t *info,
                          const char *medium_device_name,
                          const char *changer_device_name) {
   BOOL result = FALSE; /* failed */
   child_data_t *child_data = (child_data_t *) malloc(sizeof(child_data_t));

   if (child_data) {
      child_context_t child_context;

      child_data->session = NULL;
      child_data->object.medium.medium_device_name = medium_device_name;
      child_data->object.medium.changer_device_name = changer_device_name;

      debug_log(DEBUG_LEVEL_INFO,
		"Launching (%s, %s).\n",
		medium_device_name,
		changer_device_name);

      child_context = process_launch(info->process_context,
				     handle_medium,
				     (void *) child_data);
      if (child_context) {
	 result = TRUE;

	 /* NOTES
	  *    - The child, when properly started, will ask us to set
	  *      a session, in a way similar to the dumpers.
	  *      This is done at poll time: failure of one or even all tape
	  *      devices shouldn't prevent us from starting the server and
	  *      do dump to disk only). This is done in main loop.
	  */

	 debug_log(DEBUG_LEVEL_INFO,
		   "Medium child 0x%x started.\n",
		   child_context);         
      }
      else {
	 debug_log(DEBUG_LEVEL_ERROR,
		   "Can't launch (%s, %s): process creation failed.\n",
		   medium_device_name,
		   changer_device_name);
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "Can't launch (%s, %s): malloc failed.\n",
                medium_device_name,
                changer_device_name);
   }
   
   return result;
}

static void stop_all_launched_children(server_info_t *info) {
   debug_log(DEBUG_LEVEL_FATAL,
             "stop_all_launched_children() not yet implemented.\n");
}

static int handle_medium(child_context_t child_context) {
   medium_data_t *medium_data = &((child_data_t *)
                                     process_get_private_of(child_context))
                                    ->object.medium;
   int result = 1; /* FAILED */
   changer_context_t changer_context;

   /* BUGS
    *    - For now we will create a changer context for each medium
    *      context. That's wrong: main should do the work, see other
    *      comments.
    */

#if 0
   changer_context = changer_initialize(medium_data->changer_device_name);
   if (changer_context) {
#else
   changer_context = NULL;
   if (1) {
#endif
      taper_context_t taper_context;

      /* TODO
       *    - This function should lock the drive, and only an EJECT
       *      command from main could unlock it.
       */
      taper_context = taper_initialize(medium_data->medium_device_name,
				       changer_context);
      if (taper_context) {
	 message_to_scheduler_set_session_medium_t message;

	 debug_log(DEBUG_LEVEL_INFO, "taper: Taper open.\n");      

	 debug_log(DEBUG_LEVEL_INFO, "taper: Setting session (MEDIUM).\n");

	 message.header.type = MESSAGE_TYPE_TO_SCHEDULER_SET_SESSION_MEDIUM;
         /* @@@ fill message.medium_info @@@ */

	 if (process_child_send_message(child_context,
					(unsigned char *) &message,
					sizeof(message))) {
	    message_from_scheduler_set_session_t message;
	    int message_length = sizeof(message);

	    if (process_child_receive_message_blocking(child_context,
						       (unsigned char *)
                                                          &message,
						       &message_length,
                                                       FALSE)) {
	       if (message.header.type
		   == MESSAGE_TYPE_FROM_SCHEDULER_SET_SESSION) {
		  if (message.error == SCHEDULER_ERROR_NONE) {
                     BOOL shut_down = FALSE;

		     debug_log(DEBUG_LEVEL_INFO, "taper: SET_SESSION ok.\n");

                     while (!shut_down) {
                        message_to_scheduler_get_part_t message;

                        /* NOTES
                         *    - We issue a GET_PART, to which the scheduler
                         *      can reply with a part to read or write, or
                         *      an operation.
                         */

                        message.header.type
                           = MESSAGE_TYPE_TO_SCHEDULER_GET_PART;
                        if (process_child_send_message
                               (child_context,
                                (unsigned char *) &message,
                                sizeof(message))) {
                           unsigned char reply
                              [MAX(sizeof(message_from_scheduler_get_part_t),
                                   sizeof
                                   (message_type_from_scheduler_medium_op_t))];
    	                   int message_length = sizeof(reply);

			   debug_log(DEBUG_LEVEL_INFO,
				     "taper: waiting for command/part.\n");

			   if (process_child_receive_message_blocking
				  (child_context,
				   (unsigned char *) &reply,
				   &message_length,
                                   TRUE /* accepting smaller sizes too */)) {
                              message_type_t type
                                 = ((message_header_t *) &reply[0])->type;
                                  
			      if ((type
 				   == MESSAGE_TYPE_FROM_SCHEDULER_GET_PART)
                                  && (message_length
                                      ==
                                      sizeof
                                        (message_from_scheduler_get_part_t))) {
				 debug_log(DEBUG_LEVEL_INFO,
                                           "taper: Got cmd/part ###UNIMPL.\n");
			      }
                              else if ((type
                                       == MESSAGE_TYPE_FROM_SCHEDULER_MEDIUM_OP)
                                       && (message_length
                                           ==
                                           sizeof(message_type_from_scheduler_medium_op_t))) {
                                 message_type_from_scheduler_medium_op_t *msg;

                                 msg
                                  = (message_type_from_scheduler_medium_op_t *)
                                       &reply[0];
 
                                 debug_log(DEBUG_LEVEL_INFO,
                                           "taper: got op 0x%x.\n",
                                           msg->op);
                              }
			      else {
				 debug_log(DEBUG_LEVEL_ERROR,
					   "taper: Incorrect type/len 0x%x/%u"
                                           ".\n",
					   message.header.type,
                                           message_length);
				 shut_down = TRUE;
			      }
                           }
                           else {
			      debug_log(DEBUG_LEVEL_ERROR,
					"taper: GET_PART: message couldn't be "
					"received.\n");
			      shut_down = TRUE;
                           }
                        }
			else {
			   debug_log(DEBUG_LEVEL_ERROR,
				     "taper: GET_PART: message couldn't be "
                                     "sent.\n");
                           shut_down = TRUE;
			}                        
                     }

                     debug_log(DEBUG_LEVEL_INFO,
                               "taper: taper shutting down.\n");

		     result = 0;
		  }
		  else {
		     debug_log(DEBUG_LEVEL_ERROR,
			       "taper: Scheduler error 0x%x.\n",
			       message.error);
		  }
	       }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR,
			    "taper: Incorrect type 0x%x.\n",
			    message.header.type);
	       }
	    }
	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
			 "taper: Setting session: message couldn't be "
                         "received.\n");
	    }
	 }
	 else {
	    debug_log(DEBUG_LEVEL_ERROR,
		      "taper: Setting session: message couldn't be sent.\n");
	 }

	 taper_deinitialize(&taper_context);
      }
      else {
	 debug_log(DEBUG_LEVEL_ERROR, "taper: Can't initialize taper.\n");
      }

#if 0
      changer_deinitialize(&changer_context);
#endif
   }
   else {
      debug_log(DEBUG_LEVEL_FATAL,
                "taper: Can't initialize changer device %s.\n",
                medium_data->changer_device_name);
   }
   
   return result;
}
