/* Project filed -- Intelligent file cache and migration system
 * taper.c -- High level tape support module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the implementation for the function offered
 *    by the taper module. The taper module offers a high level
 *    tape interface.
 * NOTES
 *    - The idea is to keep the tape fd in use as long as we can, so
 *      that we can cache the tape label and position for more efficient
 *      operations.
 * BUGS
 *    - Needs testing, then cleanup, then better error handling.
 *    - Change the WARN fd NOT open into a fatal error.
 * TODO
 *    - Optimize the positionning by knowing where we are: for
 *      example appending a part, if we didn't move, should be quick.
 *      Maybe just add a quick double-check that the tape agrees with what
 *      we think is its positionning.
 *    - More precize errors (positionning, etc).
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: taper.c,v 1.8 2000/08/06 18:45:17 schaefer Exp $
 */

#include <stdlib.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <stdarg.h>
#include "global.h"
#include "debug_interface.h"
#include "tapeio_interface.h"
#include "changer_interface.h" /* for now. Later it will be IPC */
#include "buffer_interface.h"
#include "taper_interface.h"

/* Private type definitions */

#define MAX_WAIT_TIMEOUT 180 /* Wait for tape ready */

typedef struct {
   char *tape_device_name;
   changer_context_t changer_context;

   fd_t fd; /* not -1 if open */
} taper_info_t;

/* Private function definitions */

/* NAME
 *    wait_for_tape_ready
 * DESCRIPTION
 *    Waits upto timeout seconds for the tape to become ready.
 * RESULT
 *    fd_t not -1 if tape ready. -1 if timeout occured.
 * NOTES
 *    - If read_only, returns a read-only descriptor.
 * BUGS
 *    - Will always return a read/write descriptor since we can't
 *      seem to be able to change that state (see tapeio_interface.h).
 * TODO
 */
static fd_t wait_for_tape_ready(const char *tape_device_name,
                                unsigned int timeout,
                                BOOL read_only);

/* NAME
 *    get_label
 * DESCRIPTION
 *    Shows the tape label if it is valid.
 * RESULT
 *    char * (label) to be freed, or NULL,
 * NOTES
 * BUGS
 * TODO
 */
static char *get_label(fd_t tape_fd);

/* NAME
 *    write_label
 * DESCRIPTION
 *    Writes label to the tape.
 * RESULT
 * NOTES
 * BUGS
 * TODO
 */
static BOOL write_label(fd_t tape_fd, const char *label);

/* NAME
 *    do_scan_for_tape
 * DESCRIPTION
 *    Browse the slots looking for the specified tape. If label is NULL,
 *    just browse all the slots. If finds the specified tape (not NULL),
 *    opens the handle in a way which makes it secure.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE
 *          Loaded.
 *       CHANGER_ERROR_NO_TAPE
 *          Definitely not found
 *    other errors from changer_load_tape, including
 *       CHANGER_ERROR_MEDIUM_BUSY
 *       CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 * NOTES
 *    - If CHANGER_ERROR_NONE, the tape descriptor is now valid
 *      in info (and should be closed if no longer used, but then
 *      it makes the tape search invalid anyway).
 * BUGS
 * TODO
 */
static changer_error_t do_scan_for_tape(taper_info_t *info,
                                        const char *tape_label);

/* NAME
 *    load_and_verify_tape
 * DESCRIPTION
 *    Loads the specified tape and verify its label in a way which
 *    makes it secure.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE
 *          Loaded and matches.
 *       CHANGER_ERROR_LABEL_NOT_KNOWN
 *          Doesn't match and changer database updated, or not known
 *          in changer database, requires a scan.
 *       other errors from changer_load_tape, including
 *          CHANGER_ERROR_MEDIUM_BUSY
 *          CHANGER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 * NOTES
 *    - If CHANGER_ERROR_NONE, the tape descriptor is now valid
 *      in info (and should be closed if no longer used, but then
 *      it makes the tape search invalid anyway).
 * BUGS
 * TODO
 */
static changer_error_t load_and_verify_tape(taper_info_t *info,
     				            const char *tape_label);

/* NAME
 *    verify_tape_label
 * DESCRIPTION
 *    Opens the tape device, rewind, verify the label, and if it matches,
 *    keep the device open.
 * RESULT
 *    changer_error_t
 *       CHANGER_ERROR_NONE
 *          Loaded and matches.
 *       CHANGER_ERROR_LABEL_NOT_KNOWN
 *          Doesn't match and changer database updated, or not known
 *          in changer database, requires a scan.
 * NOTES
 *    - tape_label can be NULL, in that case we simply tell the changer
 *      subsystem about the labels found and return
 *      CHANGER_ERROR_LABEL_NOT_KNOWN.
 *    - If CHANGER_ERROR_NONE, the tape descriptor is now valid
 *      in info (and should be closed if no longer used, but then
 *      it makes the tape search invalid anyway).
 * BUGS
 * TODO
 */
static changer_error_t verify_tape_label(taper_info_t *info,
                                         const char *tape_label);

/* Function implementation */

taper_context_t taper_initialize(const char *tape_device_name,
                                 const changer_context_t changer_context) {
   taper_info_t *info = (taper_info_t *) malloc(sizeof(taper_info_t));

   if (info) {
      info->fd = -1;
      info->tape_device_name = strdup(tape_device_name);
      if (info->tape_device_name) {
         info->changer_context = changer_context;        
      }
      else {
         taper_deinitialize((taper_context_t *) &info);
      }
   }

   return info;
}

void taper_deinitialize(taper_context_t *context) {
   taper_info_t *info = (taper_info_t *) *context;

   if (info->tape_device_name) {
      free(info->tape_device_name);
   }

   if (info->fd != -1) {
      if (close(info->fd)) {
         debug_perror(DEBUG_LEVEL_ERROR, "closing tape fd");
      }

      info->fd = -1;
   }

   free(*context);
   *context = NULL;
}

const char *taper_error_description(const taper_error_t error) {
   switch (error) {
      case TAPER_ERROR_NONE:
         return "NONE";
         break;
      case TAPER_ERROR_NO_TAPE:
         return "No tape";
         break;
      case TAPER_ERROR_LABEL_NOT_KNOWN:
         return "No such tape";
         break;
      case TAPER_ERROR_LABEL_EXISTS:
         return "Can't overwrite valid label";
         break;
      case TAPER_ERROR_MEDIUM_BUSY:
         return "Medium busy";
         break;
      case TAPER_ERROR_IO:
         return "I/O error";
         break;
      case TAPER_ERROR_FS_FULL:
         return "Filesystem full";
         break;
      case TAPER_ERROR_SAFE_INCOMPLETE_IO:
         return "Safe incomplete I/O";
         break;
      case TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED:
         return "Operator intervention required";
         break;
      default:
         return "UNKNOWN";
         break;
   }
   /* NOT REACHED */
}

taper_error_t taper_load_tape(taper_context_t context,
		              const char *tape_label) {
   taper_info_t *info = (taper_info_t *) context;
   changer_error_t changer_error = CHANGER_ERROR_LABEL_NOT_KNOWN;
   taper_error_t taper_error = TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED;

   if (info->fd != -1) {
      if (close(info->fd)) {
         debug_perror(DEBUG_LEVEL_ERROR, "closing tape fd");
      }

      info->fd = -1;
   }

   if (tape_label) {   
      changer_error = load_and_verify_tape(info,
					   tape_label);
   }
      
   if (changer_error == CHANGER_ERROR_NONE) {
      debug_log(DEBUG_LEVEL_INFO,
                "Successfully loaded %s in %s.\n",
                tape_label,
                info->tape_device_name);
      taper_error = TAPER_ERROR_NONE;
   }
   else if (changer_error == CHANGER_ERROR_LABEL_NOT_KNOWN) {
      debug_log(DEBUG_LEVEL_INFO,
                "Changer doesn't know yet about this tape.\n");
      changer_error = do_scan_for_tape(info,
                                       tape_label);
      if (changer_error == CHANGER_ERROR_NONE) {
         debug_log(DEBUG_LEVEL_INFO,
                   "Successfully loaded %s in %s.\n",
                   tape_label,
                   info->tape_device_name);
   	 
         taper_error = TAPER_ERROR_NONE;
      }
      else if (changer_error == CHANGER_ERROR_NO_TAPE) {
         if (tape_label) {
	    debug_log(DEBUG_LEVEL_ERROR,
		      "Definitely not finding tape %s.\n",
		      tape_label);
         }
         taper_error = TAPER_ERROR_LABEL_NOT_KNOWN;
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR,
                   "Error %d while scanning tape %s: %s.\n",
                   changer_error,
                   tape_label,
                   changer_error_description(changer_error));
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "Error %d while loading tape %s: %s.\n",
                changer_error,
                tape_label,
                changer_error_description(changer_error));
      if (changer_error == CHANGER_ERROR_MEDIUM_BUSY) {
         taper_error = TAPER_ERROR_MEDIUM_BUSY;
      }
   }

   return taper_error;
}

taper_error_t taper_unload_tape(taper_context_t context) {
   taper_info_t *info = (taper_info_t *) context;
   changer_error_t changer_error;
   taper_error_t taper_error = TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED;

   if (info->fd != -1) {
      if (close(info->fd)) {
         debug_perror(DEBUG_LEVEL_ERROR, "closing tape fd");
      }

      info->fd = -1;
   }   

   changer_error = changer_unload_tape(info->changer_context,
                                       info->tape_device_name);
   if (changer_error == CHANGER_ERROR_NONE) {
      taper_error = TAPER_ERROR_NONE;
   }

   return taper_error;
}

taper_error_t taper_label_tape(taper_context_t context,
                               const unsigned int slot,
                               const char *tape_label,
                               BOOL force) {
   taper_info_t *info = (taper_info_t *) context;
   changer_error_t changer_error;
   taper_error_t taper_error = TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED;

   if (info->fd != -1) {
      debug_log(DEBUG_LEVEL_WARN, "tlt: fd open.\n");
   }

   changer_error = changer_load_specific(info->changer_context,
                                         info->tape_device_name,
                                         slot);
   if (changer_error == CHANGER_ERROR_NONE) {
      if ((info->fd
           = wait_for_tape_ready(info->tape_device_name,
                                 MAX_WAIT_TIMEOUT,
                                 FALSE))
          != -1) {
         debug_log(DEBUG_LEVEL_WARN, "No previous label verification.\n");
      	 
         debug_log(DEBUG_LEVEL_INFO,
                   "Writing label %s.\n",
                   tape_label);
         if (write_label(info->fd, tape_label)) {
            char *verify = get_label(info->fd);
      	 
            if (verify) {
               if (!strcmp(tape_label, verify)) {
                  debug_log(DEBUG_LEVEL_INFO,
                            "Successfully wrote and verified %s.\n",
                            tape_label);
                  changer_error
		     = changer_tell_label(info->changer_context,
					  tape_label,
					  info->tape_device_name);
		  if (changer_error != CHANGER_ERROR_NONE) {
		     debug_log(DEBUG_LEVEL_ERROR,
			       "Error %d while telling new label: %s.\n",
			       changer_error,
			       changer_error_description(changer_error));
		  }
                  /* Error ignored: not a problem */

                  taper_error = TAPER_ERROR_NONE;
               }
               else {
                  debug_log(DEBUG_LEVEL_INFO,
                            "Tape label mismatch (%s/%s).\n", 
                            tape_label,
                            verify);
               }
      	 
               free(verify);
            }
         }
         else {
            debug_log(DEBUG_LEVEL_ERROR, "Can't write label.\n");
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "Tape doesn't become ready.\n");
         taper_error = TAPER_ERROR_NO_TAPE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "Error %d while going to slot %d: %s.\n",
                changer_error,
                slot,
                changer_error_description(changer_error));
   }

   if (taper_error == TAPER_ERROR_NONE) {
      if (info->fd != -1) {
         if (!tapeio_enable_writing(info->fd, FALSE)) {
            debug_log(DEBUG_LEVEL_WARN, "disable write failed.\n");
            /* don't tell error */
         }
      }
   }
   else {
      if (info->fd != -1) {
	 if (close(info->fd)) {
	    debug_perror(DEBUG_LEVEL_ERROR, "closing tape fd");
	 }

	 info->fd = -1;
      }
   }

   return taper_error;
}

taper_error_t taper_write_part(taper_context_t context,
                               const char *part_file_name,
                               taper_file_number_t *file_number,
                               size_t *residual) {
   fd_t input_fd;
   taper_info_t *info = (taper_info_t *) context;
   taper_error_t result = TAPER_ERROR_IO;

   if (info->fd == -1) {
      debug_log(DEBUG_LEVEL_WARN, "twp: fd NOT open.\n");
   }

   if (tapeio_set_position_for_appending(info->fd)) {
      debug_log(DEBUG_LEVEL_INFO, "positionned ok.\n");

      if (tapeio_enable_writing(info->fd, TRUE)) {
	 input_fd = open(part_file_name, O_RDONLY | O_EXCL);
	 if (input_fd != -1) {
	    int the_errno;
	    BOOL input_error;

	    debug_log(DEBUG_LEVEL_INFO, "Started I/O.\n");
	    if (buffer_do_io(input_fd, info->fd, &the_errno, &input_error)) {
	       debug_log(DEBUG_LEVEL_INFO, "Completed I/O.\n");
	       if (tapeio_write_file_mark(info->fd)) {
		  debug_log(DEBUG_LEVEL_INFO, "Wrote filemarks.\n");
		  if (tapeio_get_position(info->fd,
					  file_number)) {
		     debug_log(DEBUG_LEVEL_INFO, "Got position.\n");
                     (*file_number)--; /* compensate for filemark */
		     result = TAPER_ERROR_NONE;
		  }
		  else {
		     debug_log(DEBUG_LEVEL_ERROR, "getting position.\n");
		  }
	       }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR, "writing filemark.\n");
	       }
	    }
	    else {
	       debug_log(DEBUG_LEVEL_ERROR,
			 "buffer I/O error: %s [%d] %s.\n",
			 strerror(the_errno),
			 the_errno,
			 input_error ? "IN"
				     : "OUT");
	    }

	    if (close(input_fd)) {
	       debug_perror(DEBUG_LEVEL_ERROR, "close()");
	       result = TAPER_ERROR_IO;
	    }
	 }
	 else {
	    debug_perror(DEBUG_LEVEL_ERROR, "open()");
	 }

         if (!tapeio_enable_writing(info->fd, FALSE)) {
            debug_log(DEBUG_LEVEL_WARN, "disable write failed.\n");
            /* don't tell error */
         }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "write enable failed.\n");
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "positionning for appending failed.\n");
   }


   *residual = 0; /* for now */
   return result;
}

taper_error_t taper_read_part(taper_context_t context,
                              const char *part_file_name,
                              taper_file_number_t file_number) {
   fd_t output_fd;
   taper_info_t *info = (taper_info_t *) context;
   taper_error_t result = TAPER_ERROR_IO;

   if (info->fd == -1) {
      debug_log(DEBUG_LEVEL_WARN, "trp: fd NOT open.\n");
   }

   if (tapeio_set_position(info->fd,
                           file_number)) {
      output_fd = open(part_file_name, O_WRONLY | O_CREAT | O_EXCL);
      if (output_fd != -1) {
         int the_errno;
         BOOL input_error;

	 if (buffer_do_io(info->fd, output_fd, &the_errno, &input_error)) {
	    result = TAPER_ERROR_NONE;
	 }
	 else {
	    debug_log(DEBUG_LEVEL_ERROR,
                      "buffer I/O error: %s [%d] %s.\n",
                      strerror(the_errno),
                      the_errno,
                      input_error ? "IN"
                                  : "OUT");
	 }

	 if (close(output_fd)) {
	    debug_perror(DEBUG_LEVEL_ERROR, "close()");
	    result = TAPER_ERROR_IO;
	 }
      }
      else {
	 debug_perror(DEBUG_LEVEL_ERROR, "open()");
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "setting position.\n");
   }

   return result;
}

/* Private function implementation */

static fd_t wait_for_tape_ready(const char *tape_device_name,
                                unsigned int timeout,
                                BOOL read_only) {
   /* BUGS
    *    - read_only doesn't work if we want to change our mind later
    *      without losing the fd and risking rewind + stuff.
    */
   read_only = FALSE;

   /* BUGS
    *    - Not exactly a real precize timeout.
    *    - Not the right way to wait for tape ready.
    *    - Function should be implemented by tapeio.
    */

   while (timeout--) {
      fd_t tape_fd = tapeio_open(tape_device_name, read_only);

      if (tape_fd != -1) {
	 if (tapeio_rewind(tape_fd)) {
            return tape_fd;
	 }

         close(tape_fd); /* result ignored */
      }
   
      sleep(1); /* don't really care about signals */
   }

   return (-1);
}

static char *get_label(fd_t tape_fd) {
   char *the_label = NULL;

   if (tape_fd == -1) {
      debug_log(DEBUG_LEVEL_WARN, "gl: invalid fd.\n");      
   }
   else {
      tapeio_label_t label;

      if (tapeio_rewind(tape_fd)) {
	 if (tapeio_label_read(tape_fd, &label)) {
            the_label = (char *) malloc(strlen(label.name)
                                        + strlen(label.date)
                                        + 2);
            if (the_label) {
               strcpy(the_label, label.name);
               the_label[strlen(label.name)] = ' ';
               strcpy(the_label + 1 + strlen(label.name), label.date);
            }

            tapeio_label_free(&label);
	 }
	 else {
	    debug_log(DEBUG_LEVEL_ERROR, "Can't read label.\n");
	 }
      }
      else {
         debug_log(DEBUG_LEVEL_ERROR, "Can't rewind.\n");
      }
   }

   return the_label;
}

static BOOL write_label(fd_t tape_fd, const char *label) {
   BOOL result = FALSE;

   if (tape_fd == -1) {
      debug_log(DEBUG_LEVEL_WARN, "wl: invalid fd.\n");      
   }
   else {
      tapeio_label_t the_label;

      the_label.name = strdup(label);
      if (the_label.name) {
	 the_label.date = strchr(the_label.name, ' ');

	 if (the_label.date) {
	    *the_label.date = '\0';
	    the_label.date++;
	    if (tapeio_rewind(tape_fd)) {
	       if (tapeio_label_write(tape_fd, &the_label)) {
 		  if (tapeio_write_file_mark(tape_fd)) {
		     result = TRUE;
		     /* No free. */
                  }
                  else {
		     debug_log(DEBUG_LEVEL_ERROR, "Can't write filemark.\n");
                  }
	       }
	       else {
		  debug_log(DEBUG_LEVEL_ERROR, "Can't write label.\n");
	       }
	    }
	    else {
	       debug_log(DEBUG_LEVEL_ERROR, "Can't rewind.\n");
	    }
	 }
	 else {
	    debug_perror(DEBUG_LEVEL_ERROR, "Can't split label");
	 }

	 free(the_label.name);
      }
      else {
	 debug_perror(DEBUG_LEVEL_ERROR, "Can't strdup() label");
      }
   }

   return result;
}

static changer_error_t do_scan_for_tape(taper_info_t *info,
                                        const char *tape_label) {
   unsigned int initial_slot_identification;
   unsigned int slot_identification;
   changer_error_t changer_error;
   BOOL found = FALSE;

   if (info->fd != -1) {
      debug_log(DEBUG_LEVEL_WARN, "dsft: fd open.\n");
   }
   
   changer_error = changer_next_tape(info->changer_context,
                                     info->tape_device_name,
                                     TRUE,
                                     NULL,
                                     &initial_slot_identification);
   
   slot_identification = initial_slot_identification;
   
   while ((changer_error == CHANGER_ERROR_NONE) && !found) {
      debug_log(DEBUG_LEVEL_INFO,
                "Changed to element #%d.\n",
                slot_identification);

      changer_error = verify_tape_label(info, tape_label);
      if (changer_error == CHANGER_ERROR_LABEL_NOT_KNOWN) {
	 changer_error = changer_next_tape(info->changer_context,
					   info->tape_device_name,
					   FALSE,
					   &slot_identification,
					   &initial_slot_identification);
      }
      else {
         found = TRUE;
      }
   }
   
   if ((changer_error == CHANGER_ERROR_NO_TAPE)
       || (found && (changer_error == CHANGER_ERROR_NONE))) {
      changer_dump_element_status(info->changer_context);
      debug_log(DEBUG_LEVEL_INFO, "Seen all full slots.\n");

      if (found) {
         changer_error = CHANGER_ERROR_NONE;
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR,
                "Error %d while changing tapes: %s.\n",
                changer_error,
                changer_error_description(changer_error));
   }

   return changer_error;
}

static changer_error_t load_and_verify_tape(taper_info_t *info,
     				            const char *tape_label) {
   changer_error_t changer_error
      = changer_load_tape(info->changer_context,
                          tape_label,
                          info->tape_device_name);

   if (changer_error == CHANGER_ERROR_NONE) {
      return verify_tape_label(info, tape_label);
   }

   return changer_error;
}

static changer_error_t verify_tape_label(taper_info_t *info,
                                         const char *tape_label) {
   BOOL found = FALSE;

   if (info->fd != -1) {
      debug_log(DEBUG_LEVEL_WARN, "vtl: fd open.\n");
   }

   if ((info->fd = wait_for_tape_ready(info->tape_device_name,
                                       MAX_WAIT_TIMEOUT,
                                       TRUE))
       != -1) {
      char *the_label = get_label(info->fd);
   
      if (the_label) {
         changer_error_t changer_error;

         debug_log(DEBUG_LEVEL_INFO,
                   "Tape label: %s.\n", 
                   the_label);
   
         changer_error = changer_tell_label(info->changer_context,
                                            the_label,
                                            info->tape_device_name);   
         if (changer_error != CHANGER_ERROR_NONE) {
            debug_log(DEBUG_LEVEL_ERROR,
                      "Error %d while telling label: %s.\n",
                      changer_error,
                      changer_error_description(changer_error));
         }
   
         if (tape_label && !strcmp(the_label, tape_label)) {
            found = TRUE;
         }
   
         free(the_label);
      }
   }
   else {
      debug_log(DEBUG_LEVEL_ERROR, "Tape doesn't become ready.\n");
   }

   if (found) {
      return CHANGER_ERROR_NONE;
   }
   else {
      if (info->fd != -1) {
         close(info->fd); /* res. ign. */
         info->fd = -1;
      }
      return CHANGER_ERROR_LABEL_NOT_KNOWN;
   }
   /* NOT REACHED */   
}

