/* Project filed -- Intelligent file cache and migration system
 * taper_interface.h -- Interface definitions for the taper module.
 * Marc SCHAEFER <schaefer@alphanet.ch>
 * Created: Sun Apr 9 2000
 * LICENSE
 *    (C) 2000 by Marc SCHAEFER, licensed under the terms of the
 *     GNU General Public License as published by the Free Software
 *     Foundation; either version 2 or later (at your option).
 * DESCRIPTION
 *    This file contains the type and function definitions for the
 *    taper module. The taper modules offers high-level tape I/O services.
 * NOTES
 *    - Taper operations are done within a taper process, which communicates
 *      with the server through pipes, in the same way as the dumpers.
 *    - The idea being that the taper process can cache status for its
 *      drive, such as current position, current tape, etc, and is also
 *      supposed to request tape changes from the changer subsystem.
 *    - Communication with the server is done by the module using this
 *      module. The first version will, like the test_chg program in
 *      ../changer, allow us to develop and test taper independantly from
 *      the server.
 * BUGS
 *    - Cleaning ? Idea: taper would generate a cleaning exception
 *      when required, and scheduler will not give another job to
 *      the tape process, but instead tell it to clean when finished, which
 *      would communicate with the taper. See also changer.
 * TODO
 * BASED-ON
 *    Original work.
 * MODIFICATION-HISTORY
 * $Id: taper_interface.h,v 1.6 2000/08/06 09:32:48 schaefer Exp $
 */

#ifndef _TAPER_INTERFACE_H
#define _TAPER_INTERFACE_H

/* Type definitions */

typedef enum {
   TAPER_ERROR_NONE,
   TAPER_ERROR_NO_TAPE,
   TAPER_ERROR_LABEL_NOT_KNOWN,
   TAPER_ERROR_LABEL_EXISTS,
   TAPER_ERROR_MEDIUM_BUSY,
   TAPER_ERROR_IO,
   TAPER_ERROR_FS_FULL,
   TAPER_ERROR_SAFE_INCOMPLETE_IO,
   TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED
} taper_error_t;

typedef unsigned int taper_file_number_t;

typedef void * taper_context_t;

/* Function definitions */

/* NAME
 *    taper_initialize
 * DESCRIPTION
 *    Initializes a taper instance. Passed is the tape_device_name,
 *    which describes a tape device name.
 * RESULT
 *    taper_context_t, NULL if initialization failed.
 * NOTES
 *    - changer_context is a communication point to a changer device.
 *    - This doesn't open the tape device. Look for other functions.
 * BUGS
 * TODO
 */
taper_context_t taper_initialize(const char *tape_device_name,
                                 const changer_context_t changer_context);

/* NAME
 *    taper_deinitialize
 * DESCRIPTION
 *    Deinitializes a taper module instance.
 * RESULT
 *    NONE
 * NOTES
 *    - context will be set to NULL
 *    - Closes the tape drive descriptor if required.
 * BUGS
 * TODO
 */
void taper_deinitialize(taper_context_t *context);

/* NAME
 *    taper_error_description
 * DESCRIPTION
 *    Returns an error description for the given taper error.
 * RESULT
 *    const char *
 * NOTES
 * BUGS
 * TODO
 */
const char *taper_error_description(const taper_error_t error);

/* NAME
 *    taper_load_tape
 * DESCRIPTION
 *    Loads the specified tape into this taper drive. This will also
 *    unload a present tape as required.
 * RESULT
 *    taper_error_t
 *       TAPER_ERROR_NONE  The named tape was loaded properly.
 *       TAPER_ERROR_LABEL_NOT_KNOWN
 *                         The requested label cannot be found currently.
 *       TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 *                         A hardware error occured.
 *       TAPER_ERROR_MEDIUM_BUSY
 *                         The medium is in use in another drive. The caller
 *                         program should have known that.
 * NOTES
 *    - If TAPER_ERROR_NONE, the tape descriptor is now open internally
 *      and I/O can be done through the taper_io_*() functions, until
 *      taper_unload_tape(), taper_load_tape() or other things.
 *    - If tape_label is specified as NULL, will just browse through all
 *      the available tapes, telling the changer about their labels.
 * BUGS
 * TODO
 */
taper_error_t taper_load_tape(taper_context_t context,
		              const char *tape_label);

/* NAME
 *    taper_unload_tape
 * DESCRIPTION
 *    Forces the tape to be removed from the drive.
 * RESULT
 *    taper_error_t
 *       TAPER_ERROR_NONE  Operation was successfully executed.
 *       TAPER_ERROR_NO_TAPE
 *                           There was no tape in the drive.
 *       TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 *                           A hardware error occured.
 *       TAPER_ERROR_DRIVE_NOT_KNOWN
 *                           Unknown drive in this taper.
 * NOTES
 *    - This is normally not needed since taper_load_tape() will
 *      automatically unload the current tape. One reason to call this
 *      is for example to ensure it's not in the drive for safety
 *      reasons, e.g.
 *    - Closes the tape drive descriptor if required.
 * BUGS
 * TODO
 */
taper_error_t taper_unload_tape(taper_context_t context);

/* NAME
 *    taper_label_tape
 * DESCRIPTION
 *    Labels the tape in the named slot. If successful, this function
 *    will put the module in the same state as if taper_load_tape()
 *    had been executed.
 * RESULT
 *    taper_error_t
 *       TAPER_ERROR_NONE
 *          Operation was successful
 *       TAPER_ERROR_NO_TAPE
 *          There was no tape in the specified slot.
 *       TAPER_ERROR_OPERATOR_INTERVENTION_REQUIRED
 *          An operator intervention is required (hardware error).
 *       TAPER_ERROR_LABEL_EXISTS
 *          Label exists already, and no force action specified.
 * NOTES
 *    - If force is set, will overwrite a filed label.
 * BUGS
 * TODO
 */
taper_error_t taper_label_tape(taper_context_t context,
                               const unsigned int slot,
                               const char *tape_label,
                               BOOL force);

/* NAME
 *    taper_write_part
 * DESCRIPTION
 *    Transfers a filesystem file at the end of a tape file. Sets the
 *    file number to the actual file number. If the tape drive reports
 *    EOM, and the residual (unwritten size) can be safely determined, sets
 *    *residual and return TAPER_ERROR_SAFE_INCOMPLETE_IO. Else,
 *    return TAPER_ERROR_IO. On a successful transfer, flushes the
 *    data and writes a file mark.
 * RESULT
 *    taper_error_t
 *       TAPER_ERROR_NONE
 *          Transfer successful
 *       TAPER_ERROR_IO
 *          IO Error (e.g. unsafe EOM)
 *       TAPER_ERROR_SAFE_INCOMPLETE_IO
 *          Safe EOM, see *residual
 * NOTES
 * BUGS
 * TODO
 */
taper_error_t taper_write_part(taper_context_t context,
                               const char *part_file_name,
                               taper_file_number_t *file_number,
                               size_t *residual);

/* NAME
 *    taper_read_part
 * DESCRIPTION
 *    Transfers a numbered tape file into the named file.
 * RESULT
 *    taper_error_t
 *       TAPER_ERROR_NONE
 *          Transfer successful
 *       TAPER_ERROR_IO
 *          IO Error
 *       TAPER_ERROR_FS_FULL
 *          The destination filesystem is full.
 * NOTES
 * BUGS
 * TODO
 */
taper_error_t taper_read_part(taper_context_t context,
                              const char *part_file_name,
                              taper_file_number_t file_number);

#endif /* !_TAPER_INTERFACE_H */
